# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import serial

from htf.communication.modbus.exceptions import ModbusNoResponseException
from htf.core import __version__ as htf_version
from htf.communication.modbus.interface import ModbusSerialRTUClientInterface, ModbusTCPClientInterface, \
    ModbusUDPClientInterface, ModbusTLSClientInterface, ModbusSerialASCIIClientInterface
from htf.communication.modbus.protocol import SerialRTUModbusRequest, SerialRTUModbusResponse, \
    TCPModbusRequest, TCPModbusResponse, ModbusException, SerialASCIIModbusRequest, SerialASCIIModbusResponse
from htf.communication.modbus.core import ModbusBaseClient as _ModbusBaseClient
import threading
import oser
import sys
import math

BYTE_SIZE = 8


class ModbusBaseClient(_ModbusBaseClient):

    def __init__(self,
                 interface,
                 request_parser,
                 response_parser,
                 retries=None,
                 references_start_at_one=True,
                 debuglevel=None):
        """
        Args:
            interface: an interface to be used for communication
            request_parser: a parser for requests
            response_parser: a parser for responses
            retries=None (int or None): the number of retries until an exception is raised
            references_start_at_one=True: if set to ``True`` all references start at ``1`` else ``0``
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        self.request_parser = request_parser
        self.response_parser = response_parser
        self.interface = interface
        self.retries = retries
        self.reference_offset = int(references_start_at_one)
        self.debuglevel = debuglevel

        self.query_lock = threading.Lock()

    def read(self):
        """
        Read a message from the interface and decode it.

        Returns:
            adu: the decoded adu as is

        Raises:
            Exception if no answer was received
        """
        adu = self.response_parser()
        data, address = self.interface.read()
        if data is None or data == b"":
            raise ModbusNoResponseException("No answer was received!")
        adu.decode(data)
        return adu

    def write(self, adu):
        """
        Write ``adu`` encoded using the interface.

        Args:
            adu: a serializable adu
        """
        self.interface.write(adu)

    def query(self, adu):
        """
        Query ``adu`` using the interface and return the decoded response.
        This method is thread-safe.

        Args:
            adu: a serializable adu

        Returns:
            adu: the decoded response if no exception occurred

        Raises:
            ModbusException: in case of an exception response
        """
        with self.query_lock:
            if self.debuglevel:  # pragma: no cover
                # debug
                print("=" * 80, file=sys.stderr)
                print("Client request:", file=sys.stderr)
                if self.debuglevel == 1:
                    print(adu, file=sys.stderr)
                else:
                    print(adu.introspect(), file=sys.stderr)

            retries = self.retries or 0

            while retries >= -1:
                self.write(adu)
                try:
                    retries -= 1
                    response = self.read()
                except Exception as e:
                    if retries >= 0:  # pragma: no cover
                        if self.debuglevel:
                            print("{}: {}".format(e.__class__.__name__, str(e)), file=sys.stderr)
                            print("retrying ({} tries left)".format(retries + 1), file=sys.stderr)
                        continue
                    else:
                        raise
                else:  # pragma: no cover
                    break

            if self.debuglevel:  # pragma: no cover
                # debug
                print("Client response:", file=sys.stderr)
                if self.debuglevel == 1:
                    print(response, file=sys.stderr)
                else:
                    print(response.introspect(), file=sys.stderr)

            if response.pdu.data.exception_code.get() is None:
                return response
            else:
                raise ModbusException('{} (Code: {})'.format(response.pdu.data.exception_code.get(),
                                                             response.pdu.data.exception_code.get_value()))

    def build_request(self):
        """
        Build and return an instance of the desired request parser.

        Returns:
            request parser
        """
        return self.request_parser()

    def build_response(self):  # pragma: no cover
        """
        Build and return an instance of the desired response parser.

        Returns:
            response parser
        """
        return self.response_parser()

    def close(self):
        """
        Close the interface.
        """
        self.interface.close()

    # Coils
    # =====

    def read_coil(self, reference):
        """
        Read a single coil using Modbus function code 1.

        Args:
            reference (int): the register reference.

        Returns:
            int: ``1`` if the coil is set and ``0`` else.
        """
        return self.read_multiple_coils(reference, 1)[0]

    def write_coil(self, reference, value):
        """
        Write a single coil using Modbus function code 5.

        Args:
            reference (int): the register reference.
            value (int): the value to be written.
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()

        adu.pdu.function_code.set(5)
        adu.pdu.data.output_address.set(reference - self.reference_offset)
        if value != 0:
            value = 0xff00
        else:
            value = 0x0000
        adu.pdu.data.output_value.set(value)
        return self.query(adu)

    def read_multiple_coils(self, reference, quantity):
        """
        Read multiple coils using Modbus function code 1.

        Args:
            reference (int): the register reference.
            quantity (int): the number of coils to be read.

        Returns:
            list of int: each element is ``1`` if the corresponding coil is set and ``0`` else.
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()
        adu.pdu.function_code.set(1)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_coils.set(quantity)
        resp = self.query(adu)

        coil_status = []
        for byte in resp.pdu.data.coil_status:
            for i in range(8):
                coil_status.append((byte.get() >> i) & 1)
        coil_status = coil_status[:quantity]
        return coil_status

    def write_multiple_coils(self, reference, values):
        """
        Write multiple coils using Modbus function code 15.

        Args:
            reference (int): the register reference.
            values (list of int): the values to be written.
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        quantity = len(values)
        adu = self.build_request()

        adu.pdu.function_code.set(15)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_outputs.set(quantity)
        bytes_required = int(math.ceil(quantity / BYTE_SIZE))
        adu.pdu.data.byte_count.set(bytes_required)

        value_slices = [values[i * BYTE_SIZE:(i + 1) * BYTE_SIZE]
                        for i in range((len(values) + (BYTE_SIZE - 1)) // BYTE_SIZE)]
        outputs_ints = []
        outputs_value = 0
        for s in value_slices:
            for val in s:
                outputs_value = outputs_value << 1
                outputs_value |= val
            outputs_ints.append(outputs_value)
            outputs_value = 0
        for i in range(len(outputs_ints)):
            adu.pdu.data.outputs_value[i] = outputs_ints[i]

        return self.query(adu)

    def write_coil_data(self, reference, data):
        """
        Write data to coils using Modbus function code 15.

        Args:
            reference (int): the coil reference.
            data (oser.ByteStruct): the type or instance or the decoder for the data
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = max_byte_size * 8

        adu = self.build_request()
        adu.pdu.function_code.set(15)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_outputs.set(quantity)
        adu.pdu.data.byte_count.set(max_byte_size)
        adu.pdu.data.outputs_value.decode(data.encode())
        return self.query(adu)

    def read_coil_data(self, reference, data):
        """
        Read data from coils using Modbus function code 1.

        Args:
            reference (int): the coil reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)

        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = max_byte_size * 8

        adu = self.build_request()
        adu.pdu.function_code.set(1)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_coils.set(quantity)
        response = self.query(adu)

        data.decode(response.pdu.data.coil_status.encode())

        return data

    # Discrete Inputs
    # ===============

    def read_discrete_input(self, reference):
        """
        Read single discrete input using Modbus function code 2.

        Args:
            reference (int): the register reference.

        Returns:
            int: ``1`` if the discete input is set and ``0`` else.
        """
        return self.read_multiple_discrete_inputs(reference, 1)[0]

    def read_multiple_discrete_inputs(self, reference, quantity):
        """
        Read multiple discrete inputs using Modbus function code 2.

        Args:
            reference (int): the register reference.
            quantity (int): the number of discrete inputs to be read.

        Returns:
            list of int: each element is ``1`` if the corresponding discrete input is set and ``0`` else.
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()
        adu.pdu.function_code.set(2)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_inputs.set(quantity)
        resp = self.query(adu)

        input_status = []
        for byte in resp.pdu.data.input_status:
            for i in range(8):
                input_status.append((byte.get() >> i) & 1)
                input_status = input_status[:quantity]
        return input_status

    def read_discrete_input_data(self, reference, data):
        """
        Read data from discrete inputs using Modbus function code 2.

        Args:
            reference (int): the discrete input reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)

        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = max_byte_size * 8

        adu = self.build_request()
        adu.pdu.function_code.set(2)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_inputs.set(quantity)
        response = self.query(adu)
        data.decode(response.pdu.data.input_status.encode())

        return data

    # Holding Registers
    # =================

    def read_holding_register(self, reference):
        """
        Read a single holding register using Modbus function code 3.

        Args:
            reference (int): the register reference.

        Returns:
            int: the holding register value
        """
        return self.read_multiple_holding_registers(reference, 1)[0]

    def write_holding_register(self, reference, value):
        """
        Write single holding register using Modbus function code 6.

        Args:
            reference: the register reference.
            value: the value to be written into the corresponding register.
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()
        adu.pdu.function_code.set(6)
        adu.pdu.data.register_address.set(reference - self.reference_offset)
        adu.pdu.data.register_value.set(value)
        return self.query(adu)

    def read_multiple_holding_registers(self, reference, quantity):
        """
        Read multiple holding registers using Modbus function code 3.

        Args:
            reference (int): the register reference.
            quantity (int): the number of registers to be read.

        Returns:
            list of int: the holding register values
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()
        adu.pdu.function_code.set(3)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_registers.set(quantity)
        resp = self.query(adu)
        return [r.get() for r in resp.pdu.data.register_value.get()]

    def write_multiple_holding_registers(self, reference, values):
        """
        Write multiple holding registers using Modbus function code 16.

        Args:
            reference (int): the register reference.
            values (list of int): the values to be written.
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        quantity = len(values)
        adu = self.build_request()
        adu.pdu.function_code.set(16)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_registers.set(quantity)
        adu.pdu.data.byte_count.set(quantity * 2)
        for i in range(quantity):
            adu.pdu.data.outputs_value[i] = values[i]
        return self.query(adu)

    def read_write_multiple_holding_registers(self, read_reference, read_quantity,
                                              write_reference, write_values):
        """
        Read/Write multiple holding registers using Modbus function code 23.

        Args:
            read_reference (int): the register reference to read from.
            read_quantity (int): the number of registers to read.
            write_reference (int): the register reference to to write to.
            write_values (list of int): the values to be written.

        Returns:
            list of int: the read registers
        """
        adu = self.build_request()
        adu.pdu.function_code.set(23)
        adu.pdu.data.read_starting_address.set(read_reference)
        adu.pdu.data.quantity_to_read.set(read_quantity)

        write_quantity = len(write_values)

        adu.pdu.data.write_starting_address.set(write_reference)
        adu.pdu.data.quantity_to_write.set(len(write_values))
        adu.pdu.data.write_byte_count.set(write_quantity * 2)
        for i in range(write_quantity):
            adu.pdu.data.write_registers_value[i] = write_values[i]

        resp = self.query(adu)
        return [r.get() for r in resp.pdu.data.read_register_value.get()]

    def mask_holding_register(self, reference, and_mask, or_mask):
        """
        Mask write single holding register using Modbus function code 22.

        Args:
            reference (int): the register reference
            and_mask (int): the and mask
            or_mask (int): the or mask
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()
        adu.pdu.function_code.set(22)
        adu.pdu.data.reference_address.set(reference - self.reference_offset)
        adu.pdu.data.and_mask.set(and_mask)
        adu.pdu.data.or_mask.set(or_mask)
        return self.query(adu)

    def mask_multiple_holding_registers(self, reference, quantity, and_mask, or_mask):
        """
        Mask write multiple holding registers using Modbus function code 16.

        Args:
            reference (int): the register reference
            quantity (int): the number of registers to be masked
            and_mask (int): the and mask
            or_mask (int): the or mask
        """
        for ref in range(reference, reference+quantity+1):
            self.mask_holding_register(ref, and_mask, or_mask)

    def write_holding_register_data(self, reference, data):
        """
        Write data to holding registers using Modbus function code 16.

        Args:
            reference (int): the register reference.
            data (oser.ByteStruct): the type or instance or the decoder for the data
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = (max_byte_size + 1) // 2

        adu = self.build_request()
        adu.pdu.function_code.set(16)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_registers.set(quantity)
        adu.pdu.data.byte_count.set(quantity * 2)
        adu.pdu.data.outputs_value.decode(data.encode() + b"\x00")
        return self.query(adu)

    def read_holding_register_data(self, reference, data):
        """
        Read data from holding registers using Modbus function code 3.

        Args:
            reference (int): the register reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)

        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = (max_byte_size + 1) // 2

        adu = self.build_request()
        adu.pdu.function_code.set(3)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_registers.set(quantity)
        response = self.query(adu)

        data.decode(response.pdu.data.register_value.encode())

        return data

    # Input Registers
    # ===============

    def read_input_register(self, reference):
        """
        Read a single input register using Modbus function code 4.

        Args:
            reference (int): the register reference.

        Returns:
            int: the input register value
        """
        return self.read_multiple_input_registers(reference, 1)[0]

    def read_multiple_input_registers(self, reference, quantity):
        """
        Read multiple input registers using Modbus function code 4.

        Args:
            reference (int): the register reference.
            quantity (int): the number of registers to be read.

        Returns:
            list of int: the input register values
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)
        adu = self.build_request()
        adu.pdu.function_code.set(4)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_registers.set(quantity)
        resp = self.query(adu)
        return [r.get() for r in resp.pdu.data.input_registers.get()]

    def read_input_register_data(self, reference, data):
        """
        Read data from input registers using Modbus function code 4.

        Args:
            reference (int): the register reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        assert reference >= self.reference_offset, "references start at {}".format(self.reference_offset)

        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = (max_byte_size + 1) // 2

        adu = self.build_request()
        adu.pdu.function_code.set(4)
        adu.pdu.data.starting_address.set(reference - self.reference_offset)
        adu.pdu.data.quantity_of_registers.set(quantity)
        response = self.query(adu)
        data.decode(response.pdu.data.input_registers.encode())

        return data

    # Commands
    # ========

    def get_fifo_queue_content(self, address):
        """
        Query and return the fifo content of ``address`` using Modbus function code 24.

        Args:
             address: the fifo queue pointer address

        Returns:
            list: list of content from the fifo queue at ``address``
        """
        adu = self.build_request()
        adu.pdu.function_code.set(24)
        adu.pdu.data.fifo_pointer_address.set(address)
        response = self.query(adu)
        return [data.get() for data in response.pdu.data.fifo_value_register[:]]

    def get_device_identification(self):
        """
        Query and return the device identification using Modbus function code 43.14.

        Returns:
            dict: containing key values pairs with the device identifiers
        """
        adu = self.build_request()
        adu.pdu.function_code.set(43)
        adu.pdu.data.mei_type.set(14)

        device_identification = {}

        object_id = 0x00

        while object_id < 0xFF:  # query each object id only once  # pragma: no branch
            if 0 <= object_id <= 0x02:
                adu.pdu.data.read_device_id_code.set("basic")
            elif 3 <= object_id <= 0x7f:
                adu.pdu.data.read_device_id_code.set("regular")
            else:
                adu.pdu.data.read_device_id_code.set("extended")
            adu.pdu.data.object_id.set(object_id)  # read first basic object

            response = self.query(adu)

            # add response data objects to device identification
            for data in response.pdu.data.list_of_objects:
                device_identification[data.object_id.get()] = data.value.get().replace(b"\x00", b"")

            object_id += 1

            # parse response for more information and next data object
            if response.pdu.data.more_follows.get() == 0xff:
                object_id = response.pdu.data.next_object_id.get()
            else:
                return device_identification

        return device_identification  # pragma: no cover

    def read_extended_memory(self, memory_number, record_number, record_length):
        """
        Query and return extended memory from device using Modbus function code 20.

        Args:
            memory_number (int): the memory number (0x0001 to 0xffff)
            record_number (int): the record to start reading from
            record_length (int): the length of records

        Returns:
            list of int: the read records
        """
        assert isinstance(memory_number, int), "memory_number must be an integer"
        assert 0x0001 <= memory_number <= 0xffff, "memory_number must be an integer between 0x0001 and 0xffff"

        assert isinstance(record_number, int), "record_number must be an integer"
        assert 0x0000 <= record_number <= 0x270F, "record_number must be an integer between 0x0001 and 0x270F"

        request = self.build_request()
        request.pdu.function_code.set(20)  # read file record

        request.pdu.data.sub_requests.set_length(1)

        request.pdu.data.sub_requests[0].reference_type.set(0x06)
        request.pdu.data.sub_requests[0].memory_number.set(memory_number)
        request.pdu.data.sub_requests[0].record_number.set(record_number)
        request.pdu.data.sub_requests[0].record_length.set(record_length)

        response = self.query(request)

        return [data.get() for data in response.pdu.data.sub_responses[0].record_data]

    def write_extended_memory(self, memory_number, record_number, records):
        """
        Write extended memory using Modbus function code 21.

        Args:
            memory_number (int): the memory number (0x0001 to 0xffff)
            record_number (int): the record to write to
            records (list of int): the records to be written
        """
        assert isinstance(memory_number, int), "memory_number must be an integer"
        assert 0x0001 <= memory_number <= 0xffff, "memory_number must be an integer between 0x0001 and 0xffff"

        assert isinstance(record_number, int), "record_number must be an integer"
        assert 0x0000 <= record_number <= 0x270F, "record_number must be an integer between 0x0001 and 0x270F"

        assert all(map(lambda x: isinstance(x, int), records)), "records must be a list of integers"

        request = self.build_request()
        request.pdu.function_code.set(21)  # write file record

        request.pdu.data.sub_requests.set_length(1)

        request.pdu.data.sub_requests[0].reference_type.set(0x06)
        request.pdu.data.sub_requests[0].memory_number.set(memory_number)
        request.pdu.data.sub_requests[0].record_number.set(record_number)
        request.pdu.data.sub_requests[0].record_length.set(len(records))
        request.pdu.data.sub_requests[0].record_data[:] = records

        response = self.query(request)

        assert request.pdu.data.sub_requests == response.pdu.data.sub_responses, \
            "the echoed data do not match the sent data"

    # Convenience functions
    # =====================

    # write Flag to coil
    def write_coil_flag(self, reference, value):
        """
        Write a bit to coil memory.

        Args:
            reference: the reference to write to
            value: the bool value to be written
        """
        data = oser.Flag(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data=data)

    # read Flag from coil
    def read_coil_flag(self, reference):
        """
        Read a bit from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_coil_data(reference, data).data.get()

    # write list of Flag to coil
    def write_coil_flag_list(self, reference, values):
        """
        Write a list of bits to coil memory.

        Args:
            reference: the reference to write to
            values: a list of bool values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Flag)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data)

    # read list of Flag from coil
    def read_coil_flag_list(self, reference, quantity):
        """
        Read a list of bits from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_coil_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Nibble to coil
    def write_coil_nibble(self, reference, value):
        """
        Write a nibble to coil memory.

        Args:
            reference: the reference to write to
            value: the nibble value to be written
        """
        data = oser.Nibble(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data=data)

    # read Nibble from coil
    def read_coil_nibble(self, reference):
        """
        Read a nibble from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_coil_data(reference, data).data.get()

    # write list of Nibble to coil
    def write_coil_nibble_list(self, reference, values):
        """
        Write a list of nibbles to coil memory.

        Args:
            reference: the reference to write to
            values: a list of nibble values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Nibble)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data)

    # read list of Nibble from coil
    def read_coil_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_coil_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write BitField to coil
    def write_coil_bitfield(self, reference, value, length):
        """
        Write a bitfield to coil memory.

        Args:
            reference: the reference to write to
            value: the bitfield value to be written
            length: the length of the value
        """
        data = oser.BitField(value=value, length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data=data)

    # read BitField from coil
    def read_coil_bitfield(self, reference, length):
        """
        Read a bitfield from coil memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_coil_data(reference, data).data.get()

    # write list of BitField to coil
    def write_coil_bitfield_list(self, reference, values, length):
        """
        Write a list of bitfields to coil memory.

        Args:
            reference: the reference to write to
            values: a list of bitfield values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.BitField, kwargs={'length': length})
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data)

    # read list of BitField from coil
    def read_coil_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from coil memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_coil_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # read Flag from discrete input
    def read_discrete_input_flag(self, reference):
        """
        Read a bit from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_discrete_input_data(reference, data).data.get()

    # read list of Flag from discrete input
    def read_discrete_input_flag_list(self, reference, quantity):
        """
        Read a list of bits from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_discrete_input_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # read Nibble from discrete input
    def read_discrete_input_nibble(self, reference):
        """
        Read a nibble from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_discrete_input_data(reference, data).data.get()

    # read list of Nibble from discrete input
    def read_discrete_input_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_discrete_input_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # read BitField from discrete input
    def read_discrete_input_bitfield(self, reference, length):
        """
        Read a bitfield from discrete input memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_discrete_input_data(reference, data).data.get()

    # read list of BitField from discrete input
    def read_discrete_input_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from discrete input memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_discrete_input_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write UBInt8 to holding register
    def write_holding_register_ubint8(self, reference, value):
        """
        Write a unsigned 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt8 from holding register
    def read_holding_register_ubint8(self, reference):
        """
        Read a unsigned 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt8 to holding register
    def write_holding_register_ubint8_list(self, reference, values):
        """
        Write a list of unsigned 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt8 from holding register
    def read_holding_register_ubint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt8 to holding register
    def write_holding_register_ulint8(self, reference, value):
        """
        Write a unsigned 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt8 from holding register
    def read_holding_register_ulint8(self, reference):
        """
        Read a unsigned 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt8 to holding register
    def write_holding_register_ulint8_list(self, reference, values):
        """
        Write a list of unsigned 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt8 from holding register
    def read_holding_register_ulint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt8 to holding register
    def write_holding_register_sbint8(self, reference, value):
        """
        Write a signed 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt8 from holding register
    def read_holding_register_sbint8(self, reference):
        """
        Read a signed 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt8 to holding register
    def write_holding_register_sbint8_list(self, reference, values):
        """
        Write a list of signed 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt8 from holding register
    def read_holding_register_sbint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt8 to holding register
    def write_holding_register_slint8(self, reference, value):
        """
        Write a signed 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt8 from holding register
    def read_holding_register_slint8(self, reference):
        """
        Read a signed 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt8 to holding register
    def write_holding_register_slint8_list(self, reference, values):
        """
        Write a list of signed 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt8 from holding register
    def read_holding_register_slint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt16 to holding register
    def write_holding_register_ubint16(self, reference, value):
        """
        Write a unsigned 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt16 from holding register
    def read_holding_register_ubint16(self, reference):
        """
        Read a unsigned 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt16 to holding register
    def write_holding_register_ubint16_list(self, reference, values):
        """
        Write a list of unsigned 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt16 from holding register
    def read_holding_register_ubint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt16 to holding register
    def write_holding_register_ulint16(self, reference, value):
        """
        Write a unsigned 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt16 from holding register
    def read_holding_register_ulint16(self, reference):
        """
        Read a unsigned 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt16 to holding register
    def write_holding_register_ulint16_list(self, reference, values):
        """
        Write a list of unsigned 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt16 from holding register
    def read_holding_register_ulint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt16 to holding register
    def write_holding_register_sbint16(self, reference, value):
        """
        Write a signed 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt16 from holding register
    def read_holding_register_sbint16(self, reference):
        """
        Read a signed 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt16 to holding register
    def write_holding_register_sbint16_list(self, reference, values):
        """
        Write a list of signed 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt16 from holding register
    def read_holding_register_sbint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt16 to holding register
    def write_holding_register_slint16(self, reference, value):
        """
        Write a signed 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt16 from holding register
    def read_holding_register_slint16(self, reference):
        """
        Read a signed 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt16 to holding register
    def write_holding_register_slint16_list(self, reference, values):
        """
        Write a list of signed 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt16 from holding register
    def read_holding_register_slint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt32 to holding register
    def write_holding_register_ubint32(self, reference, value):
        """
        Write a unsigned 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt32 from holding register
    def read_holding_register_ubint32(self, reference):
        """
        Read a unsigned 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt32 to holding register
    def write_holding_register_ubint32_list(self, reference, values):
        """
        Write a list of unsigned 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt32 from holding register
    def read_holding_register_ubint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt32 to holding register
    def write_holding_register_ulint32(self, reference, value):
        """
        Write a unsigned 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt32 from holding register
    def read_holding_register_ulint32(self, reference):
        """
        Read a unsigned 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt32 to holding register
    def write_holding_register_ulint32_list(self, reference, values):
        """
        Write a list of unsigned 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt32 from holding register
    def read_holding_register_ulint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt32 to holding register
    def write_holding_register_sbint32(self, reference, value):
        """
        Write a signed 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt32 from holding register
    def read_holding_register_sbint32(self, reference):
        """
        Read a signed 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt32 to holding register
    def write_holding_register_sbint32_list(self, reference, values):
        """
        Write a list of signed 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt32 from holding register
    def read_holding_register_sbint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt32 to holding register
    def write_holding_register_slint32(self, reference, value):
        """
        Write a signed 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt32 from holding register
    def read_holding_register_slint32(self, reference):
        """
        Read a signed 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt32 to holding register
    def write_holding_register_slint32_list(self, reference, values):
        """
        Write a list of signed 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt32 from holding register
    def read_holding_register_slint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt2143 to holding register
    def write_holding_register_uint2143(self, reference, value):
        """
        Write a unsigned 2143-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt2143(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt2143 from holding register
    def read_holding_register_uint2143(self, reference):
        """
        Read a unsigned 2143-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt2143()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt2143 to holding register
    def write_holding_register_uint2143_list(self, reference, values):
        """
        Write a list of unsigned 2143-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt2143)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt2143 from holding register
    def read_holding_register_uint2143_list(self, reference, quantity):
        """
        Read a list of unsigned 2143-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt2143)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt3412 to holding register
    def write_holding_register_uint3412(self, reference, value):
        """
        Write a unsigned 3412-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt3412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt3412 from holding register
    def read_holding_register_uint3412(self, reference):
        """
        Read a unsigned 3412-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt3412()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt3412 to holding register
    def write_holding_register_uint3412_list(self, reference, values):
        """
        Write a list of unsigned 3412-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt3412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt3412 from holding register
    def read_holding_register_uint3412_list(self, reference, quantity):
        """
        Read a list of unsigned 3412-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt3412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt2143 to holding register
    def write_holding_register_sint2143(self, reference, value):
        """
        Write a signed 2143-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt2143(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt2143 from holding register
    def read_holding_register_sint2143(self, reference):
        """
        Read a signed 2143-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt2143()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt2143 to holding register
    def write_holding_register_sint2143_list(self, reference, values):
        """
        Write a list of signed 2143-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt2143)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt2143 from holding register
    def read_holding_register_sint2143_list(self, reference, quantity):
        """
        Read a list of signed 2143-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt2143)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt3412 to holding register
    def write_holding_register_sint3412(self, reference, value):
        """
        Write a signed 3412-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt3412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt3412 from holding register
    def read_holding_register_sint3412(self, reference):
        """
        Read a signed 3412-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt3412()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt3412 to holding register
    def write_holding_register_sint3412_list(self, reference, values):
        """
        Write a list of signed 3412-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt3412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt3412 from holding register
    def read_holding_register_sint3412_list(self, reference, quantity):
        """
        Read a list of signed 3412-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt3412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt64 to holding register
    def write_holding_register_ubint64(self, reference, value):
        """
        Write a unsigned 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt64 from holding register
    def read_holding_register_ubint64(self, reference):
        """
        Read a unsigned 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt64 to holding register
    def write_holding_register_ubint64_list(self, reference, values):
        """
        Write a list of unsigned 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt64 from holding register
    def read_holding_register_ubint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt64 to holding register
    def write_holding_register_ulint64(self, reference, value):
        """
        Write a unsigned 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt64 from holding register
    def read_holding_register_ulint64(self, reference):
        """
        Read a unsigned 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt64 to holding register
    def write_holding_register_ulint64_list(self, reference, values):
        """
        Write a list of unsigned 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt64 from holding register
    def read_holding_register_ulint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt64 to holding register
    def write_holding_register_sbint64(self, reference, value):
        """
        Write a signed 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt64 from holding register
    def read_holding_register_sbint64(self, reference):
        """
        Read a signed 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt64 to holding register
    def write_holding_register_sbint64_list(self, reference, values):
        """
        Write a list of signed 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt64 from holding register
    def read_holding_register_sbint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt64 to holding register
    def write_holding_register_slint64(self, reference, value):
        """
        Write a signed 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt64 from holding register
    def read_holding_register_slint64(self, reference):
        """
        Read a signed 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt64 to holding register
    def write_holding_register_slint64_list(self, reference, values):
        """
        Write a list of signed 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt64 from holding register
    def read_holding_register_slint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt21436587 to holding register
    def write_holding_register_uint21436587(self, reference, value):
        """
        Write a unsigned 21436587-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt21436587(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt21436587 from holding register
    def read_holding_register_uint21436587(self, reference):
        """
        Read a unsigned 21436587-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt21436587()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt21436587 to holding register
    def write_holding_register_uint21436587_list(self, reference, values):
        """
        Write a list of unsigned 21436587-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt21436587)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt21436587 from holding register
    def read_holding_register_uint21436587_list(self, reference, quantity):
        """
        Read a list of unsigned 21436587-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt21436587)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt78563412 to holding register
    def write_holding_register_uint78563412(self, reference, value):
        """
        Write a unsigned 78563412-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt78563412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt78563412 from holding register
    def read_holding_register_uint78563412(self, reference):
        """
        Read a unsigned 78563412-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt78563412()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt78563412 to holding register
    def write_holding_register_uint78563412_list(self, reference, values):
        """
        Write a list of unsigned 78563412-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt78563412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt78563412 from holding register
    def read_holding_register_uint78563412_list(self, reference, quantity):
        """
        Read a list of unsigned 78563412-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt78563412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt21436587 to holding register
    def write_holding_register_sint21436587(self, reference, value):
        """
        Write a signed 21436587-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt21436587(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt21436587 from holding register
    def read_holding_register_sint21436587(self, reference):
        """
        Read a signed 21436587-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt21436587()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt21436587 to holding register
    def write_holding_register_sint21436587_list(self, reference, values):
        """
        Write a list of signed 21436587-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt21436587)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt21436587 from holding register
    def read_holding_register_sint21436587_list(self, reference, quantity):
        """
        Read a list of signed 21436587-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt21436587)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt78563412 to holding register
    def write_holding_register_sint78563412(self, reference, value):
        """
        Write a signed 78563412-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt78563412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt78563412 from holding register
    def read_holding_register_sint78563412(self, reference):
        """
        Read a signed 78563412-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt78563412()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt78563412 to holding register
    def write_holding_register_sint78563412_list(self, reference, values):
        """
        Write a list of signed 78563412-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt78563412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt78563412 from holding register
    def read_holding_register_sint78563412_list(self, reference, quantity):
        """
        Read a list of signed 78563412-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt78563412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write BFloat to holding register
    def write_holding_register_bfloat(self, reference, value):
        """
        Write a float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.BFloat(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read BFloat from holding register
    def read_holding_register_bfloat(self, reference):
        """
        Read a float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.BFloat()
        return self.read_holding_register_data(reference, data).get()

    # write list of BFloat to holding register
    def write_holding_register_bfloat_list(self, reference, values):
        """
        Write a list of floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.BFloat)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of BFloat from holding register
    def read_holding_register_bfloat_list(self, reference, quantity):
        """
        Read a list of floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BFloat)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write LFloat to holding register
    def write_holding_register_lfloat(self, reference, value):
        """
        Write a float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.LFloat(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read LFloat from holding register
    def read_holding_register_lfloat(self, reference):
        """
        Read a float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.LFloat()
        return self.read_holding_register_data(reference, data).get()

    # write list of LFloat to holding register
    def write_holding_register_lfloat_list(self, reference, values):
        """
        Write a list of floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.LFloat)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of LFloat from holding register
    def read_holding_register_lfloat_list(self, reference, quantity):
        """
        Read a list of floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LFloat)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Float2143 to holding register
    def write_holding_register_float2143(self, reference, value):
        """
        Write a 2143-endian float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.Float2143(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Float2143 from holding register
    def read_holding_register_float2143(self, reference):
        """
        Read a 2143-endian float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float2143()
        return self.read_holding_register_data(reference, data).get()

    # write list of Float2143 to holding register
    def write_holding_register_float2143_list(self, reference, values):
        """
        Write a list of 2143-endian floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Float2143)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Float2143 from holding register
    def read_holding_register_float2143_list(self, reference, quantity):
        """
        Read a list of 2143-endian floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float2143)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Float3412 to holding register
    def write_holding_register_float3412(self, reference, value):
        """
        Write a 3412-endian float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.Float3412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Float3412 from holding register
    def read_holding_register_float3412(self, reference):
        """
        Read a 3412-endian float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float3412()
        return self.read_holding_register_data(reference, data).get()

    # write list of Float3412 to holding register
    def write_holding_register_float3412_list(self, reference, values):
        """
        Write a list of 3412-endian floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Float3412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Float3412 from holding register
    def read_holding_register_float3412_list(self, reference, quantity):
        """
        Read a list of 3412-endian floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float3412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write BDouble to holding register
    def write_holding_register_bdouble(self, reference, value):
        """
        Write a double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.BDouble(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read BDouble from holding register
    def read_holding_register_bdouble(self, reference):
        """
        Read a double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.BDouble()
        return self.read_holding_register_data(reference, data).get()

    # write list of BDouble to holding register
    def write_holding_register_bdouble_list(self, reference, values):
        """
        Write a list of doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.BDouble)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of BDouble from holding register
    def read_holding_register_bdouble_list(self, reference, quantity):
        """
        Read a list of doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BDouble)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write LDouble to holding register
    def write_holding_register_ldouble(self, reference, value):
        """
        Write a double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.LDouble(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read LDouble from holding register
    def read_holding_register_ldouble(self, reference):
        """
        Read a double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.LDouble()
        return self.read_holding_register_data(reference, data).get()

    # write list of LDouble to holding register
    def write_holding_register_ldouble_list(self, reference, values):
        """
        Write a list of doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.LDouble)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of LDouble from holding register
    def read_holding_register_ldouble_list(self, reference, quantity):
        """
        Read a list of doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LDouble)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Double21436587 to holding register
    def write_holding_register_double21436587(self, reference, value):
        """
        Write a 21436587-endian double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.Double21436587(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Double21436587 from holding register
    def read_holding_register_double21436587(self, reference):
        """
        Read a 21436587-endian double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double21436587()
        return self.read_holding_register_data(reference, data).get()

    # write list of Double21436587 to holding register
    def write_holding_register_double21436587_list(self, reference, values):
        """
        Write a list of 21436587-endian doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Double21436587)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Double21436587 from holding register
    def read_holding_register_double21436587_list(self, reference, quantity):
        """
        Read a list of 21436587-endian doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double21436587)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Double78563412 to holding register
    def write_holding_register_double78563412(self, reference, value):
        """
        Write a 78563412-endian double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.Double78563412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Double78563412 from holding register
    def read_holding_register_double78563412(self, reference):
        """
        Read a 78563412-endian double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double78563412()
        return self.read_holding_register_data(reference, data).get()

    # write list of Double78563412 to holding register
    def write_holding_register_double78563412_list(self, reference, values):
        """
        Write a list of 78563412-endian doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Double78563412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Double78563412 from holding register
    def read_holding_register_double78563412_list(self, reference, quantity):
        """
        Read a list of 78563412-endian doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double78563412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write String to holding register
    def write_holding_register_string(self, reference, value, length):
        """
        Write a string to holding register memory.

        Args:
            reference: the reference to write to
            value: the data value to be written
            length: the length of the value
        """
        data = oser.String(value=value, length=length)
        return self.write_holding_register_data(reference, data=data)

    # read String from holding register
    def read_holding_register_string(self, reference, length):
        """
        Read a string from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            data: value read from ``reference``
        """
        data = oser.String(length=length)
        return self.read_holding_register_data(reference, data).get()

    # write list of String to holding register
    def write_holding_register_string_list(self, reference, values, length):
        """
        Write a list of strings to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of data values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.String, kwargs={'length': length})
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of String from holding register
    def read_holding_register_string_list(self, reference, quantity, length):
        """
        Read a list of strings from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of data: data values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.String, kwargs={'length': length})
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Flag to holding register
    def write_holding_register_flag(self, reference, value):
        """
        Write a bit to holding register memory.

        Args:
            reference: the reference to write to
            value: the bool value to be written
        """
        data = oser.Flag(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data=data)

    # read Flag from holding register
    def read_holding_register_flag(self, reference):
        """
        Read a bit from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_holding_register_data(reference, data).data.get()

    # write list of Flag to holding register
    def write_holding_register_flag_list(self, reference, values):
        """
        Write a list of bits to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of bool values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Flag)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data)

    # read list of Flag from holding register
    def read_holding_register_flag_list(self, reference, quantity):
        """
        Read a list of bits from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Nibble to holding register
    def write_holding_register_nibble(self, reference, value):
        """
        Write a nibble to holding register memory.

        Args:
            reference: the reference to write to
            value: the nibble value to be written
        """
        data = oser.Nibble(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data=data)

    # read Nibble from holding register
    def read_holding_register_nibble(self, reference):
        """
        Read a nibble from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_holding_register_data(reference, data).data.get()

    # write list of Nibble to holding register
    def write_holding_register_nibble_list(self, reference, values):
        """
        Write a list of nibbles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of nibble values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Nibble)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data)

    # read list of Nibble from holding register
    def read_holding_register_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write BitField to holding register
    def write_holding_register_bitfield(self, reference, value, length):
        """
        Write a bitfield to holding register memory.

        Args:
            reference: the reference to write to
            value: the bitfield value to be written
            length: the length of the value
        """
        data = oser.BitField(value=value, length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data=data)

    # read BitField from holding register
    def read_holding_register_bitfield(self, reference, length):
        """
        Read a bitfield from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_holding_register_data(reference, data).data.get()

    # write list of BitField to holding register
    def write_holding_register_bitfield_list(self, reference, values, length):
        """
        Write a list of bitfields to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of bitfield values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.BitField, kwargs={'length': length})
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data)

    # read list of BitField from holding register
    def read_holding_register_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # read UBInt8 from input register
    def read_input_register_ubint8(self, reference):
        """
        Read a unsigned 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt8()
        return self.read_input_register_data(reference, data).get()

    # read list of UBInt8 from input register
    def read_input_register_ubint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read ULInt8 from input register
    def read_input_register_ulint8(self, reference):
        """
        Read a unsigned 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt8()
        return self.read_input_register_data(reference, data).get()

    # read list of ULInt8 from input register
    def read_input_register_ulint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SBInt8 from input register
    def read_input_register_sbint8(self, reference):
        """
        Read a signed 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt8()
        return self.read_input_register_data(reference, data).get()

    # read list of SBInt8 from input register
    def read_input_register_sbint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SLInt8 from input register
    def read_input_register_slint8(self, reference):
        """
        Read a signed 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt8()
        return self.read_input_register_data(reference, data).get()

    # read list of SLInt8 from input register
    def read_input_register_slint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UBInt16 from input register
    def read_input_register_ubint16(self, reference):
        """
        Read a unsigned 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt16()
        return self.read_input_register_data(reference, data).get()

    # read list of UBInt16 from input register
    def read_input_register_ubint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read ULInt16 from input register
    def read_input_register_ulint16(self, reference):
        """
        Read a unsigned 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt16()
        return self.read_input_register_data(reference, data).get()

    # read list of ULInt16 from input register
    def read_input_register_ulint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SBInt16 from input register
    def read_input_register_sbint16(self, reference):
        """
        Read a signed 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt16()
        return self.read_input_register_data(reference, data).get()

    # read list of SBInt16 from input register
    def read_input_register_sbint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SLInt16 from input register
    def read_input_register_slint16(self, reference):
        """
        Read a signed 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt16()
        return self.read_input_register_data(reference, data).get()

    # read list of SLInt16 from input register
    def read_input_register_slint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UBInt32 from input register
    def read_input_register_ubint32(self, reference):
        """
        Read a unsigned 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt32()
        return self.read_input_register_data(reference, data).get()

    # read list of UBInt32 from input register
    def read_input_register_ubint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read ULInt32 from input register
    def read_input_register_ulint32(self, reference):
        """
        Read a unsigned 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt32()
        return self.read_input_register_data(reference, data).get()

    # read list of ULInt32 from input register
    def read_input_register_ulint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SBInt32 from input register
    def read_input_register_sbint32(self, reference):
        """
        Read a signed 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt32()
        return self.read_input_register_data(reference, data).get()

    # read list of SBInt32 from input register
    def read_input_register_sbint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SLInt32 from input register
    def read_input_register_slint32(self, reference):
        """
        Read a signed 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt32()
        return self.read_input_register_data(reference, data).get()

    # read list of SLInt32 from input register
    def read_input_register_slint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UInt2143 from input register
    def read_input_register_uint2143(self, reference):
        """
        Read a unsigned 2143-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt2143()
        return self.read_input_register_data(reference, data).get()

    # read list of UInt2143 from input register
    def read_input_register_uint2143_list(self, reference, quantity):
        """
        Read a list of unsigned 2143-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt2143)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UInt3412 from input register
    def read_input_register_uint3412(self, reference):
        """
        Read a unsigned 3412-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt3412()
        return self.read_input_register_data(reference, data).get()

    # read list of UInt3412 from input register
    def read_input_register_uint3412_list(self, reference, quantity):
        """
        Read a list of unsigned 3412-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt3412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SInt2143 from input register
    def read_input_register_sint2143(self, reference):
        """
        Read a signed 2143-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt2143()
        return self.read_input_register_data(reference, data).get()

    # read list of SInt2143 from input register
    def read_input_register_sint2143_list(self, reference, quantity):
        """
        Read a list of signed 2143-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt2143)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SInt3412 from input register
    def read_input_register_sint3412(self, reference):
        """
        Read a signed 3412-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt3412()
        return self.read_input_register_data(reference, data).get()

    # read list of SInt3412 from input register
    def read_input_register_sint3412_list(self, reference, quantity):
        """
        Read a list of signed 3412-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt3412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UBInt64 from input register
    def read_input_register_ubint64(self, reference):
        """
        Read a unsigned 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt64()
        return self.read_input_register_data(reference, data).get()

    # read list of UBInt64 from input register
    def read_input_register_ubint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read ULInt64 from input register
    def read_input_register_ulint64(self, reference):
        """
        Read a unsigned 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt64()
        return self.read_input_register_data(reference, data).get()

    # read list of ULInt64 from input register
    def read_input_register_ulint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SBInt64 from input register
    def read_input_register_sbint64(self, reference):
        """
        Read a signed 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt64()
        return self.read_input_register_data(reference, data).get()

    # read list of SBInt64 from input register
    def read_input_register_sbint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SLInt64 from input register
    def read_input_register_slint64(self, reference):
        """
        Read a signed 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt64()
        return self.read_input_register_data(reference, data).get()

    # read list of SLInt64 from input register
    def read_input_register_slint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UInt21436587 from input register
    def read_input_register_uint21436587(self, reference):
        """
        Read a unsigned 21436587-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt21436587()
        return self.read_input_register_data(reference, data).get()

    # read list of UInt21436587 from input register
    def read_input_register_uint21436587_list(self, reference, quantity):
        """
        Read a list of unsigned 21436587-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt21436587)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read UInt78563412 from input register
    def read_input_register_uint78563412(self, reference):
        """
        Read a unsigned 78563412-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt78563412()
        return self.read_input_register_data(reference, data).get()

    # read list of UInt78563412 from input register
    def read_input_register_uint78563412_list(self, reference, quantity):
        """
        Read a list of unsigned 78563412-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt78563412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SInt21436587 from input register
    def read_input_register_sint21436587(self, reference):
        """
        Read a signed 21436587-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt21436587()
        return self.read_input_register_data(reference, data).get()

    # read list of SInt21436587 from input register
    def read_input_register_sint21436587_list(self, reference, quantity):
        """
        Read a list of signed 21436587-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt21436587)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read SInt78563412 from input register
    def read_input_register_sint78563412(self, reference):
        """
        Read a signed 78563412-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt78563412()
        return self.read_input_register_data(reference, data).get()

    # read list of SInt78563412 from input register
    def read_input_register_sint78563412_list(self, reference, quantity):
        """
        Read a list of signed 78563412-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt78563412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read BFloat from input register
    def read_input_register_bfloat(self, reference):
        """
        Read a float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.BFloat()
        return self.read_input_register_data(reference, data).get()

    # read list of BFloat from input register
    def read_input_register_bfloat_list(self, reference, quantity):
        """
        Read a list of floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BFloat)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read LFloat from input register
    def read_input_register_lfloat(self, reference):
        """
        Read a float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.LFloat()
        return self.read_input_register_data(reference, data).get()

    # read list of LFloat from input register
    def read_input_register_lfloat_list(self, reference, quantity):
        """
        Read a list of floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LFloat)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read Float2143 from input register
    def read_input_register_float2143(self, reference):
        """
        Read a 2143-endian float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float2143()
        return self.read_input_register_data(reference, data).get()

    # read list of Float2143 from input register
    def read_input_register_float2143_list(self, reference, quantity):
        """
        Read a list of 2143-endian floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float2143)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read Float3412 from input register
    def read_input_register_float3412(self, reference):
        """
        Read a 3412-endian float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float3412()
        return self.read_input_register_data(reference, data).get()

    # read list of Float3412 from input register
    def read_input_register_float3412_list(self, reference, quantity):
        """
        Read a list of 3412-endian floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float3412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read BDouble from input register
    def read_input_register_bdouble(self, reference):
        """
        Read a double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.BDouble()
        return self.read_input_register_data(reference, data).get()

    # read list of BDouble from input register
    def read_input_register_bdouble_list(self, reference, quantity):
        """
        Read a list of doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BDouble)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read LDouble from input register
    def read_input_register_ldouble(self, reference):
        """
        Read a double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.LDouble()
        return self.read_input_register_data(reference, data).get()

    # read list of LDouble from input register
    def read_input_register_ldouble_list(self, reference, quantity):
        """
        Read a list of doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LDouble)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read Double21436587 from input register
    def read_input_register_double21436587(self, reference):
        """
        Read a 21436587-endian double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double21436587()
        return self.read_input_register_data(reference, data).get()

    # read list of Double21436587 from input register
    def read_input_register_double21436587_list(self, reference, quantity):
        """
        Read a list of 21436587-endian doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double21436587)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read Double78563412 from input register
    def read_input_register_double78563412(self, reference):
        """
        Read a 78563412-endian double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double78563412()
        return self.read_input_register_data(reference, data).get()

    # read list of Double78563412 from input register
    def read_input_register_double78563412_list(self, reference, quantity):
        """
        Read a list of 78563412-endian doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double78563412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read String from input register
    def read_input_register_string(self, reference, length):
        """
        Read a string from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            data: value read from ``reference``
        """
        data = oser.String(length=length)
        return self.read_input_register_data(reference, data).get()

    # read list of String from input register
    def read_input_register_string_list(self, reference, quantity, length):
        """
        Read a list of strings from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of data: data values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.String, kwargs={'length': length})
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # read Flag from input register
    def read_input_register_flag(self, reference):
        """
        Read a bit from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_input_register_data(reference, data).data.get()

    # read list of Flag from input register
    def read_input_register_flag_list(self, reference, quantity):
        """
        Read a list of bits from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # read Nibble from input register
    def read_input_register_nibble(self, reference):
        """
        Read a nibble from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_input_register_data(reference, data).data.get()

    # read list of Nibble from input register
    def read_input_register_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # read BitField from input register
    def read_input_register_bitfield(self, reference, length):
        """
        Read a bitfield from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_input_register_data(reference, data).data.get()

    # read list of BitField from input register
    def read_input_register_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))


class ModbusSerialRTUClient(ModbusBaseClient):

    def __init__(self, com_port, baud=9600, bytesize=serial.EIGHTBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False,
                 device_address=1,
                 timeout=1.0,
                 retries=None,
                 references_start_at_one=True,
                 request_parser=SerialRTUModbusRequest,
                 response_parser=SerialRTUModbusResponse,
                 debuglevel=None):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            device_address (int): the device address (Slave ID)
            timeout (float): the timeout in seconds
            retries=None (int or None): the number of retries until an exception is raised
            references_start_at_one=True: if set to ``True`` all references start at ``1`` else ``0``
            request_parser: a parser for requests
            response_parser: a parser for responses
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        self.device_address = device_address
        interface = ModbusSerialRTUClientInterface(com_port, baud=baud, bytesize=bytesize, parity=parity,
                                                   stopbits=stopbits, rtscts=rtscts, timeout=timeout)
        super(ModbusSerialRTUClient, self).__init__(
                                                interface=interface,
                                                request_parser=request_parser,
                                                response_parser=response_parser,
                                                retries=retries,
                                                references_start_at_one=references_start_at_one,
                                                debuglevel=debuglevel)

    def set_device_address(self, device_address):
        """
        Set device address.

        Args:
            device_address (int): the device address to be used for the next requests.
        """
        self.device_address = device_address

    def build_request(self):
        """
        Build and return an instance of the desired request parser.

        Returns:
            request parser
        """
        request = super(ModbusSerialRTUClient, self).build_request()
        request.address.set(self.device_address)
        return request

    # Commands
    # ========

    def get_exception_status(self):
        """
        Read and return exception status using Modbus function code 7.

        Returns:
            int: the current exception status
        """
        adu = self.build_request()
        adu.pdu.function_code.set(7)
        response = self.query(adu)
        return response.pdu.data.output_data.get()

    def echo(self, data):
        """
        Query data and return echo using Modbus function code 8.0.
        An exception is raised if the received data differ from the sent data or the request is too long.

        Args:
            data (bytes): The bytes to be sent.

        Returns:
            response
        """
        assert len(data) <= 65533, "the data is too long to be transfered"
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(0)
        adu.pdu.data.data.set(data)
        response = self.query(adu)
        assert response.pdu.data.data.get() == data, "the echoed data differ (original: {}, echoed: {})".format(
            repr(data), repr(response.pdu.data.data.get()))
        return response

    def get_diagnostic_register(self):
        """
        Query and return the diagnostic register using Modbus function 8.2.

        Returns:
            bytes: the content of the diagnostic register
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(2)
        response = self.query(adu)
        return response.pdu.data.data.get()

    def set_listen_only_mode(self):
        """
        Set listen only mode in server using Modbus function code 8.4.
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(4)  # Set listen only mode
        try:
            self.query(adu)
        except:
            return

        raise Exception("An answer was received!")  # pragma: no cover

    def restart_communications(self, reset_eventlog=False):
        """
        Restart communications using Modbus function code 8.1.
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(1)  # Reset listen only mode and clear counters
        if reset_eventlog:
            adu.pdu.data.data.set(b"\xff\x00")
        self.query(adu)

    def clear_counters_and_diagnostic_register(self):
        """
        Clear counters and diagnostic register using Modbus function code 8.10.
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(10)
        response = self.query(adu)

    def get_bus_message_count(self):
        """
        Query and return the bus message count from the device using Modbus function code 8.11.

        Returns:
            int: the bus message count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(11)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_bus_communication_error_count(self):
        """
        Query and return the bus communication error count from the device using Modbus function code 8.12.

        Returns:
            int: the bus communication error count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(12)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_bus_exception_error_count(self):
        """
        Query and return the bus exception error count from the device using Modbus function code 8.13.

        Returns:
            int: the bus exception error count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(13)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_slave_message_count(self):
        """
        Query and return the slave message count from the device using Modbus function code 8.14.

        Returns:
            int: the slave message count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(14)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_slave_no_response_count(self):
        """
        Query and return the slave no response count from the device using Modbus function code 8.15.

        Returns:
            int: the slave no response count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(15)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_slave_nak_count(self):
        """
        Query and return the slave nak count from the device using Modbus function code 8.16.

        Returns:
            int: the slave nak count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(16)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_busy_count(self):
        """
        Query and return the slave busy count from the device using Modbus function code 8.17.

        Returns:
            int: the slave busy count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(17)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def get_bus_character_overrun_count(self):
        """
        Query and return the bus character overrun count from the device using Modbus function code 8.18.

        Returns:
            int: the bus character overrun count
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(18)
        response = self.query(adu)
        decoder = oser.UBInt16()
        decoder.decode(response.pdu.data.data.get())
        return decoder.get()

    def clear_overrun_counter_and_flag(self):
        """
        Clear overrun counter and reset the error flag using Modbus function code 8.20.
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(20)
        self.query(adu)

    def get_communication_event_counter_and_busy_status(self):
        """
        Query and return communication event counter and busy status using Modbus function code 11.

        Returns:
            tuple: communication_event_counter (int), busy_status (bool)
        """
        adu = self.build_request()
        adu.pdu.function_code.set(11)
        response = self.query(adu)
        busy_status = response.pdu.data.status.get()
        assert busy_status in [0x0000, 0xffff]
        event_count = response.pdu.data.event_count.get()
        return event_count, bool(busy_status)

    def get_event_count(self):
        """
        Query and return event count status using Modbus function code 11.

        Returns:
            int: communication_event_counter
        """
        return self.get_communication_event_counter_and_busy_status()[0]

    def get_busy_status(self):
        """
        Query and return busy status using Modbus function code 11.

        Returns:
            bool: busy_status
        """
        return self.get_communication_event_counter_and_busy_status()[1]

    def get_event_log(self):
        """
        Query and return the event log using Modbus function code 12.

        Returns:
            tuple: busy_status (int), event_count (int), message_count (int), events (list of int)
        """
        adu = self.build_request()
        adu.pdu.function_code.set(12)
        response = self.query(adu)
        busy_status = response.pdu.data.status.get()
        assert busy_status in [0x0000, 0xffff]
        event_count = response.pdu.data.event_count.get()
        message_count = response.pdu.data.message_count.get()
        events = [event.get() for event in response.pdu.data.events]

        return bool(busy_status), event_count, message_count, events

    def get_slave_id(self):
        """
        Query and return the slave id using Modbus function code 17.

        Returns:
            3-tuple: slave-id, running, data
        """
        adu = self.build_request()
        adu.pdu.function_code.set(17)
        response = self.query(adu)

        slave_id = response.pdu.data.id.get()
        slave_is_running = response.pdu.data.run_indicator_status.get() == 0xFF
        data = response.pdu.data.data.get()
        return slave_id, slave_is_running, data


class ModbusSerialASCIIClient(ModbusSerialRTUClient):

    def __init__(self, com_port, baud=9600, bytesize=serial.EIGHTBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False,
                 device_address=1,
                 timeout=1.0,
                 retries=None,
                 references_start_at_one=True,
                 ascii_delimiter=b'\x0A',
                 request_parser=SerialASCIIModbusRequest,
                 response_parser=SerialASCIIModbusResponse,
                 debuglevel=None):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            device_address (int): the device address (Slave ID)
            timeout (float): the timeout in seconds
            retries=None (int or None): the number of retries until an exception is raised
            references_start_at_one=True: if set to ``True`` all references start at ``1`` else ``0``
            ascii_delimiter (bytes): the ascii delimiter to be used
            request_parser: a parser for requests
            response_parser: a parser for responses
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        self.device_address = device_address
        self._request_parser_type = request_parser
        interface = ModbusSerialASCIIClientInterface(com_port, baud=baud, bytesize=bytesize, parity=parity,
                                                     stopbits=stopbits, rtscts=rtscts, timeout=timeout)
        ModbusBaseClient.__init__(self,
                                  interface=interface,
                                  request_parser=request_parser,
                                  response_parser=response_parser,
                                  retries=retries,
                                  references_start_at_one=references_start_at_one,
                                  debuglevel=debuglevel)

        self.interface.change_ascii_delimiter(ascii_delimiter)

    def change_ascii_delimiter(self, delimiter):
        """
        Change ASCII delimiter in the device.

        Args:
            delimiter (bytes): the new ASCII delimiter.
        """
        adu = self.build_request()
        adu.pdu.function_code.set(8)
        adu.pdu.data.sub_function.set(3)
        if not isinstance(delimiter, bytes) or not len(delimiter) == 1:  # pragma: no cover
            raise ValueError("ascii_delimiter must be bytes of length one")
        adu.pdu.data.data.set(delimiter + b"\x00")
        self.write(adu)
        # change ascii delimiter locally
        self.interface.change_ascii_delimiter(delimiter)
        resp = self.read()
        if resp.pdu.data.exception_code.get() is not None:  # pragma: no cover
            raise ModbusException('{} (Code: {})'.format(resp.pdu.data.exception_code.get(),
                                                         resp.pdu.data.exception_code.get_value()))


class ModbusTCPClient(ModbusBaseClient):
    _interface_type = ModbusTCPClientInterface

    def __init__(self, address,
                 port=502,
                 unit_identifier=0,
                 timeout=1.0,
                 retries=None,
                 references_start_at_one=True,
                 transaction_identifier=0,
                 auto_disconnect=False,
                 request_parser=TCPModbusRequest,
                 response_parser=TCPModbusResponse,
                 debuglevel=None):
        """
        Args:
            address (str): the ip address to connect to
            port (int): the port to be used
            unit_identifier (int): the unit identifier for Modbus TCP
            timeout (float): the timeout in seconds
            retries=None (int or None): the number of retries until an exception is raised
            references_start_at_one=True: if set to ``True`` all references start at ``1`` else ``0``
            transaction_identifier (int): the initial transaction identifier
            auto_disconnect=False (bool): if set to ``True`` the TCP connection is closed immediately and re-opened on
                next request
            request_parser: a parser for requests
            response_parser: a parser for responses
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """

        interface = self._interface_type(address, port=port, timeout=timeout, connect=not auto_disconnect)
        self.unit_identifier = unit_identifier
        self.transaction_identifier = transaction_identifier
        self.lock = threading.Lock()
        self.auto_disconnect = auto_disconnect
        super(ModbusTCPClient, self).__init__(
                                            interface=interface,
                                            request_parser=request_parser,
                                            response_parser=response_parser,
                                            retries=retries,
                                            references_start_at_one=references_start_at_one,
                                            debuglevel=debuglevel)

    def reconnect(self):
        """
        Reconnect socket.
        """
        self.interface.reconnect()

    def get_transaction_identifier(self):
        """
        Create and return a new transaction identifer.
        Is thread-safe.

        Returns:
            int: the next transaction identifier
        """
        with self.lock:
            self.transaction_identifier += 1
            return self.transaction_identifier

    def write(self, adu, auto_length=True):
        """
        Write ``adu`` encoded using the interface.
        The transaction identifier is set automatically.
        The unit identifier is put into the adu automatically, too.

        Args:
            adu: a serializable adu
            auto_length=True (bool): if set to ``True`` the length is calculated automatically
        """
        if auto_length:  # pragma: no cover
            adu.length.set(len(adu.pdu.encode()) + 1)
        adu.transaction_identifier.set(self.get_transaction_identifier())
        adu.unit_identifier.set(self.unit_identifier)
        self.interface.write(adu)

    def query(self, adu):
        """
        Query ``adu`` using the interface and return the decoded response.

        Args:
            adu: a serializable adu

        Returns:
            adu: the decoded response if no exception occurred

        Raises:
            ModbusException: in case of an exception response
        """
        if self.auto_disconnect:
            self.interface.open()
        ret = super(ModbusTCPClient, self).query(adu)
        if self.auto_disconnect:
            self.interface.close()
        return ret


class ModbusTLSClient(ModbusTCPClient):
    _interface_type = ModbusTLSClientInterface

    def __init__(self, address,
                 port=502,
                 unit_identifier=0,
                 timeout=1.0,
                 retries=None,
                 references_start_at_one=True,
                 transaction_identifier=0,
                 auto_disconnect=False,
                 verify=True,
                 cafile=None,
                 capath=None,
                 cadata=None,
                 request_parser=TCPModbusRequest,
                 response_parser=TCPModbusResponse,
                 debuglevel=None):
        """
        Args:
            address (str): the ip address to connect to
            port (int): the port to be used
            unit_identifier (int): the unit identifier for Modbus TCP
            timeout (float): the timeout in seconds
            retries=None (int or None): the number of retries until an exception is raised
            references_start_at_one=True: if set to ``True`` all references start at ``1`` else ``0``
            transaction_identifier (int): the initial transaction identifier
            auto_disconnect=False (bool): if set to ``True`` the TCP connection is closed immediately and re-opened on
                next request
            cafile (str): path to a file of concatenated CA certificates in PEM format
            capath (str): path to a directory containing several CA certificates in PEM format
            cadata (str): either an ASCII string of one or more PEM-encoded certificates or a bytes-like
                object of DER-encoded certificates
            request_parser: a parser for requests
            response_parser: a parser for responses
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        interface = self._interface_type(address, port=port, timeout=timeout, connect=not auto_disconnect,
                                         verify=verify, cafile=cafile, capath=capath, cadata=cadata)
        self.unit_identifier = unit_identifier
        self.transaction_identifier = transaction_identifier
        self.lock = threading.Lock()
        self.auto_disconnect = auto_disconnect
        ModbusBaseClient.__init__(self,
                                  interface=interface,
                                  request_parser=request_parser,
                                  response_parser=response_parser,
                                  retries=retries,
                                  references_start_at_one=references_start_at_one,
                                  debuglevel=debuglevel)


class ModbusUDPClient(ModbusTCPClient):
    _interface_type = ModbusUDPClientInterface

    def __init__(self, address,
                 port=502,
                 unit_identifier=0,
                 timeout=1.0,
                 retries=None,
                 references_start_at_one=True,
                 transaction_identifier=0,
                 request_parser=TCPModbusRequest,
                 response_parser=TCPModbusResponse,
                 debuglevel=None):
        """
        Args:
            address (str): the ip address to connect to
            port (int): the port to be used
            unit_identifier (int): the unit identifier for Modbus TCP
            timeout (float): the timeout in seconds
            retries=None (int or None): the number of retries until an exception is raised
            references_start_at_one=True: if set to ``True`` all references start at ``1`` else ``0``
            transaction_identifier (int): the initial transaction identifier
            request_parser: a parser for requests
            response_parser: a parser for responses
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        interface = self._interface_type(address, port=port, timeout=timeout)
        self.unit_identifier = unit_identifier
        self.transaction_identifier = transaction_identifier
        self.lock = threading.Lock()
        ModbusBaseClient.__init__(self,
                                  interface=interface,
                                  request_parser=request_parser,
                                  response_parser=response_parser,
                                  retries=retries,
                                  references_start_at_one=references_start_at_one,
                                  debuglevel=debuglevel)

    def query(self, adu):
        """
        Query ``adu`` using the interface and return the decoded response.

        Args:
            adu: a serializable adu

        Returns:
            adu: the decoded response if no exception occurred

        Raises:
            ModbusException: in case of an exception response
        """
        return ModbusBaseClient.query(self, adu)
