# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import sys

import math

import oser
import traceback

from htf.communication.modbus.server import ModbusTLSServer, ModbusSerialRTUServer, ModbusTCPServer, ModbusUDPServer, \
    ModbusSerialASCIIServer
from htf.communication.modbus.protocol import SerialRTUModbusRequest, SerialRTUModbusResponse, TCPModbusResponse, \
    TCPModbusRequest, SerialASCIIModbusRequest, SerialASCIIModbusResponse
from htf.communication.modbus.memory import ModbusMemory
from htf.communication.modbus.core import ModbusBaseDevice as _ModbusBaseDevice
from htf.__version__ import __version__ as htf_version

if not isinstance(htf_version, type(b"")):  # pragma: no cover
    htf_version = htf_version.encode()


REGISTER_BYTE_SIZE = 2


class ModbusBaseDevice(_ModbusBaseDevice):

    def __init__(self, server, memory, debuglevel=None):
        """
        Args:
            server: a server instance to be used
            memory: a memory instance to be used
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        self.server = server
        self.memory = memory
        self.debuglevel = debuglevel
        self.busy_status = False
        self.exception_status = oser.UBInt8()
        self.fifo_queues = {}
        self.negative_acknowledge_count = 0
        self.events = []
        self.event_counter = 0
        self.device_identifier = {
            0x00: b"HILSTER",  # "vendor_name"
            0x01: b"htf",  # "product_code"
            0x02: htf_version,  # "major_minor_version"
            0x03: b"https://hilster.io",  # "vendor_url"
            0x04: b"HILSTER Testing Framework",  # "product_name"
            0x05: b"Modbus Testing Framework",  # "model_name"
            0x06: b"None",  # "user_application_name"
        }
        self.extended_memory = {}

        try:
            self.server.set_receive_event_callback(self.add_receive_event)
        except AttributeError:
            pass

        self.server.start()

    def close(self):
        """
        Close the device simulator and the server.
        """
        self.server.close()

    # Coils
    # =====

    def read_coil(self, reference):
        """
        Read a single coil from device memory.

        Args:
            reference (int): the register reference.

        Returns:
            int: ``1`` if the coil is set and ``0`` else.
        """
        return self.read_multiple_coils(reference, 1)[0]

    def write_coil(self, reference, value):
        """
        Write a single coil to device memory.

        Args:
            reference (int): the register reference.
            value (int): the value to be written.
        """
        self.write_multiple_coils(reference, [value])

    def read_multiple_coils(self, reference, quantity):
        """
        Read multiple coils from device memory.

        Args:
            reference (int): the register reference.
            quantity (int): the number of coils to be read.

        Returns:
            list of int: each element is ``1`` if the corresponding coil is set and ``0`` else.
        """
        coils = []
        for i in range(quantity):
            coils.append(self.memory.read_coil(reference+i))
        return coils

    def write_multiple_coils(self, reference, values):
        """
        Write multiple coils to device memory.

        Args:
            reference (int): the register reference.
            values (list of int): the values to be written.
        """
        quantity = len(values)
        for i in range(quantity):
            self.memory.write_coil(reference+i, values[i])

    def write_coil_data(self, reference, data):
        """
        Write data to coils device memory.

        Args:
            reference (int): the coil reference.
            data (oser.ByteStruct): the type or instance or the decoder for the data
        """
        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be an instance of oser.ByteStruct or oser.ByteType")

        encoded = data.encode()
        quantity = len(encoded) * 8

        self.memory.write_multiple_coils(reference, quantity,
                                         bytearray(encoded))

    def read_coil_data(self, reference, data):
        """
        Read data from coils device memory.

        Args:
            reference (int): the coil reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = max_byte_size * 8

        coil_status = []

        while quantity > 0:
            coil_status.append(self.memory.read_coils_as_int(reference, min(quantity, 8)))
            reference += 8
            quantity -= 8

        encoded = bytes(bytearray(coil_status))

        data.decode(encoded)

        return data

    # Discrete Inputs
    # ===============

    def read_discrete_input(self, reference):
        """
        Read single discrete input from device memory

        Args:
            reference (int): the register reference.

        Returns:
            int: ``1`` if the discete input is set and ``0`` else.
        """
        return self.read_multiple_discrete_inputs(reference, 1)[0]

    def write_discrete_input(self, reference, value):
        """
        Read multiple discrete inputs from device memory

        Args:
            reference (int): the register reference.
            quantity (int): the number of discrete inputs to be read.

        Returns:
            list of int: each element is ``1`` if the corresponding discrete input is set and ``0`` else.
        """
        self.write_multiple_discrete_inputs(reference, [value])

    def read_multiple_discrete_inputs(self, reference, quantity):
        """
        Read multiple discrete inputs from device memory.

        Args:
            reference (int): the register reference.
            quantity (int): the number of discrete inputs to be read.

        Returns:
            list of int: each element is ``1`` if the corresponding discrete input is set and ``0`` else.
        """
        inputs = []
        for i in range(quantity):
            inputs.append(self.memory.read_discrete_input(reference + i))
        return inputs

    def write_multiple_discrete_inputs(self, reference, values):
        """
        Write multiple discrete inputs to device memory.

        Args:
            reference (int): the register reference.
            values (list of int): the values to be written.
        """
        quantity = len(values)
        for i in range(quantity):
            self.memory.write_discrete_input(reference + i, values[i])

    def read_discrete_input_data(self, reference, data):
        """
        Read data from discrete inputs device memory.

        Args:
            reference (int): the discrete input reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = max_byte_size * 8

        coil_status = []

        while quantity > 0:
            coil_status.append(self.memory.read_discrete_input_as_int(reference, min(quantity, 8)))
            reference += 8
            quantity -= 8

        encoded = bytes(bytearray(coil_status))

        data.decode(encoded)

        return data

    def write_discrete_input_data(self, reference, data):
        """
        Write data to discrete input device memory.

        Args:
            reference (int): the discrete input reference.
            data (oser.ByteStruct): the type or instance or the decoder for the data
        """
        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be an instance of oser.ByteStruct or oser.ByteType")

        encoded = data.encode()
        quantity = len(encoded) * 8

        self.memory.write_multiple_discrete_inputs(reference, quantity, bytearray(encoded))

    # Holding Registers
    # =================

    def read_holding_register(self, reference):
        """
        Read a single holding register from device memory.

        Args:
            reference (int): the register reference.

        Returns:
            int: the holding register value
        """
        register = self.memory.read_holding_register(reference)
        decoder = oser.UBInt16()
        decoder.decode(register)
        return decoder.get()

    def write_holding_register(self, reference, value):
        """
        Write single holding register to device memory.

        Args:
            reference: the register reference.
            value: the value to be written into the corresponding register.
        """
        encoder = oser.UBInt16(value)
        self.memory.write_holding_register(reference, encoder.encode())

    def read_multiple_holding_registers(self, reference, quantity):
        """
        Read multiple holding registers from device memory.

        Args:
            reference (int): the register reference.
            quantity (int): the number of registers to be read.

        Returns:
            list of int: the holding register values
        """
        registers = [self.memory.read_holding_register(reference + i) for i in range(quantity)]
        decoder = oser.UBInt16()
        integers = [decoder.get() for reg in registers if decoder.decode(reg)]
        return integers

    def write_multiple_holding_registers(self, reference, values):
        """
        Write multiple holding registers to device memory.

        Args:
            reference (int): the register reference.
            values (list of int): the values to be written.
        """
        encoder = oser.UBInt16()
        register_values = []
        for value in values:
            encoder.set(value)
            register_values.append(encoder.encode())

        self.memory.write_multiple_holding_registers(reference, register_values)

    def mask_holding_register(self, reference, and_mask, or_mask):
        """
        Mask write single holding register from device memory.

        Args:
            reference (int): the register reference
            and_mask (int): the and mask
            or_mask (int): the or mask
        """
        current_content = self.memory.read_holding_register(reference)
        var = oser.UBInt16()
        var.decode(current_content)
        current_content_int = var.get()
        masked_value = (current_content_int & and_mask) | (or_mask & (~and_mask))
        self.memory.write_holding_register(reference, oser.UBInt16(masked_value).encode())

    def mask_multiple_holding_registers(self, reference, quantity, and_mask, or_mask):
        """
        Mask write multiple holding registers from device memory.

        Args:
            reference (int): the register reference
            quantity (int): the number of registers to be masked
            and_mask (int): the and mask
            or_mask (int): the or mask
        """
        for i in range(quantity):
            self.mask_holding_register(reference+i, and_mask=and_mask, or_mask=or_mask)

    def write_holding_register_data(self, reference, data):
        """
        Write data to holding registers device memory.

        Args:
            reference (int): the register reference.
            data (oser.ByteStruct): the type or instance or the decoder for the data
        """
        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be an instance of oser.ByteStruct or oser.ByteType")

        encoded = data.encode()
        slices = [encoded[i * REGISTER_BYTE_SIZE:(i + 1) * REGISTER_BYTE_SIZE]
                  for i in range((len(encoded) + (REGISTER_BYTE_SIZE - 1)) // REGISTER_BYTE_SIZE)]
        if len(encoded) % 2 != 0:
            slices[-1] += b'\x00'
        self.memory.write_multiple_holding_registers(reference, slices)

    def read_holding_register_data(self, reference, data):
        """
        Read data from holding registers device memory.

        Args:
            reference (int): the register reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = (max_byte_size + 1) // 2

        registers = b"".join([bytes(self.memory.read_holding_register(reference + i)) for i in range(quantity)])

        data.decode(registers)

        return data

    # Input Registers
    # ===============

    def read_input_register(self, reference):
        """
        Read a single input register from device memory.

        Args:
            reference (int): the register reference.

        Returns:
            int: the input register value
        """
        register = self.memory.read_input_register(reference)
        decoder = oser.UBInt16()
        decoder.decode(register)
        return decoder.get()

    def write_input_register(self, reference, value):
        """
        Write single input register to device memory.

        Args:
            reference: the register reference.
            value: the value to be written into the corresponding register.
        """
        self.write_multiple_input_registers(reference, [value])

    def read_multiple_input_registers(self, reference, quantity):
        """
        Read multiple input registers from device memory.

        Args:
            reference (int): the register reference.
            quantity (int): the number of registers to be read.

        Returns:
            list of int: the inputregister values
        """
        registers = []
        for i in range(quantity):
            registers.append(self.read_input_register(reference+i))
        return registers

    def write_multiple_input_registers(self, reference, values):
        """
        Write multiple input registers to device memory.

        Args:
            reference (int): the register reference.
            values (list of int): the values to be written.
        """
        u = oser.UBInt16()
        register_values = []
        for reg in values:
            u.set(reg)
            register_values.append(u.encode())
        self.memory.write_multiple_input_registers(reference, register_values)

    def read_input_register_data(self, reference, data):
        """
        Read data from input registers from device memory.

        Args:
            reference (int): the register reference
            data (oser.ByteStruct): the type or instance or the decoder for the data

        Returns:
            instance of ``datatype`` containing the read data
        """
        try:
            issubclass(data, (oser.ByteStruct, oser.ByteType))
            data = data()
        except TypeError:
            pass

        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be a subclass or an instance of oser.ByteStruct or oser.ByteType")

        max_byte_size = data._get_max_byte_size()
        quantity = (max_byte_size + 1) // 2

        registers = b"".join([bytes(self.memory.read_input_register(reference + i)) for i in range(quantity)])

        data.decode(registers)

        return data

    def write_input_register_data(self, reference, data):
        """
        Write data to input registers device memory.

        Args:
            reference (int): the register reference.
            data (oser.ByteStruct): the type or instance or the decoder for the data
        """
        if not isinstance(data, (oser.ByteStruct, oser.ByteType)):
            raise TypeError("data must be an instance of oser.ByteStruct or oser.ByteType")

        encoded = data.encode()
        slices = [encoded[i * REGISTER_BYTE_SIZE:(i + 1) * REGISTER_BYTE_SIZE]
                  for i in range((len(encoded) + (REGISTER_BYTE_SIZE - 1)) // REGISTER_BYTE_SIZE)]
        if len(encoded) % 2 != 0:
            slices[-1] += b'\x00'
        self.memory.write_multiple_input_registers(reference, slices)

    # Special Methods
    # ===============

    def set_fifo_queue_content(self, address, content):
        """
        Set the FIFO queue content.

        Args:
            address: the address
            content: a list of integers or ``None`` to delete the queue contents
        """
        assert 0x0000 <= address <= 0xffff, "address must be within 0x0000 and 0xffff"

        if content is None:
            del self.fifo_queues[address]
        else:
            assert len(content) <= 31, "content must not have more than 31 elements"
            self.fifo_queues[address] = content

    def set_device_identifier(self, object_id, value):
        """
        Set a device identifier.
        The objects ``"vendor_name"``, ``"product_code"``, ``"major_minor_version"``, ``"vendor_url"``,
        ``"product_name"``, ``"model_name"`` and ``"user_application_name"`` are predefined.

        Set ``value`` to ``None`` to remove an entry.

        Entries can be read using Modbus function code 43.14.

        Args:
            object_id (int or None): the object identifier
            value (str or bytes): the value to be used
        """

        if value is None:
            del self.device_identifier[object_id]
        else:
            assert isinstance(value, (str, bytes)), "value must be str or bytes"
            if not isinstance(value, type(b"")):
                value = value.encode()
            self.device_identifier[object_id] = value

    def set_extended_memory(self, memory_number, records):
        """
        Set extended memory that can be read with Modbus function code 20 and written with Modbus function code 21.

        If ``records`` is ``None`` the memore is removed.

        Args:
            memory_number (int): the memory number (0x0001 to 0xffff)
            records (list of int or None): the records to be stored
        """
        assert isinstance(memory_number, int), "memory_number must be an integer"
        assert 0x0001 <= memory_number <= 0xffff, "memory_number must be an integer between 0x0001 and 0xffff"

        if records is None:
            del self.extended_memory[memory_number]
        else:
            assert all(map(lambda x: isinstance(x, int), records)), "records must be a list of integers"
            self.extended_memory[memory_number] = records

    def get_extended_memory(self, memory_number):
        """
        Get extended memory.

        Args:
            memory_number (int): the memory number (0x0001 to 0xffff)

        Returns:
             list of int: the extended memory at ``memory_number``

        Raises:
            KeyError: if the ``memory_number`` was not found in extended memory.
        """
        assert isinstance(memory_number, int), "memory_number must be an integer"
        assert 0x0001 <= memory_number <= 0xffff, "memory_number must be an integer between 0x0001 and 0xffff"

        return self.extended_memory[memory_number]

    # Request Handlers
    # ================

    def handle_request(self, adu):
        """
        Request handler for an incoming request.

        Args:
            adu: a request adu

        Returns:
            handler: the command or exception handler
        """
        if self.debuglevel:  # pragma: no cover
            # debug
            print("=" * 80, file=sys.stderr)
            print("Device request:", file=sys.stderr)
            if self.debuglevel == 1:
                print(adu, file=sys.stderr)
            else:
                print(adu.introspect(), file=sys.stderr)

        try:
            function_code = adu.pdu.function_code.get()
        except Exception:  # pragma: no cover
            print(traceback.print_exc())
            return self.send_exception_response(adu.pdu, "server_device_failure")

        try:
            handler = getattr(self, 'handle_function_code_{}'.format(function_code))
        except AttributeError:  # pragma: no cover
            return self.create_exception_response(adu.pdu, "illegal_function")

        try:
            response = handler(adu.pdu)
            if self.debuglevel:  # pragma: no cover
                response.encode()
                # debug
                print("Device response:", file=sys.stderr)
                if self.debuglevel == 1:
                    print(response, file=sys.stderr)
                else:
                    print(response.introspect(), file=sys.stderr)
            return response
        except Exception:
            traceback.print_exc()
            return self.create_exception_response(adu.pdu, "server_device_failure")

    def create_exception_response(self, request_pdu, exception_code):
        """
        Request handler to response with an exception.

        Args:
            request_pdu: the request pdu
            exception_code: the exception code to be sent

        Returns:
            the response to be sent to the client
        """
        self.server.bus_exception_error_count += 1
        response = self.server.response_parser()
        response.pdu.function_code.set(request_pdu.function_code.get() | 0x80)
        response.pdu.data.exception_code.set(exception_code)

        if self.debuglevel:  # pragma: no cover
            # debug
            response.encode()
            print("Device exception: {}".format(exception_code), file=sys.stderr)
            if self.debuglevel == 1:
                print(response, file=sys.stderr)
            else:
                print(response.introspect(), file=sys.stderr)
        return response

    def handle_function_code_1(self, pdu):
        """
        Request handler for function code 1 ````Read Coils``.``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 1 <= pdu.data.quantity_of_coils.get() <= 0x07D0:
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not (pdu.data.starting_address.get() + pdu.data.quantity_of_coils.get()) <= 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            response = self.server.response_parser()
            response.pdu.function_code.set(1)
            byte_count = int(math.ceil(pdu.data.quantity_of_coils.get() / 8))
            response.pdu.data.byte_count.set(byte_count)
            coil_status = []
            quantity = pdu.data.quantity_of_coils.get()
            reference = pdu.data.starting_address.get()
            while quantity > 0:
                coil_status.append(self.memory.read_coils_as_int(reference, min(quantity, 8)))
                reference += 8
                quantity -= 8
            response.pdu.data.coil_status[:] = coil_status
            return response

    def handle_function_code_2(self, pdu):
        """
        Request handler for function code 2 ``Read Discrete Inputs``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 1 <= pdu.data.quantity_of_inputs.get() <= 0x07D0:
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not (pdu.data.starting_address.get() + pdu.data.quantity_of_inputs.get()) <= 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            response = self.server.response_parser()
            response.pdu.function_code.set(2)
            byte_count = int(math.ceil(pdu.data.quantity_of_inputs.get() / 8))
            response.pdu.data.byte_count.set(byte_count)
            input_status = []
            quantity = pdu.data.quantity_of_inputs.get()
            reference = pdu.data.starting_address.get()
            while quantity > 0:
                input_status.append(self.memory.read_discrete_input_as_int(reference, min(quantity, 8)))
                reference += 8
                quantity -= 8
            response.pdu.data.input_status[:] = input_status
            return response

    def handle_function_code_3(self, pdu):
        """
        Request handler for function code 3 ``Read Holding Registers``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 1 <= pdu.data.quantity_of_registers.get() <= 0x7D:
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not (pdu.data.starting_address.get() + pdu.data.quantity_of_registers.get()) <= 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            response = self.server.response_parser()
            response.pdu.function_code.set(3)
            response.pdu.data.byte_count.set(pdu.data.quantity_of_registers.get() * 2)
            registers = [self.memory.read_holding_register(pdu.data.starting_address.get() + i)
                         for i in range(pdu.data.quantity_of_registers.get())]

            var = oser.UBInt16()
            integers = [var.get() for reg in registers if var.decode(reg)]
            response.pdu.data.register_value[:] = integers
            return response

    def handle_function_code_4(self, pdu):
        """
        Request handler for function code 4 ``Read Input Registers``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 1 <= pdu.data.quantity_of_registers.get() <= 0x7D:  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not (pdu.data.starting_address.get() + pdu.data.quantity_of_registers.get()) <= 0xFFFF:  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            response = self.server.response_parser()
            response.pdu.function_code.set(4)
            response.pdu.data.byte_count.set(pdu.data.quantity_of_registers.get() * 2)
            registers = [self.memory.read_input_register(pdu.data.starting_address.get() + i)
                         for i in range(pdu.data.quantity_of_registers.get())]
            var = oser.UBInt16()
            integers = [var.get() for reg in registers if var.decode(reg)]
            response.pdu.data.input_registers[:] = integers
            return response

    def handle_function_code_5(self, pdu):
        """
        Request handler for function code 5 ``Write Single Coil``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not (pdu.data.output_value.get() == 0x0000 or pdu.data.output_value.get() == 0xFF00):  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not pdu.data.output_address.get() < 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            if pdu.data.output_value.get():
                self.memory.write_coil(pdu.data.output_address.get(), 1)
            else:
                self.memory.write_coil(pdu.data.output_address.get(), 0)
            response = self.server.response_parser()
            response.pdu.function_code.set(5)
            response.pdu.data.output_address.set(pdu.data.output_address.get())
            response.pdu.data.output_value.set(pdu.data.output_value.get())
            return response

    def handle_function_code_6(self, pdu):
        """
        Request handler for function code 6 ``Write Single Register``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 0x0000 <= pdu.data.register_value.get() <= 0xFFFF:  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not pdu.data.register_address.get() < 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            self.memory.write_holding_register(pdu.data.register_address.get(), pdu.data.register_value.encode())
            response = self.server.response_parser()
            response.pdu.function_code.set(6)
            response.pdu.data.register_address.set(pdu.data.register_address.get())
            response.pdu.data.register_value.set(pdu.data.register_value.get())
            return response

    def handle_function_code_15(self, pdu):
        """
        Request handler for function code 15 ``Write Multiple Coils``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        quantity_of_outputs = pdu.data.quantity_of_outputs.get()
        expected_byte_count = int(math.ceil(quantity_of_outputs / 8))
        if not 1 <= quantity_of_outputs <= 0x07B0 or pdu.data.byte_count.get() != expected_byte_count:
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not (pdu.data.starting_address.get() + expected_byte_count) <= 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            self.memory.write_multiple_coils(pdu.data.starting_address.get(), pdu.data.quantity_of_outputs.get(),
                                             pdu.data.outputs_value.get())
            response = self.server.response_parser()
            response.pdu.function_code.set(15)
            response.pdu.data.starting_address.set(pdu.data.starting_address.get())
            response.pdu.data.quantity_of_outputs.set(pdu.data.quantity_of_outputs.get())
            return response

    def handle_function_code_16(self, pdu):
        """
        Request handler for function code 16 ``Write Multiple Registers``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        quantity_of_registers = pdu.data.quantity_of_registers.get()
        expected_byte_count = quantity_of_registers * 2
        if not 1 <= quantity_of_registers <= 0x007B or pdu.data.byte_count.get() != expected_byte_count:
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not (pdu.data.starting_address.get() + expected_byte_count) <= 0xFFFF+1:
            return self.create_exception_response(pdu, "illegal_data_address")
        else:
            register_values = [reg.encode() for reg in pdu.data.outputs_value]
            assert len(register_values) == quantity_of_registers
            self.memory.write_multiple_holding_registers(pdu.data.starting_address.get(),
                                                         register_values)
            response = self.server.response_parser()
            response.pdu.function_code.set(16)
            response.pdu.data.starting_address.set(pdu.data.starting_address.get())
            response.pdu.data.quantity_of_registers.set(pdu.data.quantity_of_registers.get())
            return response

    def handle_function_code_20(self, pdu):
        """
        Request handler for function code 20 ``Read File Record``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 0x07 <= pdu.data.byte_count.get() <= 0xF5:  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_value")
        for sub_request in pdu.data.sub_requests:
            if not sub_request.reference_type.get() == 0x06 or \
                    not 0x0001 <= sub_request.memory_number.get() <= 0xFFFF or \
                    not 0x0000 <= sub_request.record_number.get() <= 0x270F or \
                    not sub_request.record_number.get() + sub_request.record_length.get() <= 0x270F:
                return self.create_exception_response(pdu, "illegal_data_address")

        response = self.server.response_parser()
        response.pdu.function_code.set(20)
        response.pdu.data.sub_responses.set_length(len(pdu.data.sub_requests))

        for sub_request, sub_response in zip(pdu.data.sub_requests, response.pdu.data.sub_responses):
            sub_response.reference_type.set(sub_request.reference_type.get())
            sub_response.response_length.set(2 * sub_request.record_length.get() + 1)

            memory_number = sub_request.memory_number.get()
            record_number = sub_request.record_number.get()
            record_length = sub_request.record_length.get()

            try:
                sub_response.record_data[:] = \
                    self.extended_memory[memory_number][record_number:record_number+record_length+1]
            except (KeyError, IndexError):
                return self.create_exception_response(pdu, "illegal_data_address")

        return response

    def handle_function_code_21(self, pdu):
        """
        Request handler for function code 21 ``Write File Record``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 0x09 <= pdu.data.byte_count.get() <= 0xFB:  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_value")
        for sub_request in pdu.data.sub_requests:
            if not sub_request.reference_type.get() == 0x06 or \
                    not 0x0001 <= sub_request.memory_number.get() <= 0xFFFF or \
                    not 0x0000 <= sub_request.record_number.get() <= 0x270F or \
                    not sub_request.record_number.get() + sub_request.record_length.get() <= 0x270F:
                return self.create_exception_response(pdu, "illegal_data_address")

        response = self.server.response_parser()
        response.pdu.function_code.set(21)
        response.pdu.data.sub_responses.set_length(len(pdu.data.sub_requests))

        for sub_request, sub_response in zip(pdu.data.sub_requests, response.pdu.data.sub_responses):
            sub_response.reference_type.set(sub_request.reference_type.get())

            memory_number = sub_request.memory_number.get()
            record_number = sub_request.record_number.get()
            record_length = sub_request.record_length.get()
            records = [data.get() for data in sub_request.record_data]

            sub_response.memory_number.set(memory_number)
            sub_response.record_number.set(record_number)
            sub_response.record_length.set(record_length)
            sub_response.record_data[:] = records

            try:
                self.extended_memory[memory_number]
            except KeyError:
                self.extended_memory[memory_number] = []

            if len(self.extended_memory[memory_number]) < record_number:  # initially fill memory with zeros
                self.extended_memory[memory_number] = [0 for _ in range(record_number)]
            self.extended_memory[memory_number][record_number:record_number+record_length] = records

        return response

    def handle_function_code_22(self, pdu):
        """
        Request handler for function code 22 ``Mask Write Register``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 0x0000 <= pdu.data.reference_address.get() < 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")
        elif not 0x0000 <= pdu.data.and_mask.get() <= 0xFFFF \
                and 0x0000 <= pdu.data.or_mask.get() <= 0xFFFF:  # pragma: no cover
            return self.create_exception_response(pdu, "illegal_data_value")
        reference = pdu.data.reference_address.get()
        current_content = self.memory.read_holding_register(pdu.data.reference_address.get())
        var = oser.UBInt16()
        var.decode(current_content)
        current_content_int = var.get()
        and_mask = pdu.data.and_mask.get()
        or_mask = pdu.data.or_mask.get()
        masked_value = (current_content_int & and_mask) | (or_mask & (~and_mask))
        self.memory.write_holding_register(reference, oser.UBInt16(masked_value).encode())

        response = self.server.response_parser()
        response.pdu.function_code.set(22)
        response.pdu.data.reference_address.set(reference)
        response.pdu.data.and_mask.set(and_mask)
        response.pdu.data.or_mask.set(or_mask)
        return response

    def handle_function_code_23(self, pdu):
        """
        Request handler for function code 23 ``Read/Write Multiple Registers``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if not 0x0001 <= pdu.data.quantity_to_read.get() <= 0x007D or \
                not 0x0001 <= pdu.data.quantity_to_write.get() <= 0x0079 or \
                not pdu.data.write_byte_count.get() == 2 * pdu.data.quantity_to_write.get():
            return self.create_exception_response(pdu, "illegal_data_value")
        elif not 0x0000 <= pdu.data.read_starting_address.get() <= 0xFFFF or \
                not 0x0000 <= pdu.data.write_starting_address.get() <= 0xFFFF or \
                not 0x0000 <= pdu.data.read_starting_address.get() + pdu.data.quantity_to_read.get() - 1 <= 0xFFFF or \
                not 0x0000 <= pdu.data.write_starting_address.get() + (
                        pdu.data.write_byte_count.get() // 2) - 1 <= 0xFFFF:
            return self.create_exception_response(pdu, "illegal_data_address")

        register_values = [reg.encode() for reg in pdu.data.write_registers_value]
        self.memory.write_multiple_holding_registers(pdu.data.write_starting_address.get(),
                                                     register_values)

        response = self.server.response_parser()
        response.pdu.function_code.set(23)
        response.pdu.data.byte_count.set(pdu.data.quantity_to_read.get() * 2)
        registers = [self.memory.read_holding_register(pdu.data.read_starting_address.get() + i) for i in
                     range(pdu.data.quantity_to_read.get())]
        var = oser.UBInt16()
        integers = [var.get() for reg in registers if var.decode(reg)]

        response.pdu.data.read_register_value[:] = integers
        return response

    def handle_function_code_24(self, pdu):
        """
        Request handler for function code 24 ``Read FIFO Queue``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        fifo_pointer_address = pdu.data.fifo_pointer_address.get()

        if fifo_pointer_address not in self.fifo_queues:
            return self.create_exception_response(pdu, "illegal_data_address")

        fifo_queue = self.fifo_queues[fifo_pointer_address][:31 + 1]  # maximum 31 elements

        response = self.server.response_parser()
        response.pdu.function_code.set(24)
        response.pdu.data.byte_count.set(len(fifo_queue) * 2)
        response.pdu.data.fifo_count.set(len(fifo_queue))
        response.pdu.data.fifo_value_register[:] = fifo_queue

        return response

    def handle_function_code_43(self, pdu):
        """
        Encapsulated Interface Transport.

        Calls a handler by mei type with the name ``handle_function_code_43_mei_type_<mei type>``. To add more
        mei types more handlers have to be added.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        mei_type = pdu.data.mei_type.get()
        try:
            return getattr(self, 'handle_function_code_43_mei_type_{}'.format(mei_type))(pdu)
        except AttributeError:
            return self.create_exception_response(pdu, "illegal_function")

    def handle_function_code_43_mei_type_14(self, pdu):
        """
        Encapsulated Interface Transport: Read Device Identification.
        Basic object ids can be read.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """

        if not 0x01 <= pdu.data.read_device_id_code.get_value() <= 0x04:
            return self.create_exception_response(pdu, "illegal_data_value")

        object_id = pdu.data.object_id.get_value()

        response = self.server.response_parser()
        response.pdu.function_code.set(43)
        response.pdu.data.mei_type.set(14)
        response.pdu.data.read_device_id_code.set(pdu.data.read_device_id_code.get())
        response.pdu.data.conformity_level.set(0x83)
        response.pdu.data.number_of_objects.set(0x01)

        data_object = response.pdu.data.list_of_objects[0]
        data_object.object_id.set(object_id)

        object_id_keys = list(sorted(self.device_identifier.keys()))

        if object_id in self.device_identifier:
            try:
                index = object_id_keys.index(object_id)
                response.pdu.data.next_object_id.set(object_id_keys[index + 1])
                response.pdu.data.more_follows.set(0xFF)
            except (ValueError, IndexError):
                # no more data objects that can be served
                response.pdu.data.more_follows.set(0x00)

            # set data
            value = self.device_identifier[object_id]
            data_object.length.set(len(value))
            data_object.value.set(value)
        else:
            return self.create_exception_response(pdu, "illegal_data_address")

        return response

    # Convenience functions
    # =====================

    # write Flag to coil
    def write_coil_flag(self, reference, value):
        """
        Write a bit to coil memory.

        Args:
            reference: the reference to write to
            value: the bool value to be written
        """
        data = oser.Flag(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data=data)

    # read Flag from coil
    def read_coil_flag(self, reference):
        """
        Read a bit from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_coil_data(reference, data).data.get()

    # write list of Flag to coil
    def write_coil_flag_list(self, reference, values):
        """
        Write a list of bits to coil memory.

        Args:
            reference: the reference to write to
            values: a list of bool values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Flag)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data)

    # read list of Flag from coil
    def read_coil_flag_list(self, reference, quantity):
        """
        Read a list of bits from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_coil_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Nibble to coil
    def write_coil_nibble(self, reference, value):
        """
        Write a nibble to coil memory.

        Args:
            reference: the reference to write to
            value: the nibble value to be written
        """
        data = oser.Nibble(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data=data)

    # read Nibble from coil
    def read_coil_nibble(self, reference):
        """
        Read a nibble from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_coil_data(reference, data).data.get()

    # write list of Nibble to coil
    def write_coil_nibble_list(self, reference, values):
        """
        Write a list of nibbles to coil memory.

        Args:
            reference: the reference to write to
            values: a list of nibble values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Nibble)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data)

    # read list of Nibble from coil
    def read_coil_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from coil memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_coil_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write BitField to coil
    def write_coil_bitfield(self, reference, value, length):
        """
        Write a bitfield to coil memory.

        Args:
            reference: the reference to write to
            value: the bitfield value to be written
            length: the length of the value
        """
        data = oser.BitField(value=value, length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data=data)

    # read BitField from coil
    def read_coil_bitfield(self, reference, length):
        """
        Read a bitfield from coil memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_coil_data(reference, data).data.get()

    # write list of BitField to coil
    def write_coil_bitfield_list(self, reference, values, length):
        """
        Write a list of bitfields to coil memory.

        Args:
            reference: the reference to write to
            values: a list of bitfield values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.BitField, kwargs={'length': length})
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_coil_data(reference, data)

    # read list of BitField from coil
    def read_coil_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from coil memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_coil_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Flag to discrete input
    def write_discrete_input_flag(self, reference, value):
        """
        Write a bit to discrete input memory.

        Args:
            reference: the reference to write to
            value: the bool value to be written
        """
        data = oser.Flag(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_discrete_input_data(reference, data=data)

    # read Flag from discrete input
    def read_discrete_input_flag(self, reference):
        """
        Read a bit from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_discrete_input_data(reference, data).data.get()

    # write list of Flag to discrete input
    def write_discrete_input_flag_list(self, reference, values):
        """
        Write a list of bits to discrete input memory.

        Args:
            reference: the reference to write to
            values: a list of bool values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Flag)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_discrete_input_data(reference, data)

    # read list of Flag from discrete input
    def read_discrete_input_flag_list(self, reference, quantity):
        """
        Read a list of bits from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_discrete_input_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Nibble to discrete input
    def write_discrete_input_nibble(self, reference, value):
        """
        Write a nibble to discrete input memory.

        Args:
            reference: the reference to write to
            value: the nibble value to be written
        """
        data = oser.Nibble(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_discrete_input_data(reference, data=data)

    # read Nibble from discrete input
    def read_discrete_input_nibble(self, reference):
        """
        Read a nibble from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_discrete_input_data(reference, data).data.get()

    # write list of Nibble to discrete input
    def write_discrete_input_nibble_list(self, reference, values):
        """
        Write a list of nibbles to discrete input memory.

        Args:
            reference: the reference to write to
            values: a list of nibble values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Nibble)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_discrete_input_data(reference, data)

    # read list of Nibble from discrete input
    def read_discrete_input_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from discrete input memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_discrete_input_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write BitField to discrete input
    def write_discrete_input_bitfield(self, reference, value, length):
        """
        Write a bitfield to discrete input memory.

        Args:
            reference: the reference to write to
            value: the bitfield value to be written
            length: the length of the value
        """
        data = oser.BitField(value=value, length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_discrete_input_data(reference, data=data)

    # read BitField from discrete input
    def read_discrete_input_bitfield(self, reference, length):
        """
        Read a bitfield from discrete input memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_discrete_input_data(reference, data).data.get()

    # write list of BitField to discrete input
    def write_discrete_input_bitfield_list(self, reference, values, length):
        """
        Write a list of bitfields to discrete input memory.

        Args:
            reference: the reference to write to
            values: a list of bitfield values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.BitField, kwargs={'length': length})
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_discrete_input_data(reference, data)

    # read list of BitField from discrete input
    def read_discrete_input_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from discrete input memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_discrete_input_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write UBInt8 to holding register
    def write_holding_register_ubint8(self, reference, value):
        """
        Write a unsigned 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt8 from holding register
    def read_holding_register_ubint8(self, reference):
        """
        Read a unsigned 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt8 to holding register
    def write_holding_register_ubint8_list(self, reference, values):
        """
        Write a list of unsigned 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt8 from holding register
    def read_holding_register_ubint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt8 to holding register
    def write_holding_register_ulint8(self, reference, value):
        """
        Write a unsigned 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt8 from holding register
    def read_holding_register_ulint8(self, reference):
        """
        Read a unsigned 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt8 to holding register
    def write_holding_register_ulint8_list(self, reference, values):
        """
        Write a list of unsigned 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt8 from holding register
    def read_holding_register_ulint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt8 to holding register
    def write_holding_register_sbint8(self, reference, value):
        """
        Write a signed 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt8 from holding register
    def read_holding_register_sbint8(self, reference):
        """
        Read a signed 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt8 to holding register
    def write_holding_register_sbint8_list(self, reference, values):
        """
        Write a list of signed 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt8 from holding register
    def read_holding_register_sbint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt8 to holding register
    def write_holding_register_slint8(self, reference, value):
        """
        Write a signed 8-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt8(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt8 from holding register
    def read_holding_register_slint8(self, reference):
        """
        Read a signed 8-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt8()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt8 to holding register
    def write_holding_register_slint8_list(self, reference, values):
        """
        Write a list of signed 8-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt8)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt8 from holding register
    def read_holding_register_slint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt8)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt16 to holding register
    def write_holding_register_ubint16(self, reference, value):
        """
        Write a unsigned 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt16 from holding register
    def read_holding_register_ubint16(self, reference):
        """
        Read a unsigned 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt16 to holding register
    def write_holding_register_ubint16_list(self, reference, values):
        """
        Write a list of unsigned 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt16 from holding register
    def read_holding_register_ubint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt16 to holding register
    def write_holding_register_ulint16(self, reference, value):
        """
        Write a unsigned 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt16 from holding register
    def read_holding_register_ulint16(self, reference):
        """
        Read a unsigned 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt16 to holding register
    def write_holding_register_ulint16_list(self, reference, values):
        """
        Write a list of unsigned 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt16 from holding register
    def read_holding_register_ulint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt16 to holding register
    def write_holding_register_sbint16(self, reference, value):
        """
        Write a signed 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt16 from holding register
    def read_holding_register_sbint16(self, reference):
        """
        Read a signed 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt16 to holding register
    def write_holding_register_sbint16_list(self, reference, values):
        """
        Write a list of signed 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt16 from holding register
    def read_holding_register_sbint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt16 to holding register
    def write_holding_register_slint16(self, reference, value):
        """
        Write a signed 16-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt16(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt16 from holding register
    def read_holding_register_slint16(self, reference):
        """
        Read a signed 16-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt16()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt16 to holding register
    def write_holding_register_slint16_list(self, reference, values):
        """
        Write a list of signed 16-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt16)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt16 from holding register
    def read_holding_register_slint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt16)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt32 to holding register
    def write_holding_register_ubint32(self, reference, value):
        """
        Write a unsigned 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt32 from holding register
    def read_holding_register_ubint32(self, reference):
        """
        Read a unsigned 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt32 to holding register
    def write_holding_register_ubint32_list(self, reference, values):
        """
        Write a list of unsigned 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt32 from holding register
    def read_holding_register_ubint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt32 to holding register
    def write_holding_register_ulint32(self, reference, value):
        """
        Write a unsigned 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt32 from holding register
    def read_holding_register_ulint32(self, reference):
        """
        Read a unsigned 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt32 to holding register
    def write_holding_register_ulint32_list(self, reference, values):
        """
        Write a list of unsigned 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt32 from holding register
    def read_holding_register_ulint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt32 to holding register
    def write_holding_register_sbint32(self, reference, value):
        """
        Write a signed 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt32 from holding register
    def read_holding_register_sbint32(self, reference):
        """
        Read a signed 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt32 to holding register
    def write_holding_register_sbint32_list(self, reference, values):
        """
        Write a list of signed 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt32 from holding register
    def read_holding_register_sbint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt32 to holding register
    def write_holding_register_slint32(self, reference, value):
        """
        Write a signed 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt32(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt32 from holding register
    def read_holding_register_slint32(self, reference):
        """
        Read a signed 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt32()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt32 to holding register
    def write_holding_register_slint32_list(self, reference, values):
        """
        Write a list of signed 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt32)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt32 from holding register
    def read_holding_register_slint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt32)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt2143 to holding register
    def write_holding_register_uint2143(self, reference, value):
        """
        Write a unsigned 2143-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt2143(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt2143 from holding register
    def read_holding_register_uint2143(self, reference):
        """
        Read a unsigned 2143-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt2143()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt2143 to holding register
    def write_holding_register_uint2143_list(self, reference, values):
        """
        Write a list of unsigned 2143-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt2143)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt2143 from holding register
    def read_holding_register_uint2143_list(self, reference, quantity):
        """
        Read a list of unsigned 2143-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt2143)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt3412 to holding register
    def write_holding_register_uint3412(self, reference, value):
        """
        Write a unsigned 3412-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt3412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt3412 from holding register
    def read_holding_register_uint3412(self, reference):
        """
        Read a unsigned 3412-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt3412()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt3412 to holding register
    def write_holding_register_uint3412_list(self, reference, values):
        """
        Write a list of unsigned 3412-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt3412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt3412 from holding register
    def read_holding_register_uint3412_list(self, reference, quantity):
        """
        Read a list of unsigned 3412-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt3412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt2143 to holding register
    def write_holding_register_sint2143(self, reference, value):
        """
        Write a signed 2143-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt2143(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt2143 from holding register
    def read_holding_register_sint2143(self, reference):
        """
        Read a signed 2143-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt2143()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt2143 to holding register
    def write_holding_register_sint2143_list(self, reference, values):
        """
        Write a list of signed 2143-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt2143)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt2143 from holding register
    def read_holding_register_sint2143_list(self, reference, quantity):
        """
        Read a list of signed 2143-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt2143)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt3412 to holding register
    def write_holding_register_sint3412(self, reference, value):
        """
        Write a signed 3412-endian 32-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt3412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt3412 from holding register
    def read_holding_register_sint3412(self, reference):
        """
        Read a signed 3412-endian 32-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt3412()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt3412 to holding register
    def write_holding_register_sint3412_list(self, reference, values):
        """
        Write a list of signed 3412-endian 32-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt3412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt3412 from holding register
    def read_holding_register_sint3412_list(self, reference, quantity):
        """
        Read a list of signed 3412-endian 32-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt3412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt64 to holding register
    def write_holding_register_ubint64(self, reference, value):
        """
        Write a unsigned 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UBInt64 from holding register
    def read_holding_register_ubint64(self, reference):
        """
        Read a unsigned 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of UBInt64 to holding register
    def write_holding_register_ubint64_list(self, reference, values):
        """
        Write a list of unsigned 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UBInt64 from holding register
    def read_holding_register_ubint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt64 to holding register
    def write_holding_register_ulint64(self, reference, value):
        """
        Write a unsigned 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read ULInt64 from holding register
    def read_holding_register_ulint64(self, reference):
        """
        Read a unsigned 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of ULInt64 to holding register
    def write_holding_register_ulint64_list(self, reference, values):
        """
        Write a list of unsigned 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of ULInt64 from holding register
    def read_holding_register_ulint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt64 to holding register
    def write_holding_register_sbint64(self, reference, value):
        """
        Write a signed 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SBInt64 from holding register
    def read_holding_register_sbint64(self, reference):
        """
        Read a signed 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of SBInt64 to holding register
    def write_holding_register_sbint64_list(self, reference, values):
        """
        Write a list of signed 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SBInt64 from holding register
    def read_holding_register_sbint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt64 to holding register
    def write_holding_register_slint64(self, reference, value):
        """
        Write a signed 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt64(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SLInt64 from holding register
    def read_holding_register_slint64(self, reference):
        """
        Read a signed 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt64()
        return self.read_holding_register_data(reference, data).get()

    # write list of SLInt64 to holding register
    def write_holding_register_slint64_list(self, reference, values):
        """
        Write a list of signed 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt64)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SLInt64 from holding register
    def read_holding_register_slint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt64)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt21436587 to holding register
    def write_holding_register_uint21436587(self, reference, value):
        """
        Write a unsigned 21436587-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt21436587(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt21436587 from holding register
    def read_holding_register_uint21436587(self, reference):
        """
        Read a unsigned 21436587-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt21436587()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt21436587 to holding register
    def write_holding_register_uint21436587_list(self, reference, values):
        """
        Write a list of unsigned 21436587-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt21436587)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt21436587 from holding register
    def read_holding_register_uint21436587_list(self, reference, quantity):
        """
        Read a list of unsigned 21436587-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt21436587)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt78563412 to holding register
    def write_holding_register_uint78563412(self, reference, value):
        """
        Write a unsigned 78563412-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt78563412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read UInt78563412 from holding register
    def read_holding_register_uint78563412(self, reference):
        """
        Read a unsigned 78563412-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt78563412()
        return self.read_holding_register_data(reference, data).get()

    # write list of UInt78563412 to holding register
    def write_holding_register_uint78563412_list(self, reference, values):
        """
        Write a list of unsigned 78563412-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt78563412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of UInt78563412 from holding register
    def read_holding_register_uint78563412_list(self, reference, quantity):
        """
        Read a list of unsigned 78563412-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt78563412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt21436587 to holding register
    def write_holding_register_sint21436587(self, reference, value):
        """
        Write a signed 21436587-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt21436587(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt21436587 from holding register
    def read_holding_register_sint21436587(self, reference):
        """
        Read a signed 21436587-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt21436587()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt21436587 to holding register
    def write_holding_register_sint21436587_list(self, reference, values):
        """
        Write a list of signed 21436587-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt21436587)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt21436587 from holding register
    def read_holding_register_sint21436587_list(self, reference, quantity):
        """
        Read a list of signed 21436587-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt21436587)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt78563412 to holding register
    def write_holding_register_sint78563412(self, reference, value):
        """
        Write a signed 78563412-endian 64-bit integer to holding register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt78563412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read SInt78563412 from holding register
    def read_holding_register_sint78563412(self, reference):
        """
        Read a signed 78563412-endian 64-bit integer from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt78563412()
        return self.read_holding_register_data(reference, data).get()

    # write list of SInt78563412 to holding register
    def write_holding_register_sint78563412_list(self, reference, values):
        """
        Write a list of signed 78563412-endian 64-bit integers to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt78563412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of SInt78563412 from holding register
    def read_holding_register_sint78563412_list(self, reference, quantity):
        """
        Read a list of signed 78563412-endian 64-bit integers from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt78563412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write BFloat to holding register
    def write_holding_register_bfloat(self, reference, value):
        """
        Write a float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.BFloat(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read BFloat from holding register
    def read_holding_register_bfloat(self, reference):
        """
        Read a float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.BFloat()
        return self.read_holding_register_data(reference, data).get()

    # write list of BFloat to holding register
    def write_holding_register_bfloat_list(self, reference, values):
        """
        Write a list of floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.BFloat)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of BFloat from holding register
    def read_holding_register_bfloat_list(self, reference, quantity):
        """
        Read a list of floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BFloat)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write LFloat to holding register
    def write_holding_register_lfloat(self, reference, value):
        """
        Write a float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.LFloat(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read LFloat from holding register
    def read_holding_register_lfloat(self, reference):
        """
        Read a float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.LFloat()
        return self.read_holding_register_data(reference, data).get()

    # write list of LFloat to holding register
    def write_holding_register_lfloat_list(self, reference, values):
        """
        Write a list of floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.LFloat)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of LFloat from holding register
    def read_holding_register_lfloat_list(self, reference, quantity):
        """
        Read a list of floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LFloat)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Float2143 to holding register
    def write_holding_register_float2143(self, reference, value):
        """
        Write a 2143-endian float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.Float2143(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Float2143 from holding register
    def read_holding_register_float2143(self, reference):
        """
        Read a 2143-endian float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float2143()
        return self.read_holding_register_data(reference, data).get()

    # write list of Float2143 to holding register
    def write_holding_register_float2143_list(self, reference, values):
        """
        Write a list of 2143-endian floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Float2143)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Float2143 from holding register
    def read_holding_register_float2143_list(self, reference, quantity):
        """
        Read a list of 2143-endian floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float2143)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Float3412 to holding register
    def write_holding_register_float3412(self, reference, value):
        """
        Write a 3412-endian float to holding register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.Float3412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Float3412 from holding register
    def read_holding_register_float3412(self, reference):
        """
        Read a 3412-endian float from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float3412()
        return self.read_holding_register_data(reference, data).get()

    # write list of Float3412 to holding register
    def write_holding_register_float3412_list(self, reference, values):
        """
        Write a list of 3412-endian floats to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Float3412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Float3412 from holding register
    def read_holding_register_float3412_list(self, reference, quantity):
        """
        Read a list of 3412-endian floats from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float3412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write BDouble to holding register
    def write_holding_register_bdouble(self, reference, value):
        """
        Write a double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.BDouble(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read BDouble from holding register
    def read_holding_register_bdouble(self, reference):
        """
        Read a double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.BDouble()
        return self.read_holding_register_data(reference, data).get()

    # write list of BDouble to holding register
    def write_holding_register_bdouble_list(self, reference, values):
        """
        Write a list of doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.BDouble)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of BDouble from holding register
    def read_holding_register_bdouble_list(self, reference, quantity):
        """
        Read a list of doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BDouble)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write LDouble to holding register
    def write_holding_register_ldouble(self, reference, value):
        """
        Write a double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.LDouble(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read LDouble from holding register
    def read_holding_register_ldouble(self, reference):
        """
        Read a double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.LDouble()
        return self.read_holding_register_data(reference, data).get()

    # write list of LDouble to holding register
    def write_holding_register_ldouble_list(self, reference, values):
        """
        Write a list of doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.LDouble)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of LDouble from holding register
    def read_holding_register_ldouble_list(self, reference, quantity):
        """
        Read a list of doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LDouble)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Double21436587 to holding register
    def write_holding_register_double21436587(self, reference, value):
        """
        Write a 21436587-endian double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.Double21436587(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Double21436587 from holding register
    def read_holding_register_double21436587(self, reference):
        """
        Read a 21436587-endian double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double21436587()
        return self.read_holding_register_data(reference, data).get()

    # write list of Double21436587 to holding register
    def write_holding_register_double21436587_list(self, reference, values):
        """
        Write a list of 21436587-endian doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Double21436587)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Double21436587 from holding register
    def read_holding_register_double21436587_list(self, reference, quantity):
        """
        Read a list of 21436587-endian doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double21436587)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Double78563412 to holding register
    def write_holding_register_double78563412(self, reference, value):
        """
        Write a 78563412-endian double to holding register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.Double78563412(value=value)
        return self.write_holding_register_data(reference, data=data)

    # read Double78563412 from holding register
    def read_holding_register_double78563412(self, reference):
        """
        Read a 78563412-endian double from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double78563412()
        return self.read_holding_register_data(reference, data).get()

    # write list of Double78563412 to holding register
    def write_holding_register_double78563412_list(self, reference, values):
        """
        Write a list of 78563412-endian doubles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Double78563412)
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of Double78563412 from holding register
    def read_holding_register_double78563412_list(self, reference, quantity):
        """
        Read a list of 78563412-endian doubles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double78563412)
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write String to holding register
    def write_holding_register_string(self, reference, value, length):
        """
        Write a string to holding register memory.

        Args:
            reference: the reference to write to
            value: the data value to be written
            length: the length of the value
        """
        data = oser.String(value=value, length=length)
        return self.write_holding_register_data(reference, data=data)

    # read String from holding register
    def read_holding_register_string(self, reference, length):
        """
        Read a string from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            data: value read from ``reference``
        """
        data = oser.String(length=length)
        return self.read_holding_register_data(reference, data).get()

    # write list of String to holding register
    def write_holding_register_string_list(self, reference, values, length):
        """
        Write a list of strings to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of data values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.String, kwargs={'length': length})
        data[:] = values
        return self.write_holding_register_data(reference, data)

    # read list of String from holding register
    def read_holding_register_string_list(self, reference, quantity, length):
        """
        Read a list of strings from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of data: data values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.String, kwargs={'length': length})
        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Flag to holding register
    def write_holding_register_flag(self, reference, value):
        """
        Write a bit to holding register memory.

        Args:
            reference: the reference to write to
            value: the bool value to be written
        """
        data = oser.Flag(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data=data)

    # read Flag from holding register
    def read_holding_register_flag(self, reference):
        """
        Read a bit from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_holding_register_data(reference, data).data.get()

    # write list of Flag to holding register
    def write_holding_register_flag_list(self, reference, values):
        """
        Write a list of bits to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of bool values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Flag)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data)

    # read list of Flag from holding register
    def read_holding_register_flag_list(self, reference, quantity):
        """
        Read a list of bits from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Nibble to holding register
    def write_holding_register_nibble(self, reference, value):
        """
        Write a nibble to holding register memory.

        Args:
            reference: the reference to write to
            value: the nibble value to be written
        """
        data = oser.Nibble(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data=data)

    # read Nibble from holding register
    def read_holding_register_nibble(self, reference):
        """
        Read a nibble from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_holding_register_data(reference, data).data.get()

    # write list of Nibble to holding register
    def write_holding_register_nibble_list(self, reference, values):
        """
        Write a list of nibbles to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of nibble values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Nibble)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data)

    # read list of Nibble from holding register
    def read_holding_register_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from holding register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write BitField to holding register
    def write_holding_register_bitfield(self, reference, value, length):
        """
        Write a bitfield to holding register memory.

        Args:
            reference: the reference to write to
            value: the bitfield value to be written
            length: the length of the value
        """
        data = oser.BitField(value=value, length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data=data)

    # read BitField from holding register
    def read_holding_register_bitfield(self, reference, length):
        """
        Read a bitfield from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_holding_register_data(reference, data).data.get()

    # write list of BitField to holding register
    def write_holding_register_bitfield_list(self, reference, values, length):
        """
        Write a list of bitfields to holding register memory.

        Args:
            reference: the reference to write to
            values: a list of bitfield values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.BitField, kwargs={'length': length})
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_holding_register_data(reference, data)

    # read list of BitField from holding register
    def read_holding_register_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from holding register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_holding_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write UBInt8 to input register
    def write_input_register_ubint8(self, reference, value):
        """
        Write a unsigned 8-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt8(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UBInt8 from input register
    def read_input_register_ubint8(self, reference):
        """
        Read a unsigned 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt8()
        return self.read_input_register_data(reference, data).get()

    # write list of UBInt8 to input register
    def write_input_register_ubint8_list(self, reference, values):
        """
        Write a list of unsigned 8-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt8)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UBInt8 from input register
    def read_input_register_ubint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt8 to input register
    def write_input_register_ulint8(self, reference, value):
        """
        Write a unsigned 8-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt8(value=value)
        return self.write_input_register_data(reference, data=data)

    # read ULInt8 from input register
    def read_input_register_ulint8(self, reference):
        """
        Read a unsigned 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt8()
        return self.read_input_register_data(reference, data).get()

    # write list of ULInt8 to input register
    def write_input_register_ulint8_list(self, reference, values):
        """
        Write a list of unsigned 8-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt8)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of ULInt8 from input register
    def read_input_register_ulint8_list(self, reference, quantity):
        """
        Read a list of unsigned 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt8 to input register
    def write_input_register_sbint8(self, reference, value):
        """
        Write a signed 8-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt8(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SBInt8 from input register
    def read_input_register_sbint8(self, reference):
        """
        Read a signed 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt8()
        return self.read_input_register_data(reference, data).get()

    # write list of SBInt8 to input register
    def write_input_register_sbint8_list(self, reference, values):
        """
        Write a list of signed 8-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt8)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SBInt8 from input register
    def read_input_register_sbint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt8 to input register
    def write_input_register_slint8(self, reference, value):
        """
        Write a signed 8-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt8(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SLInt8 from input register
    def read_input_register_slint8(self, reference):
        """
        Read a signed 8-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt8()
        return self.read_input_register_data(reference, data).get()

    # write list of SLInt8 to input register
    def write_input_register_slint8_list(self, reference, values):
        """
        Write a list of signed 8-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt8)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SLInt8 from input register
    def read_input_register_slint8_list(self, reference, quantity):
        """
        Read a list of signed 8-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt8)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt16 to input register
    def write_input_register_ubint16(self, reference, value):
        """
        Write a unsigned 16-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt16(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UBInt16 from input register
    def read_input_register_ubint16(self, reference):
        """
        Read a unsigned 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt16()
        return self.read_input_register_data(reference, data).get()

    # write list of UBInt16 to input register
    def write_input_register_ubint16_list(self, reference, values):
        """
        Write a list of unsigned 16-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt16)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UBInt16 from input register
    def read_input_register_ubint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt16 to input register
    def write_input_register_ulint16(self, reference, value):
        """
        Write a unsigned 16-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt16(value=value)
        return self.write_input_register_data(reference, data=data)

    # read ULInt16 from input register
    def read_input_register_ulint16(self, reference):
        """
        Read a unsigned 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt16()
        return self.read_input_register_data(reference, data).get()

    # write list of ULInt16 to input register
    def write_input_register_ulint16_list(self, reference, values):
        """
        Write a list of unsigned 16-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt16)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of ULInt16 from input register
    def read_input_register_ulint16_list(self, reference, quantity):
        """
        Read a list of unsigned 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt16 to input register
    def write_input_register_sbint16(self, reference, value):
        """
        Write a signed 16-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt16(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SBInt16 from input register
    def read_input_register_sbint16(self, reference):
        """
        Read a signed 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt16()
        return self.read_input_register_data(reference, data).get()

    # write list of SBInt16 to input register
    def write_input_register_sbint16_list(self, reference, values):
        """
        Write a list of signed 16-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt16)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SBInt16 from input register
    def read_input_register_sbint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt16 to input register
    def write_input_register_slint16(self, reference, value):
        """
        Write a signed 16-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt16(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SLInt16 from input register
    def read_input_register_slint16(self, reference):
        """
        Read a signed 16-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt16()
        return self.read_input_register_data(reference, data).get()

    # write list of SLInt16 to input register
    def write_input_register_slint16_list(self, reference, values):
        """
        Write a list of signed 16-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt16)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SLInt16 from input register
    def read_input_register_slint16_list(self, reference, quantity):
        """
        Read a list of signed 16-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt16)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt32 to input register
    def write_input_register_ubint32(self, reference, value):
        """
        Write a unsigned 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt32(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UBInt32 from input register
    def read_input_register_ubint32(self, reference):
        """
        Read a unsigned 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt32()
        return self.read_input_register_data(reference, data).get()

    # write list of UBInt32 to input register
    def write_input_register_ubint32_list(self, reference, values):
        """
        Write a list of unsigned 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt32)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UBInt32 from input register
    def read_input_register_ubint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt32 to input register
    def write_input_register_ulint32(self, reference, value):
        """
        Write a unsigned 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt32(value=value)
        return self.write_input_register_data(reference, data=data)

    # read ULInt32 from input register
    def read_input_register_ulint32(self, reference):
        """
        Read a unsigned 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt32()
        return self.read_input_register_data(reference, data).get()

    # write list of ULInt32 to input register
    def write_input_register_ulint32_list(self, reference, values):
        """
        Write a list of unsigned 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt32)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of ULInt32 from input register
    def read_input_register_ulint32_list(self, reference, quantity):
        """
        Read a list of unsigned 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt32 to input register
    def write_input_register_sbint32(self, reference, value):
        """
        Write a signed 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt32(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SBInt32 from input register
    def read_input_register_sbint32(self, reference):
        """
        Read a signed 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt32()
        return self.read_input_register_data(reference, data).get()

    # write list of SBInt32 to input register
    def write_input_register_sbint32_list(self, reference, values):
        """
        Write a list of signed 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt32)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SBInt32 from input register
    def read_input_register_sbint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt32 to input register
    def write_input_register_slint32(self, reference, value):
        """
        Write a signed 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt32(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SLInt32 from input register
    def read_input_register_slint32(self, reference):
        """
        Read a signed 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt32()
        return self.read_input_register_data(reference, data).get()

    # write list of SLInt32 to input register
    def write_input_register_slint32_list(self, reference, values):
        """
        Write a list of signed 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt32)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SLInt32 from input register
    def read_input_register_slint32_list(self, reference, quantity):
        """
        Read a list of signed 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt32)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt2143 to input register
    def write_input_register_uint2143(self, reference, value):
        """
        Write a unsigned 2143-endian 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt2143(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UInt2143 from input register
    def read_input_register_uint2143(self, reference):
        """
        Read a unsigned 2143-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt2143()
        return self.read_input_register_data(reference, data).get()

    # write list of UInt2143 to input register
    def write_input_register_uint2143_list(self, reference, values):
        """
        Write a list of unsigned 2143-endian 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt2143)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UInt2143 from input register
    def read_input_register_uint2143_list(self, reference, quantity):
        """
        Read a list of unsigned 2143-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt2143)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt3412 to input register
    def write_input_register_uint3412(self, reference, value):
        """
        Write a unsigned 3412-endian 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt3412(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UInt3412 from input register
    def read_input_register_uint3412(self, reference):
        """
        Read a unsigned 3412-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt3412()
        return self.read_input_register_data(reference, data).get()

    # write list of UInt3412 to input register
    def write_input_register_uint3412_list(self, reference, values):
        """
        Write a list of unsigned 3412-endian 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt3412)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UInt3412 from input register
    def read_input_register_uint3412_list(self, reference, quantity):
        """
        Read a list of unsigned 3412-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt3412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt2143 to input register
    def write_input_register_sint2143(self, reference, value):
        """
        Write a signed 2143-endian 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt2143(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SInt2143 from input register
    def read_input_register_sint2143(self, reference):
        """
        Read a signed 2143-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt2143()
        return self.read_input_register_data(reference, data).get()

    # write list of SInt2143 to input register
    def write_input_register_sint2143_list(self, reference, values):
        """
        Write a list of signed 2143-endian 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt2143)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SInt2143 from input register
    def read_input_register_sint2143_list(self, reference, quantity):
        """
        Read a list of signed 2143-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt2143)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt3412 to input register
    def write_input_register_sint3412(self, reference, value):
        """
        Write a signed 3412-endian 32-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt3412(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SInt3412 from input register
    def read_input_register_sint3412(self, reference):
        """
        Read a signed 3412-endian 32-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt3412()
        return self.read_input_register_data(reference, data).get()

    # write list of SInt3412 to input register
    def write_input_register_sint3412_list(self, reference, values):
        """
        Write a list of signed 3412-endian 32-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt3412)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SInt3412 from input register
    def read_input_register_sint3412_list(self, reference, quantity):
        """
        Read a list of signed 3412-endian 32-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt3412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UBInt64 to input register
    def write_input_register_ubint64(self, reference, value):
        """
        Write a unsigned 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UBInt64(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UBInt64 from input register
    def read_input_register_ubint64(self, reference):
        """
        Read a unsigned 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UBInt64()
        return self.read_input_register_data(reference, data).get()

    # write list of UBInt64 to input register
    def write_input_register_ubint64_list(self, reference, values):
        """
        Write a list of unsigned 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UBInt64)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UBInt64 from input register
    def read_input_register_ubint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UBInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write ULInt64 to input register
    def write_input_register_ulint64(self, reference, value):
        """
        Write a unsigned 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.ULInt64(value=value)
        return self.write_input_register_data(reference, data=data)

    # read ULInt64 from input register
    def read_input_register_ulint64(self, reference):
        """
        Read a unsigned 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.ULInt64()
        return self.read_input_register_data(reference, data).get()

    # write list of ULInt64 to input register
    def write_input_register_ulint64_list(self, reference, values):
        """
        Write a list of unsigned 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.ULInt64)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of ULInt64 from input register
    def read_input_register_ulint64_list(self, reference, quantity):
        """
        Read a list of unsigned 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.ULInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SBInt64 to input register
    def write_input_register_sbint64(self, reference, value):
        """
        Write a signed 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SBInt64(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SBInt64 from input register
    def read_input_register_sbint64(self, reference):
        """
        Read a signed 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SBInt64()
        return self.read_input_register_data(reference, data).get()

    # write list of SBInt64 to input register
    def write_input_register_sbint64_list(self, reference, values):
        """
        Write a list of signed 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SBInt64)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SBInt64 from input register
    def read_input_register_sbint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SBInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SLInt64 to input register
    def write_input_register_slint64(self, reference, value):
        """
        Write a signed 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SLInt64(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SLInt64 from input register
    def read_input_register_slint64(self, reference):
        """
        Read a signed 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SLInt64()
        return self.read_input_register_data(reference, data).get()

    # write list of SLInt64 to input register
    def write_input_register_slint64_list(self, reference, values):
        """
        Write a list of signed 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SLInt64)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SLInt64 from input register
    def read_input_register_slint64_list(self, reference, quantity):
        """
        Read a list of signed 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SLInt64)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt21436587 to input register
    def write_input_register_uint21436587(self, reference, value):
        """
        Write a unsigned 21436587-endian 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt21436587(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UInt21436587 from input register
    def read_input_register_uint21436587(self, reference):
        """
        Read a unsigned 21436587-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt21436587()
        return self.read_input_register_data(reference, data).get()

    # write list of UInt21436587 to input register
    def write_input_register_uint21436587_list(self, reference, values):
        """
        Write a list of unsigned 21436587-endian 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt21436587)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UInt21436587 from input register
    def read_input_register_uint21436587_list(self, reference, quantity):
        """
        Read a list of unsigned 21436587-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt21436587)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write UInt78563412 to input register
    def write_input_register_uint78563412(self, reference, value):
        """
        Write a unsigned 78563412-endian 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the unsigned int value to be written
        """
        data = oser.UInt78563412(value=value)
        return self.write_input_register_data(reference, data=data)

    # read UInt78563412 from input register
    def read_input_register_uint78563412(self, reference):
        """
        Read a unsigned 78563412-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            unsigned int: value read from ``reference``
        """
        data = oser.UInt78563412()
        return self.read_input_register_data(reference, data).get()

    # write list of UInt78563412 to input register
    def write_input_register_uint78563412_list(self, reference, values):
        """
        Write a list of unsigned 78563412-endian 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of unsigned int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.UInt78563412)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of UInt78563412 from input register
    def read_input_register_uint78563412_list(self, reference, quantity):
        """
        Read a list of unsigned 78563412-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of unsigned int: unsigned int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.UInt78563412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt21436587 to input register
    def write_input_register_sint21436587(self, reference, value):
        """
        Write a signed 21436587-endian 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt21436587(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SInt21436587 from input register
    def read_input_register_sint21436587(self, reference):
        """
        Read a signed 21436587-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt21436587()
        return self.read_input_register_data(reference, data).get()

    # write list of SInt21436587 to input register
    def write_input_register_sint21436587_list(self, reference, values):
        """
        Write a list of signed 21436587-endian 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt21436587)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SInt21436587 from input register
    def read_input_register_sint21436587_list(self, reference, quantity):
        """
        Read a list of signed 21436587-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt21436587)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write SInt78563412 to input register
    def write_input_register_sint78563412(self, reference, value):
        """
        Write a signed 78563412-endian 64-bit integer to input register memory.

        Args:
            reference: the reference to write to
            value: the signed int value to be written
        """
        data = oser.SInt78563412(value=value)
        return self.write_input_register_data(reference, data=data)

    # read SInt78563412 from input register
    def read_input_register_sint78563412(self, reference):
        """
        Read a signed 78563412-endian 64-bit integer from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            signed int: value read from ``reference``
        """
        data = oser.SInt78563412()
        return self.read_input_register_data(reference, data).get()

    # write list of SInt78563412 to input register
    def write_input_register_sint78563412_list(self, reference, values):
        """
        Write a list of signed 78563412-endian 64-bit integers to input register memory.

        Args:
            reference: the reference to write to
            values: a list of signed int values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.SInt78563412)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of SInt78563412 from input register
    def read_input_register_sint78563412_list(self, reference, quantity):
        """
        Read a list of signed 78563412-endian 64-bit integers from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of signed int: signed int values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.SInt78563412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write BFloat to input register
    def write_input_register_bfloat(self, reference, value):
        """
        Write a float to input register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.BFloat(value=value)
        return self.write_input_register_data(reference, data=data)

    # read BFloat from input register
    def read_input_register_bfloat(self, reference):
        """
        Read a float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.BFloat()
        return self.read_input_register_data(reference, data).get()

    # write list of BFloat to input register
    def write_input_register_bfloat_list(self, reference, values):
        """
        Write a list of floats to input register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.BFloat)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of BFloat from input register
    def read_input_register_bfloat_list(self, reference, quantity):
        """
        Read a list of floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BFloat)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write LFloat to input register
    def write_input_register_lfloat(self, reference, value):
        """
        Write a float to input register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.LFloat(value=value)
        return self.write_input_register_data(reference, data=data)

    # read LFloat from input register
    def read_input_register_lfloat(self, reference):
        """
        Read a float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.LFloat()
        return self.read_input_register_data(reference, data).get()

    # write list of LFloat to input register
    def write_input_register_lfloat_list(self, reference, values):
        """
        Write a list of floats to input register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.LFloat)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of LFloat from input register
    def read_input_register_lfloat_list(self, reference, quantity):
        """
        Read a list of floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LFloat)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Float2143 to input register
    def write_input_register_float2143(self, reference, value):
        """
        Write a 2143-endian float to input register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.Float2143(value=value)
        return self.write_input_register_data(reference, data=data)

    # read Float2143 from input register
    def read_input_register_float2143(self, reference):
        """
        Read a 2143-endian float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float2143()
        return self.read_input_register_data(reference, data).get()

    # write list of Float2143 to input register
    def write_input_register_float2143_list(self, reference, values):
        """
        Write a list of 2143-endian floats to input register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Float2143)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of Float2143 from input register
    def read_input_register_float2143_list(self, reference, quantity):
        """
        Read a list of 2143-endian floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float2143)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Float3412 to input register
    def write_input_register_float3412(self, reference, value):
        """
        Write a 3412-endian float to input register memory.

        Args:
            reference: the reference to write to
            value: the float value to be written
        """
        data = oser.Float3412(value=value)
        return self.write_input_register_data(reference, data=data)

    # read Float3412 from input register
    def read_input_register_float3412(self, reference):
        """
        Read a 3412-endian float from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            float: value read from ``reference``
        """
        data = oser.Float3412()
        return self.read_input_register_data(reference, data).get()

    # write list of Float3412 to input register
    def write_input_register_float3412_list(self, reference, values):
        """
        Write a list of 3412-endian floats to input register memory.

        Args:
            reference: the reference to write to
            values: a list of float values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Float3412)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of Float3412 from input register
    def read_input_register_float3412_list(self, reference, quantity):
        """
        Read a list of 3412-endian floats from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of float: float values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Float3412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write BDouble to input register
    def write_input_register_bdouble(self, reference, value):
        """
        Write a double to input register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.BDouble(value=value)
        return self.write_input_register_data(reference, data=data)

    # read BDouble from input register
    def read_input_register_bdouble(self, reference):
        """
        Read a double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.BDouble()
        return self.read_input_register_data(reference, data).get()

    # write list of BDouble to input register
    def write_input_register_bdouble_list(self, reference, values):
        """
        Write a list of doubles to input register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.BDouble)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of BDouble from input register
    def read_input_register_bdouble_list(self, reference, quantity):
        """
        Read a list of doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BDouble)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write LDouble to input register
    def write_input_register_ldouble(self, reference, value):
        """
        Write a double to input register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.LDouble(value=value)
        return self.write_input_register_data(reference, data=data)

    # read LDouble from input register
    def read_input_register_ldouble(self, reference):
        """
        Read a double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.LDouble()
        return self.read_input_register_data(reference, data).get()

    # write list of LDouble to input register
    def write_input_register_ldouble_list(self, reference, values):
        """
        Write a list of doubles to input register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.LDouble)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of LDouble from input register
    def read_input_register_ldouble_list(self, reference, quantity):
        """
        Read a list of doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.LDouble)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Double21436587 to input register
    def write_input_register_double21436587(self, reference, value):
        """
        Write a 21436587-endian double to input register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.Double21436587(value=value)
        return self.write_input_register_data(reference, data=data)

    # read Double21436587 from input register
    def read_input_register_double21436587(self, reference):
        """
        Read a 21436587-endian double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double21436587()
        return self.read_input_register_data(reference, data).get()

    # write list of Double21436587 to input register
    def write_input_register_double21436587_list(self, reference, values):
        """
        Write a list of 21436587-endian doubles to input register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Double21436587)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of Double21436587 from input register
    def read_input_register_double21436587_list(self, reference, quantity):
        """
        Read a list of 21436587-endian doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double21436587)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Double78563412 to input register
    def write_input_register_double78563412(self, reference, value):
        """
        Write a 78563412-endian double to input register memory.

        Args:
            reference: the reference to write to
            value: the double value to be written
        """
        data = oser.Double78563412(value=value)
        return self.write_input_register_data(reference, data=data)

    # read Double78563412 from input register
    def read_input_register_double78563412(self, reference):
        """
        Read a 78563412-endian double from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            double: value read from ``reference``
        """
        data = oser.Double78563412()
        return self.read_input_register_data(reference, data).get()

    # write list of Double78563412 to input register
    def write_input_register_double78563412_list(self, reference, values):
        """
        Write a list of 78563412-endian doubles to input register memory.

        Args:
            reference: the reference to write to
            values: a list of double values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Double78563412)
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of Double78563412 from input register
    def read_input_register_double78563412_list(self, reference, quantity):
        """
        Read a list of 78563412-endian doubles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of double: double values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Double78563412)
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write String to input register
    def write_input_register_string(self, reference, value, length):
        """
        Write a string to input register memory.

        Args:
            reference: the reference to write to
            value: the data value to be written
            length: the length of the value
        """
        data = oser.String(value=value, length=length)
        return self.write_input_register_data(reference, data=data)

    # read String from input register
    def read_input_register_string(self, reference, length):
        """
        Read a string from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            data: value read from ``reference``
        """
        data = oser.String(length=length)
        return self.read_input_register_data(reference, data).get()

    # write list of String to input register
    def write_input_register_string_list(self, reference, values, length):
        """
        Write a list of strings to input register memory.

        Args:
            reference: the reference to write to
            values: a list of data values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.String, kwargs={'length': length})
        data[:] = values
        return self.write_input_register_data(reference, data)

    # read list of String from input register
    def read_input_register_string_list(self, reference, quantity, length):
        """
        Read a list of strings from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of data: data values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.String, kwargs={'length': length})
        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data))

    # write Flag to input register
    def write_input_register_flag(self, reference, value):
        """
        Write a bit to input register memory.

        Args:
            reference: the reference to write to
            value: the bool value to be written
        """
        data = oser.Flag(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_input_register_data(reference, data=data)

    # read Flag from input register
    def read_input_register_flag(self, reference):
        """
        Read a bit from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            bool: value read from ``reference``
        """
        data = oser.Flag()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_input_register_data(reference, data).data.get()

    # write list of Flag to input register
    def write_input_register_flag_list(self, reference, values):
        """
        Write a list of bits to input register memory.

        Args:
            reference: the reference to write to
            values: a list of bool values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Flag)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_input_register_data(reference, data)

    # read list of Flag from input register
    def read_input_register_flag_list(self, reference, quantity):
        """
        Read a list of bits from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of bool: bool values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Flag)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write Nibble to input register
    def write_input_register_nibble(self, reference, value):
        """
        Write a nibble to input register memory.

        Args:
            reference: the reference to write to
            value: the nibble value to be written
        """
        data = oser.Nibble(value=value)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_input_register_data(reference, data=data)

    # read Nibble from input register
    def read_input_register_nibble(self, reference):
        """
        Read a nibble from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            nibble: value read from ``reference``
        """
        data = oser.Nibble()
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_input_register_data(reference, data).data.get()

    # write list of Nibble to input register
    def write_input_register_nibble_list(self, reference, values):
        """
        Write a list of nibbles to input register memory.

        Args:
            reference: the reference to write to
            values: a list of nibble values to be written
        """
        data = oser.Array(length=len(values), prototype=oser.Nibble)
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_input_register_data(reference, data)

    # read list of Nibble from input register
    def read_input_register_nibble_list(self, reference, quantity):
        """
        Read a list of nibbles from input register memory.

        Args:
            reference: the reference to read from

        Returns:
            list of nibble: nibble values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.Nibble)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))

    # write BitField to input register
    def write_input_register_bitfield(self, reference, value, length):
        """
        Write a bitfield to input register memory.

        Args:
            reference: the reference to write to
            value: the bitfield value to be written
            length: the length of the value
        """
        data = oser.BitField(value=value, length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_input_register_data(reference, data=data)

    # read BitField from input register
    def read_input_register_bitfield(self, reference, length):
        """
        Read a bitfield from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            bitfield: value read from ``reference``
        """
        data = oser.BitField(length=length)
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.read_input_register_data(reference, data).data.get()

    # write list of BitField to input register
    def write_input_register_bitfield_list(self, reference, values, length):
        """
        Write a list of bitfields to input register memory.

        Args:
            reference: the reference to write to
            values: a list of bitfield values to be written
            length: the length of the value
        """
        data = oser.Array(length=len(values), prototype=oser.BitField, kwargs={'length': length})
        data[:] = values
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        return self.write_input_register_data(reference, data)

    # read list of BitField from input register
    def read_input_register_bitfield_list(self, reference, quantity, length):
        """
        Read a list of bitfields from input register memory.

        Args:
            reference: the reference to read from
            length: the length of the value

        Returns:
            list of bitfield: bitfield values read from ``reference``
        """
        data = oser.Array(length=quantity, prototype=oser.BitField, kwargs={'length': length})
        struct = oser.BitStruct()
        struct.data = data
        data = struct

        self.read_input_register_data(reference, data)
        return list(map(lambda d: d.get(), data.data))


class ModbusSerialRTUDevice(ModbusBaseDevice):

    def __init__(self, com_port, baud=9600, device_address=1, shared_memory=False,
                 request_parser=SerialRTUModbusRequest, response_parser=SerialRTUModbusResponse,
                 debuglevel=None):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            device_address (int): the device address to be used (Slave ID)
            shared_memory (bool): if set to ``True`` a single shared memory is used
            request_parser: the request parser
            response_parser: the response parser
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        server = ModbusSerialRTUServer(com_port=com_port, baud=baud, device_address=device_address,
                                       request_parser=request_parser, response_parser=response_parser,
                                       callback=self.handle_request)
        memory = ModbusMemory(shared_memory=shared_memory)
        super(ModbusSerialRTUDevice, self).__init__(server, memory, debuglevel=debuglevel)

    # Special Methods
    # ===============

    def set_exception_status(self, status):
        """
        Set exception status to ``status``.

        Args:
            status (int): the new exception status
        """
        self.exception_status.set(status)

    def get_exception_status(self):
        """
        Return the current exception status.

        Returns:
            int: the current exception status
        """
        return self.exception_status.get()

    def set_busy_status(self, busy=True):
        """
        Set busy flag which will be reported in Modbus function 11.

        Args:
            busy (bool): if set to ``True`` the busy state is set to busy else not busy.
        """
        if busy not in [True, False]:
            raise ValueError("busy must be of type bool")
        self.busy_status = busy

    def get_busy_status(self):
        """
        Return the current busy status.

        Returns:
            bool: ``True`` if the device is busy else ``False``
        """
        return self.busy_status

    def set_bus_communication_error_count(self, count):
        """
        Set the bus communication error count to ``count``.

        Args:
            count (int): the bus communication error count
        """
        self.server.bus_communication_error_count = count

    def set_bus_exception_error_count(self, count):
        """
        Set the bus exception error count to ``count``.

        Args:
            count (int): the bus exception error count
        """
        self.server.bus_exception_error_count = count

    def set_bus_message_count(self, count):
        """
        Set the bus message count to ``count``.

        Args:
            count (int): the bus message count
        """
        self.server.bus_message_count = count

    def set_busy_count(self, count):
        """
        Set the slave busy count to ``count``.

        Args:
            count (int): the slave busy count
        """
        self.server.slave_busy_count = count

    def get_busy_count(self):
        """
        Return the current busy count.

        Returns:
            int: the busy count
        """
        return self.server.slave_busy_count

    def set_bus_character_overrun_count(self, count):
        """
        Set the bus character overrun count to ``count``.

        Args:
            count (int): the bus character overrun count
        """
        self.server.bus_character_overrun_count = count

    def clear_counters_and_diagnostic_register(self):
        """
        Clear counters and diagnostic register
        """
        self.server.clear_counters()
        self.server.clear_diagnostics()
        self.negative_acknowledge_count = 0

    def get_bus_character_overrun_count(self):
        """
        Get the bus character overrun count.

        Returns:
            int: the bus character overrun count
        """
        return self.server.bus_character_overrun_count

    def clear_overrun_counter_and_flag(self):
        """
        Clear overrun counter and reset the error flag.
        """
        self.server.bus_character_overrun_count = 0

    def get_diagnostic_register(self):
        """
        Return the diagnostic register

        Returns:
            bytes: the content of the diagnostic register
        """
        return self.server.diagnostics

    def set_diagnostic_register(self, content):
        """
        Set the diagnostic register to ``content``

        Args:
            content (bytes): the new content for the diagnostic register
        """
        if not isinstance(content, bytes):
            raise ValueError("content must be of type bytes")
        self.server.diagnostics = content

    def set_listen_only_mode(self, listen_only_mode_enabled=True):
        """
        Set listen only mode.

        Args:
            listen_only_enabled (bool): ``True`` enables the listen only mode. ``False`` disables it.
        """
        self.server.set_listen_only_mode(listen_only_mode_enabled)
        if listen_only_mode_enabled:
            self.add_set_listen_only_mode_event()

    def set_slave_message_count(self, count):
        """
        Set the slave message count to ``count``.

        Args:
            count (int): the slave message count
        """
        self.server.slave_message_count = count

    def set_slave_no_response_count(self, count):
        """
        Set the slave no response count to ``count``.

        Args:
            count (int): the slave no response count
        """
        self.server.slave_no_response_count = count

    def set_slave_nak_count(self, count):
        """
        Set the slave nak count to ``count``.

        Args:
            count (int): the slave nak count
        """
        self.negative_acknowledge_count = count

    # Eventlog
    # ========

    def add_event(self, event):
        """
        Add an event to event log.

        Args:
            event (int): the event to be added.
        """
        self.events.append(event)
        self.events = self.events[-1024:]
        self.event_counter += 1

    def clear_events(self):
        """
        Clear event log.
        """
        self.events = []
        self.event_counter = 0

    def add_receive_event(self, is_communication_error=False, is_character_overrun=False, is_broadcast=False,
                          is_listen_only_mode=False):
        """
        Add a receive event.

        Args:
            is_communication_error (bool): set to ``True`` if a communication error occurred
            is_character_overrun (bool): set to ``True`` if a character overrun occurred
            is_broadcast (bool): set to ``True`` if a broadcast message was received
            is_listen_only_mode (bool): set to ``True`` if the server is currently in listen only mode

        Returns:

        """
        event = 0x80

        if is_communication_error:
            event |= 1 << 1

        if is_character_overrun:
            event |= 1 << 4

        if is_listen_only_mode:
            event |= 1 << 5

        if is_broadcast:
            event |= 1 << 6

        self.add_event(event)

    def add_send_event(self, read_exception_sent=False, slave_abort_exeception_sent=False,
                       slave_busy_exception_sent=False, slave_program_nak_exception_sent=False,
                       write_timeout_occurred=False):
        """
        Add a send event.
        The current listen only mode is determined automatically.

        Args:
            read_exception_sent (bool): set to ``True`` if a read exception sent occurred (exception codes 1-3)
            slave_abort_exeception_sent (bool): set to ``True`` if a slave abort exception sent occurred
                (execption code 4)
            slave_busy_exception_sent (bool): set to ``True`` if a slave busy exception sent occurred
                (exception codes 5-6)
            slave_program_nak_exception_sent (bool): set to ``True`` if a slave program NAK exception sent occurred
                (exception code 7)
            write_timeout_occurred (bool): set to ``True`` if a write timeout occurred
        """

        event = 0x40

        if read_exception_sent:
            event |= 1 << 0

        if slave_abort_exeception_sent:
            event |= 1 << 1

        if slave_busy_exception_sent:
            event |= 1 << 2

        if slave_program_nak_exception_sent:
            event |= 1 << 3

        if write_timeout_occurred:
            event |= 1 << 4

        if self.server.listen_only:
            event |= 1 << 5

        self.add_event(event)

    def add_set_listen_only_mode_event(self):
        """
        Add a listen only mode event.
        """
        self.add_event(0x04)

    def add_communication_restart_event(self):
        """
        Add a communication restart event.
        """
        self.add_event(0x00)

    # Handlers
    # ========

    def create_exception_response(self, request_pdu, exception_code):
        """
        Request handler to response with an exception.

        Args:
            request_pdu: the request pdu
            exception_code: the exception code to be sent

        Returns:
            the response to be sent to the client
        """
        response = self.server.response_parser()
        response.pdu.function_code.set(request_pdu.function_code.get() | 0x80)
        response.pdu.data.exception_code.set(exception_code)
        exception_code_value = response.pdu.data.exception_code.get_value()

        if 1 <= exception_code_value <= 3:  # Read Exception Sent (Exception Codes 1-3)
            self.add_send_event(read_exception_sent=True)
        elif exception_code_value == 4:  # Slave Abort Exception Sent (Exception Code 4)
            self.add_send_event(slave_abort_exeception_sent=True)
        elif 5 <= exception_code_value <= 6:  # Slave Busy Exception Sent (Exception Codes 5-6)  # pragma: no cover
            self.add_send_event(slave_busy_exception_sent=True)
        elif exception_code_value == 7:  # Slave Program NAK Exception Sent (Exception Code 7)  # pragma: no branch
            self.add_send_event(slave_program_nak_exception_sent=True)
            self.negative_acknowledge_count += 1

        return ModbusBaseDevice.create_exception_response(self, request_pdu, exception_code)

    def handle_function_code_7(self, pdu):
        """
        Request handler for function code 7 ``Read Exception Status``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        response = self.server.response_parser()
        response.pdu.function_code.set(7)
        response.pdu.data.output_data.set(self.exception_status.get())
        return response

    def handle_function_code_11(self, pdu):
        """
        Request handler for function code 11 ``Get Comm Event Counter (Serial Line only)``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        response = self.server.response_parser()
        response.pdu.function_code.set(11)
        busy_status = 0 if self.busy_status is False else 0xffff

        response.pdu.data.status.set(busy_status)
        response.pdu.data.event_count.set(self.event_counter)
        return response

    def handle_function_code_12(self, pdu):
        """
        Request handler for function code 12 ``Get Comm Event Log (Serial Line only)``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        response = self.server.response_parser()
        response.pdu.function_code.set(12)
        response.pdu.data.byte_count.set(min(len(self.events), 64) + 6)
        busy_status = 0 if self.busy_status is False else 0xffff
        response.pdu.data.status.set(busy_status)
        response.pdu.data.event_count.set(self.event_counter)
        response.pdu.data.message_count.set(self.server.bus_message_count)
        response.pdu.data.events[:] = self.events[:-64-1:-1]

        return response

    def handle_function_code_17(self, pdu):
        """
        Request handler for function code 17 ``Report Slave ID (Serial Line only)``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        response = self.server.response_parser()
        response.pdu.function_code.set(17)
        # response with default values
        return response

    def send_function_code_8_echo_response(self, pdu):
        """
        Generate an echo response using the request ``pdu``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        response.pdu.data.data.set(pdu.data.data.get())
        return response

    def handle_function_code_8_sub_function_0(self, pdu):
        """
        Request handler for function code 8.0 ``Diagnostics Send Echo Response``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        return self.send_function_code_8_echo_response(pdu)

    def handle_function_code_8_sub_function_1(self, pdu):
        """
        Request handler for function code 8.1 ``Diagnostics Restart Communications Option``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() not in [b"\x00\x00", b"\xFF\x00"]:
            return self.create_exception_response(pdu, "illegal_data_value")

        ret = self.send_function_code_8_echo_response(pdu)
        self.set_listen_only_mode(False)
        self.server.clear_counters()
        if pdu.data.data.get() == b"\xFF\x00":
            self.clear_events()
        self.add_communication_restart_event()
        return ret

    def handle_function_code_8_sub_function_2(self, pdu):
        """
        Request handler for function code 8.2 ``Diagnostics Return diagnostic register``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        response.pdu.data.data.set(self.server.diagnostics)
        return response

    def handle_function_code_8_sub_function_3(self, pdu):
        """
        Request handler for function code 8.3 ``Diagnostics Change ASCII Input Delimiter``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        data = pdu.data.data.get()
        if data[1:2] != b'\x00':
            return self.create_exception_response(pdu, 'illegal_data_value')
        new_delimiter = data[0:1]

        self.server.change_ascii_delimiter(new_delimiter)
        return self.send_function_code_8_echo_response(pdu)

    def handle_function_code_8_sub_function_4(self, pdu):
        """
        Request handler for function code 8.4 ``Diagnostics Force Listen Only Mode``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        ret = self.send_function_code_8_echo_response(pdu)
        self.set_listen_only_mode(True)
        return ret

    def handle_function_code_8_sub_function_10(self, pdu):
        """
        Request handler for function code 8.10 ``Diagnostics Clear counters and diagnostic register``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        self.server.clear_counters()
        self.server.clear_diagnostics()
        return self.send_function_code_8_echo_response(pdu)

    def handle_function_code_8_sub_function_11(self, pdu):
        """
        Request handler for function code 8.11 ``Diagnostics Bus message count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.bus_message_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_12(self, pdu):
        """
        Request handler for function code 8.12 ``Diagnostics Bus communication error count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.bus_communication_error_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_13(self, pdu):
        """
        Request handler for function code 8.13 ``Diagnostics Bus exception error count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.bus_exception_error_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_14(self, pdu):
        """
        Request handler for function code 8.14 ``Diagnostics Return slave message count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.slave_message_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_15(self, pdu):
        """
        Request handler for function code 8.15 ``Diagnostics Slave No Response Count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.slave_no_response_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_16(self, pdu):
        """
        Request handler for function code 8.16 ``Diagnostics Slave NAK Count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.negative_acknowledge_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_17(self, pdu):
        """
        Request handler for function code 8.17 ``Diagnostics Slave Busy Count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.slave_busy_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_18(self, pdu):
        """
        Request handler for function code 8.18 ``Diagnostics Slave Character Overrun Count``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        response = self.server.response_parser()
        response.pdu.function_code.set(8)
        response.pdu.data.sub_function.set(pdu.data.sub_function.get())
        encoder = oser.UBInt16(self.server.bus_character_overrun_count)
        response.pdu.data.data.set(encoder.encode())
        return response

    def handle_function_code_8_sub_function_20(self, pdu):
        """
        Request handler for function code 8.20 ``Diagnostics``.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        if pdu.data.data.get() != b"\x00\x00":
            return self.create_exception_response(pdu, "illegal_data_value")

        self.server.bus_character_overrun_count = 0
        return self.send_function_code_8_echo_response(pdu)

    def handle_function_code_8(self, pdu):
        """
        Request handler for function code 8.x Diagnostics (Serial Line only).

        Calls a handler by sub function code like ``handle_function_code_8_sub_function_<sub function code>``.
        To add more sub functions more handlers have to be added.

        Args:
            pdu: the request pdu

        Returns:
            the response to be sent to the client
        """
        sub_function = pdu.data.sub_function.get_value()
        try:
            return getattr(self, 'handle_function_code_8_sub_function_{}'.format(sub_function))(pdu)
        except AttributeError:
            return self.create_exception_response(pdu, "illegal_function")


class ModbusSerialASCIIDevice(ModbusSerialRTUDevice):

    def __init__(self, com_port, baud=9600, device_address=1, shared_memory=False,
                 request_parser=SerialASCIIModbusRequest, response_parser=SerialASCIIModbusResponse,
                 debuglevel=None):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            device_address (int): the device address to be used (Slave ID)
            shared_memory (bool): if set to ``True`` a single shared memory is used
            request_parser: the request parser
            response_parser: the response parser
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        server = ModbusSerialASCIIServer(com_port=com_port, baud=baud, device_address=device_address,
                                         request_parser=request_parser, response_parser=response_parser,
                                         callback=self.handle_request)
        memory = ModbusMemory(shared_memory=shared_memory)
        ModbusBaseDevice.__init__(self, server, memory, debuglevel=debuglevel)

    def change_ascii_delimiter(self, delimiter):
        """
        Change ASCII delimiter.

        Args:
            new_lf (bytes): the new ASCII delimiter
        """
        self.server.change_ascii_delimiter(delimiter)


class ModbusTCPDevice(ModbusBaseDevice):

    def __init__(self, bind_address, port=502, shared_memory=False,
                 request_parser=TCPModbusRequest, response_parser=TCPModbusResponse,
                 debuglevel=None):
        """
        Args:
            bind_address (str): the ip-address to listen on
            port (int): the port to listen on
            shared_memory (bool): if set to ``True`` a single shared memory is used
            request_parser: the request parser
            response_parser: the response parser
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        server = ModbusTCPServer(bind_address=bind_address, port=port, request_parser=request_parser,
                                 response_parser=response_parser, callback=self.handle_request)
        memory = ModbusMemory(shared_memory=shared_memory)
        super(ModbusTCPDevice, self).__init__(server, memory, debuglevel=debuglevel)


class ModbusUDPDevice(ModbusBaseDevice):

    def __init__(self, bind_address, port=502,
                 shared_memory=False, request_parser=TCPModbusRequest, response_parser=TCPModbusResponse,
                 debuglevel=None):
        """
        Args:
            bind_address (str): the ip-address to listen on
            port (int): the port to listen on
            shared_memory (bool): if set to ``True`` a single shared memory is used
            request_parser: the request parser
            response_parser: the response parser
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        server = ModbusUDPServer(bind_address=bind_address, port=port, request_parser=request_parser,
                                 response_parser=response_parser, callback=self.handle_request)
        memory = ModbusMemory(shared_memory=shared_memory)
        super(ModbusUDPDevice, self).__init__(server, memory, debuglevel=debuglevel)


class ModbusTLSDevice(ModbusBaseDevice):

    def __init__(self, bind_address, port=502,
                 shared_memory=False,
                 certfile=None, keyfile=None, password=None,
                 request_parser=TCPModbusRequest, response_parser=TCPModbusResponse,
                 debuglevel=None):
        """
        Args:
            bind_address (str): the ip-address to listen on
            port (int): the port to listen on
            shared_memory (bool): if set to ``True`` a single shared memory is used
            callback: the callback to handle requests
            certfile (str): path to TLS certificate
            keyfile (str): path to TLS keyfile
            password (str): TLS password
            request_parser: the request parser
            response_parser: the response parser
            debuglevel (int or None): the debuglevel (``None`` disables debug, ``1`` prints requests and responses and
                ``2`` prints introspection)
        """
        server = ModbusTLSServer(bind_address=bind_address, port=port,
                                 certfile=certfile, keyfile=keyfile, password=password,
                                 request_parser=request_parser,
                                 response_parser=response_parser, callback=self.handle_request)
        memory = ModbusMemory(shared_memory=shared_memory)
        super(ModbusTLSDevice, self).__init__(server, memory, debuglevel=debuglevel)
