# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import struct
import threading
import time
import traceback
import warnings

import serial
import socket
import ssl
import sys
from binascii import hexlify, unhexlify

from htf.communication.modbus.protocol import SerialRTUModbusRequest, SerialRTUModbusResponse, \
    SerialASCIIModbusResponse, SerialASCIIModbusRequest


def _create_ssl_client_context():  # pragma: no cover
    # we only support Python 2.7, 3.6 and 3.7+
    if sys.version_info[:2] == (2, 7):
        return ssl.SSLContext(ssl.PROTOCOL_TLS)
    if sys.version_info[:2] >= (3, 6):
        return ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    raise Exception("Python version is not supported.")


def _create_ssl_server_context():  # pragma: no cover
    # we only support Python 2.7, 3.6 and 3.7+
    if sys.version_info[:2] == (2, 7):
        return ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    if sys.version_info[:2] >= (3, 6):
        return ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    raise Exception("Python version is not supported.")


class ModbusSerialRTUClientInterface(object):

    def __init__(self, com_port=None, baud=9600, bytesize=serial.EIGHTBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False, timeout=1.0, response_parser=SerialRTUModbusResponse):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            timeout (float): the timeout in seconds
            response_parser: the parser to parse responses
        """
        self.baud = baud
        if parity == serial.PARITY_NONE and stopbits != serial.STOPBITS_TWO:
            warnings.warn(Warning('Two stop bits are required when using no parity!'))
        self.serial = serial.serial_for_url(com_port, baud, bytesize=bytesize, parity=parity, stopbits=stopbits,
                                            rtscts=rtscts, timeout=timeout, write_timeout=timeout)
        self.last_write = time.time()
        self._response_parser = response_parser

    def read(self):
        """
        Read Modbus serial RTU responses from server.

        Returns:
            tuple: buffer, ``None``
        """
        buffer = b""
        decoder = self._response_parser()
        read = 0

        while read <= 256:
            # read byte wise and try to decode the request
            byte = self.serial.read(1)

            if not byte:
                return buffer, None

            read += len(byte)

            buffer += byte

            try:
                decoder.decode(buffer)
                return buffer, None
            except Exception:
                pass

        return buffer, None

    def write(self, adu):
        """
        Write ``adu``.

        Args:
            adu: serializable RTU adu
        """
        self.serial.flushInput()
        try:
            current = time.time()
            time_diff = current - (self.last_write + (3.5 * 11) / self.baud)  # wait at least 3.5 char times
            if time_diff < .0:
                time.sleep(min(0.0, -time_diff))
            data = adu.encode()
            self.serial.write(data)
            self.last_write = current
        except Exception:  # pragma: no cover
            traceback.print_exc()

    def close(self):
        """
        Close serial port.
        """
        self.serial.close()


class ModbusSerialRTUServerInterface(ModbusSerialRTUClientInterface):
    def __init__(self, com_port=None, baud=9600, bytesize=serial.EIGHTBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False, timeout=1.0, response_parser=SerialRTUModbusRequest):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            timeout (float): the timeout in seconds
            response_parser: the parser to parse responses
        """
        super(ModbusSerialRTUServerInterface, self).__init__(com_port=com_port, baud=baud, bytesize=bytesize,
                                                             parity=parity, stopbits=stopbits, rtscts=rtscts,
                                                             timeout=timeout, response_parser=response_parser)

    def read(self):
        """
        Read Modbus serial RTU requests from client.

        Returns:
            tuple: buffer, ``None``
        """
        return super(ModbusSerialRTUServerInterface, self).read()


class ModbusSerialASCIIClientInterface(object):
    def __init__(self, com_port=None, baud=9600, bytesize=serial.SEVENBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False, timeout=1.0, response_parser=SerialASCIIModbusResponse):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            timeout (float): the timeout in seconds
            response_parser: the parser to parse responses
        """
        self.ascii_delimiter = b'\x0A'
        self.baud = baud
        if parity == serial.PARITY_NONE and stopbits != serial.STOPBITS_TWO:
            warnings.warn(Warning('Two stop bits are required when using no parity!'))
        self.serial = serial.serial_for_url(com_port, baud, bytesize=bytesize, parity=parity, stopbits=stopbits,
                                            rtscts=rtscts, timeout=timeout, write_timeout=timeout)
        self._response_parser = response_parser

    def change_ascii_delimiter(self, delimiter):
        """
        Change LF character to ``delimiter``.

        Args:
            delimiter (bytes): bytes of length 1 containing the new ASCII delimiter
        """
        assert isinstance(delimiter, (bytes, bytearray)) and len(delimiter) == 1
        self.ascii_delimiter = bytes(delimiter)

    def _bytes_to_ascii(self, encoded):  # pragma: no cover
        ret = b'\x3A'
        for byte in hexlify(encoded):
            if not isinstance(byte, (type(b""), type(""))):  # Python 3.x
                ret += struct.Struct('>B').pack(byte)
            else:  # Python 2.7
                ret += byte
        return ret + b'\x0D' + self.ascii_delimiter

    def _ascii_to_bytes(self, ascii):
        return unhexlify(ascii[1:-2])

    def read(self):
        """
        Read Modbus serial ASCII responses from server.
        Read until line feed character.

        Returns:
            tuple: buffer, ``None``
        """

        buffer = b""

        decoder = self._response_parser()
        read = 0
        while read <= 517:
            byte = self.serial.read(1)
            if not byte:
                return b"", None

            read += len(byte)

            if byte == b'\x3A' or buffer.startswith(b'\x3A'):  # pragma: no cover
                # frame start received
                buffer += byte
            else:  # pragma: no cover
                continue

            if byte == self.ascii_delimiter:
                # line feed received
                try:
                    bytes_buffer = self._ascii_to_bytes(buffer)
                    decoder.decode(bytes_buffer)
                    return bytes_buffer, None
                except Exception:
                    pass
                return self._ascii_to_bytes(buffer), None

        return self._ascii_to_bytes(buffer), None

    def write(self, adu):
        """
        Write ``adu``.

        Args:
            adu: serializable RTU adu
        """
        self.serial.flushInput()
        data = self._bytes_to_ascii(adu.encode())
        self.serial.write(data)

    def close(self):
        """
        Close serial port.
        """
        self.serial.close()


class ModbusSerialASCIIServerInterface(ModbusSerialASCIIClientInterface):
    def __init__(self, com_port=None, baud=9600, bytesize=serial.SEVENBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False, timeout=1.0, response_parser=SerialASCIIModbusRequest):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            timeout (float): the timeout in seconds
            response_parser: the parser to parse responses
        """
        super(ModbusSerialASCIIServerInterface, self).__init__(com_port=com_port, baud=baud, bytesize=bytesize,
                                                               parity=parity, stopbits=stopbits, rtscts=rtscts,
                                                               timeout=timeout, response_parser=response_parser)

    def read(self):
        """
        Read serial ASCII Modbus requests from client.

        Returns:
            tuple: buffer, ``None``
        """
        return super(ModbusSerialASCIIServerInterface, self).read()


class ModbusTCPClientInterface(object):
    _socket_type = socket.SOCK_STREAM

    def __init__(self, address, port=502, timeout=1.0, connect=True):
        """
        Args:
            address (str): the ip-address to connect to
            port (int): the port
            timeout (float): the timeout in seconds
            connect (bool): if set to ``True`` the socket connects automatically
        """
        self.socket = None
        self.address = address
        self.port = port
        self.timeout = timeout
        self.socket_lock = threading.Lock()
        if connect:
            self.open()

    def read(self):
        """
        Read a TCP request from TCP server.

        Returns:
            tuple: buffer, ``None``
        """
        try:
            return self.socket.recv(300), None
        except Exception as e:
            return None, None

    def write(self, adu):
        """
        Write ``adu``.

        Args:
            adu: serializable TCP adu
        """
        self.socket.sendall(adu.encode())

    def open(self):
        """
        Open a socket.
        """
        try:
            with self.socket_lock:
                self.socket = socket.socket(type=self._socket_type)
                self.timeout = self.timeout
                self.socket.setblocking(1)
                self.socket.settimeout(self.timeout)

                self.socket.connect((self.address, self.port))
        except Exception:
            self.close()
            raise

    def close(self):
        """
        Close TCP connection.
        """
        with self.socket_lock:
            self.socket.close()

    def reconnect(self):
        """
        Close socket and create a new one.
        """
        self.close()
        self.open()


class ModbusTLSClientInterface(ModbusTCPClientInterface):
    def __init__(self, address, port=502, timeout=1.0, verify=True, connect=True,
                 cafile=None, capath=None, cadata=None):
        """
        Args:
            address (str): the ip-address to connect to
            port (int): the port
            timeout (float): the timeout in seconds
            verify (bool): set to ``False`` to disable TLS certificate verification
            connect (bool): if set to ``True`` the socket connects automatically
            cafile (str): path to a file of concatenated CA certificates in PEM format
            capath (str): path to a directory containing several CA certificates in PEM format
            cadata (str): either an ASCII string of one or more PEM-encoded certificates or a bytes-like
                object of DER-encoded certificates
        """

        self.socket = None
        self.address = address
        self.port = port
        self.timeout = timeout
        self.cafile = cafile
        self.capath = capath
        self.cadata = cadata
        self.verify = verify
        self.socket_lock = threading.Lock()

        if connect:
            self.open()

    def open(self):
        """
        Open a socket.
        """

        try:
            with self.socket_lock:
                context = _create_ssl_client_context()
                context.load_verify_locations(cafile=self.cafile, capath=self.capath, cadata=self.cadata)

                if self.verify is False:
                    # insecure mode
                    context.check_hostname = False
                    context.verify_mode = ssl.CERT_NONE
                else:
                    # Python 2.7 does not set these values by default
                    context.verify_mode = ssl.CERT_REQUIRED
                    context.check_hostname = True

                sock = socket.socket(type=self._socket_type)
                self.timeout = self.timeout
                sock.settimeout(self.timeout)
                self.socket = context.wrap_socket(sock, server_hostname=self.address)

                self.socket.connect((self.address, self.port))
        except Exception:
            self.close()
            raise


class ModbusUDPClientInterface(ModbusTCPClientInterface):
    _socket_type = socket.SOCK_DGRAM

    def __init__(self, address, port=502, timeout=1.0):
        """
        Args:
            address (str): the ip-address to connect to
            port (int): the port
            timeout (float): the timeout in seconds
        """
        self.socket = socket.socket(type=self._socket_type)
        self.socket.setblocking(1)
        self.timeout = timeout
        self.socket_lock = threading.Lock()
        self.socket.settimeout(self.timeout)
        self.address = address
        self.port = port

    def read(self):
        """
        Read a TLS request from TLS server.

        Returns:
            tuple: buffer, ``None``
        """
        try:
            return self.socket.recvfrom(300)
        except Exception as e:
            return None, None

    def write(self, adu):
        """
        Write ``adu``.

        Args:
            adu: serializable TCP adu
        """
        self.socket.sendto(adu.encode(), (self.address, self.port))


class ModbusTCPServerInterface(object):
    _socket_type = socket.SOCK_STREAM

    def __init__(self, address, port=502, timeout=1.0):
        """
        Args:
            address (str): the ip-address to listen on
            port (int): the port to listen on
            timeout (float): the timeout in seconds
        """
        self.socket = socket.socket(type=self._socket_type)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.timeout = timeout
        self.socket.settimeout(self.timeout)
        self.socket.bind((address, port))
        self.socket.listen(100)

    def accept(self):
        """
        Accept a new connection.

        Returns:
            tuple: connection, address
        """
        conn, address = self.socket.accept()
        return conn, address

    def close(self):
        """
        Close TCP socket.
        """
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self.socket.close()

    def close_connection(self, connection):
        """
        Close TCP connection.
        """
        connection.close()

    def read(self, connection):
        """
        Read a TCP request from TLS client.

        Returns:
            tuple: buffer, ``None``
        """
        connection.settimeout(self.timeout)
        return connection.recv(300)

    def write(self, connection, adu):
        """
        Write ``adu``.

        Args:
            adu: serializable TCP adu
        """
        connection.sendall(adu.encode())


class ModbusTLSServerInterface(ModbusTCPServerInterface):
    def __init__(self, address, port=502, timeout=1.0,
                 certfile=None, keyfile=None, password=None):
        """
        Args:
            address (str): the ip-address to listen on
            port (int): the port to listen on
            timeout (float): the timeout in seconds
            certfile (str): path to TLS certificate
            keyfile (str): path to TLS keyfile
            password (str): TLS password
        """
        context = _create_ssl_server_context()
        context.load_cert_chain(certfile=certfile, keyfile=keyfile, password=password)

        sock = socket.socket(type=self._socket_type)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.timeout = timeout
        sock.settimeout(self.timeout)
        sock.bind((address, port))
        sock.listen(100)

        self.socket = context.wrap_socket(sock, server_side=True)

    def accept(self):
        """
        Accept a new connection.

        Returns:
            tuple: connection, address
        """
        conn, address = self.socket.accept()
        return conn, address

    def close(self):
        """
        Close TLS socket.
        """
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self.socket.close()


class ModbusUDPServerInterface(object):
    _socket_type = socket.SOCK_DGRAM

    def __init__(self, address, port=502, timeout=None):
        """
        Args:
            address (str): the ip-address to listen on
            port (int): the port to listen on
            timeout=None (float): the timeout in seconds
        """
        self.socket = socket.socket(type=self._socket_type)
        self.socket.setblocking(0)
        if timeout is not None:  # pragma: no cover
            self.socket.settimeout(timeout)
        self.socket.bind((address, port))

    def close(self):
        """
        Close UDP socket.
        """
        self.socket.close()

    def read(self, bufsize=300):
        """
        Read a request from an UDP client.

        Args:
            bufsize (int): buffer size

        Returns:
            bytes: the received request
        """
        return self.socket.recvfrom(bufsize)

    def write(self, address, adu):
        """
        Write ``adu``.

        Args:
            address: the target ip-address
            adu: serializable TCP adu
        """
        self.socket.sendto(adu.encode(), address)
