# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals
from threading import Lock
from functools import wraps

MEMORY_SIZE = 65536
REGISTER_BYTE_SIZE = 2
BYTE_SIZE = 8


def synchronized(func):
    @wraps(func)
    def with_lock(self, *args, **kwargs):
        with self._lock:
            return func(self, *args, **kwargs)

    return with_lock


class MemoryBase(object):  # inherit from object for Python 2.7
    """
    MemoryBase is the Modbus memory base-class.
    """

    def __init__(self, memory=None):
        """
        Args:
             memory (bytearray or None): the memory to address or ``None`` for new bytearray
        """
        self._mem = memory or bytearray(MEMORY_SIZE * REGISTER_BYTE_SIZE)
        self._lock = Lock()

    @synchronized
    def __getitem__(self, item):
        return self._mem[item]

    @synchronized
    def __setitem__(self, key, value):
        self._mem[key] = value

    @synchronized
    def __delitem__(self, key):  # pragma: no cover
        self._mem[key] = b'\x00'


class BitArray(MemoryBase):
    """
    BitArray is a bit-aligned Modbus memory for coils and discrete inputs.
    """

    def __init__(self, memory=None):
        """
        Args:
             memory (bytearray or None): the memory to address or ``None`` for new bytearray
        """
        memory = memory or bytearray(MEMORY_SIZE // BYTE_SIZE)
        super(BitArray, self).__init__(memory)

    @synchronized
    def __getitem__(self, item):
        """
        Get the value of a bit by its reference.

        Args:
            item (int): bit reference

        Returns:
            value of the bit
        """
        if isinstance(item, slice):
            ret = []
            for i in range(item.start, item.stop, item.step or 1):
                byte_index = i // BYTE_SIZE
                bit_index = i % BYTE_SIZE
                ret.append((self._mem[byte_index] >> bit_index) & 1)
            return ret
        else:
            byte_index = item // BYTE_SIZE
            bit_index = item % BYTE_SIZE
            return (self._mem[byte_index] >> bit_index) & 1

    @synchronized
    def __setitem__(self, key, value):
        """
        Set the value of a bit.

        Args:
            key (int): bit reference
            value (int): value to set the bit to
        """
        if isinstance(key, slice):
            for i, val in zip(range(key.start, key.stop, key.step or 1), value):
                byte_index = i // BYTE_SIZE
                bit_index = i % BYTE_SIZE
                erase_mask = ~(1 << bit_index)
                new_byte = self._mem[byte_index] & erase_mask
                new_byte |= (val << bit_index)
                self._mem[byte_index] = new_byte
        else:
            byte_index = key // BYTE_SIZE
            bit_index = key % BYTE_SIZE
            erase_mask = ~(1 << bit_index)
            new_byte = self._mem[byte_index] & erase_mask
            new_byte |= (value << bit_index)
            self._mem[byte_index] = new_byte

    def __delitem__(self, key):
        """
        Set the value of a bit to zero.

        Args:
            key (int): bit reference
        """
        if isinstance(key, slice):
            self.__setitem__(key, [0] * abs(key.stop - key.start))
        else:
            self.__setitem__(key, 0)


class RegisterArray(MemoryBase):
    """
    RegisterArray is a 2 byte-aligned Modbus memory for input and holding registers.
    """

    def __init__(self, memory=None):
        """
        Args:
             memory (bytearray or None): the memory to address or ``None`` for new bytearray
        """
        super(RegisterArray, self).__init__(memory)

    @synchronized
    def __getitem__(self, item):
        """
        Gets the value of a register by its reference.

        Args:
            item (int or slice): register reference
        Returns:
            value of the register
        """
        if isinstance(item, slice):
            ret = []
            for i in range(item.start, item.stop, item.step or 1):
                ret.append(self._mem[i*REGISTER_BYTE_SIZE:(i+1)*REGISTER_BYTE_SIZE])
            return ret
        else:
            return self._mem[item * REGISTER_BYTE_SIZE:item * REGISTER_BYTE_SIZE + REGISTER_BYTE_SIZE]

    @synchronized
    def __setitem__(self, key, value):
        """
        Sets the value of a register by its reference.

        Args:
            key (int or slice): register reference
            value (int or list of int): value to set the register to
        """
        if isinstance(key, slice):
            for i, val in zip(range(key.start, key.stop, key.step or 1), value):
                self._mem[i*REGISTER_BYTE_SIZE:(i+1)*REGISTER_BYTE_SIZE] = val
        else:
            self._mem[key * REGISTER_BYTE_SIZE:key * REGISTER_BYTE_SIZE + REGISTER_BYTE_SIZE] = value

    def __delitem__(self, key):
        """
        Set the value of a register to zero.

        Args:
            key (int or slice): register reference
        """
        empty_register = b'\x00' * REGISTER_BYTE_SIZE

        if isinstance(key, slice):
            self.__setitem__(key, [empty_register] * abs(key.stop - key.start))
        else:
            self.__setitem__(key, empty_register)


class ModbusMemory(object):  # inherit from object for Python 2.7
    """
    ModbusMemory represents the physical memory of a Modbus device.
    It supports reading and writing of: coils, discrete inputs,
    input and holding registers.
    """

    def __init__(self, shared_memory=False):
        """
        Args:
            shared_memory (bool): If set to ``True`` a single shared memory is used.
        """
        self._shared_memory = shared_memory

        if self._shared_memory:
            mem = MemoryBase()
        else:
            mem = None
        self._coils = BitArray(mem)
        self._discrete_inputs = BitArray(mem)
        self._input_registers = RegisterArray(mem)
        self._holding_registers = RegisterArray(mem)

    def read_discrete_input(self, reference):
        """
        Read a discrete input bit.

        Args:
            reference (int): reference index of the discrete input

        Returns:
            value of the discrete input bit
        """
        return self._discrete_inputs[reference]

    def read_discrete_input_as_int(self, reference, quantity):
        """
        Read one or more discrete inputs as an int.

        Args:
            reference (int): reference index of the first discrete input
            quantity (int): number of bits to read
        Returns:
            value of discrete inputs as integer
        """
        value = 0
        for i in range(quantity):
            value |= self.read_discrete_input(reference + i) << i
        return value

    def write_discrete_input(self, reference, value):
        """
        Write one discrete input.

        Args:
            reference (int): reference index of the discrete input
            value (int): value to set the discrete input to
        """
        self._discrete_inputs[reference] = value

    def write_multiple_discrete_inputs(self, reference, quantity, values):
        """
        Write one or multiple discrete inputs.

        Args:
            reference (int): reference index of the first discrete input
            values (list of int): values to set the discrete input to
        """
        self._write_multiple_bits(self.write_discrete_input, reference, quantity, values)

    def read_coil(self, reference):
        """
        Read a coil.

        Args:
            reference (int): reference index of the coil

        Returns:
            value of the coil
        """
        return self._coils[reference]

    def read_coils_as_int(self, reference, quantity):
        """
        Read one or more coils as an int.

        Args:
            reference (int): reference index of the first coil
            quantity (int): number of bits to read
        Returns:
            value of coils as integer
        """
        byte = 0
        for i in range(quantity):
            byte |= self.read_coil(reference + i) << i
        return byte

    def write_coil(self, reference, value):
        """
        Write one coil.

        Args:
            reference (int): reference index of the coil
            value (int): value to set the coil to
        """
        self._coils[reference] = value

    @staticmethod
    def _write_multiple_bits(write_bit_func, reference, quantity, values):
        current_reference = reference
        for byte in values:
            for i in range(BYTE_SIZE):
                current_value = (byte >> i) & 1
                if (current_reference - reference) < quantity:
                    write_bit_func(current_reference, current_value)
                    current_reference += 1
                else:
                    break

    def write_multiple_coils(self, reference, quantity, values):
        """
        Write one or multiple coils.

        Args:
            reference (int): reference index of the first coil
            values (list of int): values to set the coils to
        """
        self._write_multiple_bits(self.write_coil, reference, quantity, values)

    def read_input_register(self, reference):
        """
        Read a input register.

        Args:
            reference (int): reference index of the input register

        Returns:
            value of the input register
        """
        return self._input_registers[reference]

    def read_holding_register(self, reference):
        """
        Read a holding register.

        Args:
            reference (int): reference index of the holding register

        Returns:
            value of the holding register
        """
        return self._holding_registers[reference]

    def write_input_register(self, reference, value):
        """
        Write a input register.

        Args:
            reference (int): reference index of the input register
            value (int): value to set the input register to
        """
        self._input_registers[reference] = value

    def write_holding_register(self, reference, value):
        """
        Write a holding register.

        Args:
            reference (int): reference index of the holding register
            value (int): value to set the holding register to
        """
        self._holding_registers[reference] = value

    def write_multiple_input_registers(self, reference, values):
        """
        Write one or multiple input registers.

        Args:
            reference (int): reference index of the first input register
            values (list of int): values to set the input register to
        """
        quantity = len(values)
        for i in range(quantity):
            self.write_input_register(reference + i, values[i])

    def write_multiple_holding_registers(self, reference, values):
        """
        Write one or multiple holding registers.

        Args:
            reference (int): reference index of the first holding register
            values (list of int): values to set the holding register to
        """
        quantity = len(values)
        for i in range(quantity):
            self.write_holding_register(reference + i, values[i])
