# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import serial

from htf.core import __version__ as htf_version
from htf._thread import Thread
from htf.communication.modbus.interface import ModbusTCPServerInterface, ModbusTLSServerInterface, \
    ModbusUDPServerInterface, ModbusSerialRTUServerInterface, \
    ModbusSerialASCIIServerInterface
from htf.communication.modbus.protocol import SerialRTUModbusRequest, SerialRTUModbusResponse, \
    TCPModbusRequest, TCPModbusResponse, SerialASCIIModbusRequest, SerialASCIIModbusResponse
from htf.communication.modbus.core import ModbusBaseServer as _ModbusBaseServer
import threading
import oser

BYTE_SIZE = 8
MAX_SERIAL_ADU_SIZE = 256
MAX_TCP_ADU_SIZE = 260


class ModbusBaseServer(_ModbusBaseServer):

    def __init__(self):
        """
        Modbus Base Server is an abstract base class.
        """
        self.listen_only = False
        self.device_address = 0
        self.diagnostics = b'\x00' * 2
        self.bus_message_count = 0
        self.bus_communication_error_count = 0
        self.bus_exception_error_count = 0
        self.slave_message_count = 0
        self.slave_no_response_count = 0
        self.slave_busy_count = 0
        self.bus_character_overrun_count = 0

    def handle_request(self):  # pragma: no cover
        """
        Request handler.

        Raises:
            Exception: abstract base class has no implementation.
        """
        raise Exception('Not implemented! Abstract base class.')

    def start(self):  # pragma: no cover
        """
        Start the server.

        Raises:
            Exception: abstract base class has no implementation.
        """
        raise Exception('Not implemented! Abstract base class.')

    def stop(self):  # pragma: no cover
        """
        Stop the server.

        Raises:
            Exception: abstract base class has no implementation.
        """
        raise Exception('Not implemented! Abstract base class.')

    def set_listen_only_mode(self, listen_only_mode_enabled=True):
        """
        Set listen only mode.

        Args:
            listen_only_mode_enabled (bool): ``True`` enables the listen only mode. ``False`` disables it.
        """
        self.listen_only = listen_only_mode_enabled

    def clear_counters(self):
        """
        Clear all counters.
        """
        self.bus_message_count = 0
        self.bus_communication_error_count = 0
        self.bus_exception_error_count = 0
        self.slave_message_count = 0
        self.slave_no_response_count = 0
        self.slave_busy_count = 0
        self.bus_character_overrun_count = 0

    def clear_diagnostics(self):
        """
        Clear diagnostics.
        """
        self.diagnostics = b'\x00' * 2

    def close(self):  # pragma: no cover
        """
        Close interface and stop the server.
        """
        self.interface.close()
        self.stop()


class ModbusSerialRTUServer(ModbusBaseServer):
    _interface_type = ModbusSerialRTUServerInterface

    def __init__(self, com_port, baud=9600, bytesize=serial.EIGHTBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False, device_address=1, timeout=1.0,
                 request_parser=SerialRTUModbusRequest,
                 response_parser=SerialRTUModbusResponse, callback=None):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            device_address (int): the device address to be used (Slave ID)
            timeout (float): the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
            callback: a callback used as a request handler
        """
        super(ModbusSerialRTUServer, self).__init__()
        self.device_address = device_address
        self.request_parser = request_parser
        self.response_parser = response_parser
        self.interface = self._interface_type(com_port, baud=baud, bytesize=bytesize, parity=parity,
                                              stopbits=stopbits, rtscts=rtscts, timeout=timeout)

        self.callback = callback

        self.thread = None
        self.run = False

        self.receive_event_callback = None

    def set_receive_event_callback(self, callback):
        """
        Set the callback to emit events.

        Args:
            callback (callable): the callback to be called
        """
        self.receive_event_callback = callback

    @staticmethod
    def get_broadcast_enabled_function_codes():
        """
        Return a list of functions codes that are processed for broadcast requests.
        By default the function codes ``5``, ``6``, ``15`` and ``16`` are supported.
        This method can be overwritten easily.

        Returns:
            list: a list of int containing the supported function codes
        """
        return [5, 6, 15, 16]

    def handle_request(self, data):
        """
        Request handler.
        A request is handled by calling the callback.

        Args:
            data (bytes): a request to be handle
        """
        if len(data) < 3:  # pragma: no cover
            self.bus_communication_error_count += 1
            if self.receive_event_callback is not None:  # pragma: no cover
                self.receive_event_callback(
                    is_communication_error=True,
                    is_broadcast=data[0] == b"\x00",
                    is_listen_only_mode=self.listen_only
                )
            return

        if len(data) > MAX_SERIAL_ADU_SIZE:
            self.bus_character_overrun_count += 1
            if self.receive_event_callback is not None:  # pragma: no cover
                self.receive_event_callback(
                    is_character_overrun=True,
                    is_broadcast=data[0] == b"\x00",
                    is_listen_only_mode=self.listen_only
                )

        adu = self.request_parser()

        try:
            adu.decode(data)
            self.bus_message_count += 1
            address = adu.address.get()
            is_broadcast = address == 0
            if is_broadcast or address == self.device_address:
                self.slave_message_count += 1
                process_request = True
                if is_broadcast:  # Broadcast
                    if self.receive_event_callback is not None:  # pragma: no cover
                        self.receive_event_callback(
                            is_broadcast=True,
                        )
                    function_code = adu.pdu.function_code.get()
                    # only enabled functions codes are processed for broadcast requests
                    if function_code not in self.get_broadcast_enabled_function_codes():
                        process_request = False

                if process_request:
                    response = self.callback(adu)
                    if not self.listen_only and not is_broadcast:
                        response.address.set(self.device_address)  # set device address for response
                        self.interface.write(response)
                    else:
                        self.slave_no_response_count += 1
        except oser.ChecksumMismatchException:
            self.bus_communication_error_count += 1
            if self.receive_event_callback is not None:  # pragma: no cover
                self.receive_event_callback(
                    is_character_overrun=True,
                    is_broadcast=data[0] == b"\x00",
                    is_listen_only_mode=self.listen_only
                )

    def start(self):
        """
        Start the server.
        """
        self.run = True
        assert self.thread is None

        def _thread():
            while self.run:
                try:
                    data, _ = self.interface.read()
                    if data:
                        self.handle_request(data)
                except Exception:
                    pass
                    # traceback.print_exc()

        self.thread = threading.Thread(target=_thread)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        """
        Stop the server.
        """
        self.run = False

    def close(self):
        """
        Close interface and stop the server.
        """
        self.interface.close()
        self.stop()


class ModbusSerialASCIIServer(ModbusSerialRTUServer):
    _interface_type = ModbusSerialASCIIServerInterface

    def __init__(self, com_port, baud=9600, bytesize=serial.EIGHTBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, rtscts=False, device_address=1, timeout=1.0,
                 request_parser=SerialASCIIModbusRequest,
                 response_parser=SerialASCIIModbusResponse, callback=None):
        """
        Args:
            com_port (str): the com port to be used
            baud (int): the baud rate
            bytesize (int): number of bits in a byte
            parity (str): serial.PARITY_NONE, serial.PARITY_ODD or serial.PARITY_EVEN
            stopbits (int or float): number of stopbits (1, 1.5 or 2)
            rtscts (bool): if True use flow control
            device_address (int): the device address to be used (Slave ID)
            timeout (float): the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
            callback: a callback used as a request handler
        """
        super(ModbusSerialASCIIServer, self).__init__(
            com_port=com_port, baud=baud, bytesize=bytesize, parity=parity, stopbits=stopbits, rtscts=rtscts,
            device_address=device_address, timeout=timeout, request_parser=request_parser,
            response_parser=response_parser, callback=callback)

    def change_ascii_delimiter(self, delimiter):
        """
        Change ASCII delimiter to ``ascii_delimiter``.

        Args:
            delimiter (bytes): bytes of length 1 representing the new lf character
        """
        self.interface.change_ascii_delimiter(delimiter)


class ModbusTCPServer(ModbusBaseServer):
    _interface_type = ModbusTCPServerInterface

    def __init__(self, bind_address, port=502, timeout=1.0, callback=None,
                 request_parser=TCPModbusRequest,
                 response_parser=TCPModbusResponse):
        """
        Args:
            bind_address (str): the ip-address to listen on
            port (int): the port to listen on
            timeout (float): the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
        """
        super(ModbusTCPServer, self).__init__()
        self._init_parsers_variables_and_callback(request_parser, response_parser, callback)
        self.interface = self._interface_type(bind_address, port=port, timeout=timeout)

    def _init_parsers_variables_and_callback(self, request_parser, response_parser, callback):
        self.request_parser = request_parser
        self.response_parser = response_parser

        if callback is None:  # pragma: no cover
            raise Exception('Callback has to be set!')

        self.callback = callback

        self.connections = {}

        self.accept_thread = None
        self.threads = []
        self.run = False
        self._threads_lock = threading.Lock()
        self._connections_lock = threading.Lock()

    def handle_request(self, connection, data):
        """
        The request handler.
        A request is handled by calling the callback.

        Args:
            connection: the current TCP connection
            data (bytes): the request data
        """
        if len(data) > MAX_TCP_ADU_SIZE:
            self.bus_character_overrun_count += 1

        adu = self.request_parser()
        adu.decode(data)
        self.bus_message_count += 1
        self.slave_message_count += 1
        response = self.callback(adu)
        response.transaction_identifier.set(adu.transaction_identifier.get())
        response.protocol_identifier.set(adu.protocol_identifier.get())
        response.length.set(len(response.pdu.encode()) + 1)
        response.unit_identifier.set(adu.unit_identifier.get())
        self.interface.write(connection, response)

    def accept_connections(self):
        """
        A thread to accept new incoming connections.
        """
        while self.run:
            try:
                conn, address = self.interface.accept()
                with self._connections_lock:
                    self.connections[conn] = address

                def _thread():
                    thread_local_connection = conn
                    thread_local_address = address
                    while self.run:
                        try:
                            data = self.interface.read(thread_local_connection)
                            if data:
                                self.handle_request(thread_local_connection, data)
                        except Exception:
                            pass
                    self.interface.close_connection(thread_local_connection)
                    with self._connections_lock:
                        del self.connections[thread_local_connection]

                with self._threads_lock:
                    thread = Thread(target=_thread)
                    thread.daemon = True
                    self.threads.append(thread)
                    thread.start()
            except Exception:
                pass

    def start(self):
        """
        Start the server.
        """
        self.run = True

        assert self.accept_thread is None

        self.accept_thread = Thread(target=self.accept_connections)
        self.accept_thread.daemon = True
        self.accept_thread.start()

    def stop(self):
        """
        Stop the server.
        """
        self.run = False

        with self._threads_lock:
            for thread in self.threads:
                thread.join(timeout=30)
                assert thread.is_alive() is False

            self.threads = []

        try:
            self.accept_thread.join(timeout=30)
            assert self.accept_thread.is_alive() is False
        except AttributeError:  # pragma: no cover
            pass

    def close(self):
        """
        Close the interface and stop the server.
        """
        self.interface.close()
        self.stop()


class ModbusTLSServer(ModbusTCPServer):
    _interface_type = ModbusTLSServerInterface

    def __init__(self, bind_address, port=502, timeout=1.0, callback=None,
                 certfile=None, keyfile=None, password=None,
                 request_parser=TCPModbusRequest,
                 response_parser=TCPModbusResponse):
        """
        Args:
            bind_address (str): the ip-address to listen on
            port (int): the port to listen on
            timeout (float): the timeout in seconds
            callback: the callback to handle requests
            certfile (str): path to TLS certificate
            keyfile (str): path to TLS keyfile
            password (str): TLS password
            request_parser: the request parser
            response_parser: the response parser
        """
        ModbusBaseServer.__init__(self)
        self._init_parsers_variables_and_callback(request_parser, response_parser, callback)
        self.interface = self._interface_type(bind_address, port=port, timeout=timeout,
                                              certfile=certfile, keyfile=keyfile, password=password)


class ModbusUDPServer(ModbusTCPServer):
    _interface_type = ModbusUDPServerInterface

    def __init__(self, bind_address, port=502, timeout=None, callback=None,
                 request_parser=TCPModbusRequest,
                 response_parser=TCPModbusResponse):
        """
        Args:
            bind_address (str): the ip-address to listen on
            port (int): the port to listen on
            timeout=None (float): the timeout in seconds
            request_parser: the request parser
            response_parser: the response parser
        """
        super(ModbusUDPServer, self).__init__(
            bind_address=bind_address,
            port=port,
            timeout=timeout,
            callback=callback,
            request_parser=request_parser,
            response_parser=response_parser
        )

    def accept_connections(self):
        """
        A thread to accept new incoming connections.
        """
        while self.run:
            try:
                data, address = self.interface.read()

                def _thread():
                    try:
                        self.handle_request(address, data)
                    except Exception:
                        pass
                if data:  # pragma: no cover
                    with self._threads_lock:
                        thread = Thread(target=_thread)
                        thread.daemon = True
                        self.threads.append(thread)
                        thread.start()
            except Exception:
                pass
