# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64, DAQmx_Val_GroupByChannel, DAQmx_Val_FiniteSamps, DAQmx_Val_Rising,\
    DAQmx_Val_Volts, DAQmx_Val_Cfg_Default, uInt64, DAQmx_Val_ContSamps,\
    DAQmx_Val_DoNotAllowRegen, DAQmx_Val_OnBrdMemHalfFullOrLess,\
    DAQmx_Val_OnBrdMemMoreThanHalfFull
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageInputPort(DelegateMixin):
    """
    ``AnalogVoltageInputPort`` is used to access a single analog input port
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    def __init__(self, descriptor,
                 minimumVoltage=-10.0,
                 maximumVoltage=10.0,
                 numberOfLines=32,
                 numberOfSamples=2,
                 sampleFrequency=1000.0,
                 timeout=1.0,
                 terminalConfig=DAQmx_Val_Cfg_Default, continuousMode=False):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                ``cDaq1Mod1/line0``.
            minimumVoltage=-10.0 (float): the minimum voltage.
            maximumVoltage=10.0 (float): the maximum voltage.
            numberOfLines=32 (int): the number of lines defined in the
                descriptor. For example the descriptor "cDaq1Mod3/ai0:31"
                defines 32 lines.
            numberOfSamples=2 (int): the number of samples to acquire for one
                average value.
            sampleFrequency=1000.0 (float): the sample frequency in Hertz.
            timeout=1.0 (float): the timeout when acquiring a new value in
                seconds.
            terminalConfig=DAQmx_Val_Cfg_Default: the terminal configuration
                for DAQmxCreateAIVoltageChan(). Possible values: DAQmx_Val_RSE,
                DAQmx_Val_NRSE, DAQmx_Val_Diff, DAQmx_Val_PseudoDiff.
            continuousMode=False: if set to ``True`` continuous sampling is
                done and get() becomes blocking. Simply read samples in a loop.
        """
        taskName = "AI_" + descriptor.replace("/", "_")
        self.numberOfSamples = numberOfSamples
        self.sampleFrequency = sampleFrequency
        self.numberOfLines = numberOfLines
        self.continuousMode = continuousMode
        self.timeout = timeout
        self.data = cast(
            (float64 * self.numberOfSamples * self.numberOfLines)(),
            POINTER(float64))
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateAIVoltageChan(descriptor,
                                      None,
                                      terminalConfig,
                                      float64(minimumVoltage),
                                      float64(maximumVoltage),
                                      DAQmx_Val_Volts,
                                      None)

        if self.continuousMode is False:
            self.task.CfgSampClkTiming(
                None,
                float(self.sampleFrequency),
                DAQmx_Val_Rising,
                DAQmx_Val_FiniteSamps,
                uInt64(self.numberOfSamples * self.numberOfLines))
        else:
            self.task.CfgSampClkTiming(
                None,
                float(self.sampleFrequency),
                DAQmx_Val_Rising, DAQmx_Val_ContSamps,
                uInt64(self.numberOfSamples * self.numberOfLines))
            self.task.SetAIDataXferReqCond(descriptor,
                                           DAQmx_Val_OnBrdMemMoreThanHalfFull)
            self.task.SetBufInputBufSize(2 * int(self.numberOfSamples))

    def get(self):
        """
        Return the current analog value.

        Returned data layout::

            [
                [line0 sample0, line0 sample1, ..., line0 sampleN-1],
                [line1 sample0, line1 sample1, ..., line1 sampleN-1],
                ...
                [lineM-1 sample0, lineM-1 sample1, ..., lineM-1 sampleN-1]
            ]
        """
        self.task.ReadAnalogF64(self.numberOfSamples,
                                float64(self.timeout),
                                DAQmx_Val_GroupByChannel,
                                self.data,
                                self.numberOfSamples * self.numberOfLines,
                                byref(self.read),
                                None)

        data = []

        for line in range(self.numberOfLines):
            lineData = []
            for sample in range(self.numberOfSamples):
                lineData.append(self.data[line * self.numberOfSamples + sample])
            data.append(lineData)
        return data

    def set(self, state=True):
        raise NotImplementedError(
            "There is no meaningful implementation for "
            "a 'set' method in the context of this class.")

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    ai0 = AnalogVoltageInputPort("Dev1/ai0:0", numberOfSamples=10, sampleFrequency=10, continuousMode=True)

    for ii in range(100000):
        samples = ai0.get()[0]
        for sample in samples:
            print(sample)
        # time.sleep(.5)#
