# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64
from htf.mixins.delegation import DelegateMixin


class DigitalInputSingleLine(DelegateMixin):
    """
    ``DigitalInputSingleLine`` is used to access a single digital input line
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor):
        """

        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`.
        """
        taskName = "DI_" + descriptor.replace("/", "_")
        self.state = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateDIChan(descriptor, None, DAQmx_Val_ChanForAllLines)

    def get(self):
        """
        Return the current pin state.
        """
        self.task.ReadDigitalU32(self.N,
                                 float64(1.0),
                                 DAQmx_Val_ChanForAllLines,
                                 self.data,
                                 self.N,
                                 byref(self.read),
                                 None)
        self.task.StopTask()
        return bool(self.data[0])

    def set(self, state=True):
        raise NotImplementedError(
            "There is no meaningful implementation for "
            "a 'set' method in the context of this class.")

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    import time
    l7 = DigitalInputSingleLine("Dev1/line0")
    l8 = DigitalInputSingleLine("Dev1/line1")

    for ii in range(10):
        print(l7.get())
        print(l8.get())
        time.sleep(.5)

        print(l8.get())
        print(l7.get())
        time.sleep(.5)
