# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64, DAQmx_Val_GroupByChannel
from htf.mixins.delegation import DelegateMixin


class DigitalOutputPort(DelegateMixin):
    """
    ``DigitalOutputPort`` is used to access a digital output port
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/port0`.
        """
        taskName = "DO_" + descriptor.replace("/", "_")
        self.value = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))
        self.written = int32(0)
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateDOChan(descriptor, None, DAQmx_Val_ChanForAllLines)

    def get(self):
        """
        Return the current port state.
        """
        self.task.ReadDigitalU32(self.N,
                                 float64(1.0),
                                 DAQmx_Val_GroupByChannel,
                                 self.data,
                                 self.N,
                                 byref(self.read),
                                 None)
        return self.data[0]

    def set(self, value):
        """
        Set the port to ``value``.

        Args:
            value (float): new value for output port
        """

        self.value = value

        self.data[0] = self.data[1] = value

        self.task.WriteDigitalU32(self.N,
                                  True,
                                  float64(1.0),
                                  DAQmx_Val_ChanForAllLines,
                                  self.data,
                                  byref(self.written),
                                  None)
        self.task.StopTask()

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    a = DigitalOutputPort("Dev1/port1")
    a.set(0b10101)
    print(bin(a.get()))
    exit(0)
    a = DigitalOutputPort("cDaq4Mod1/port0")
    b = DigitalOutputPort("cDaq4Mod1/port0")

    a.set(0x12345678)
    assert a.get() == 0x12345678
    assert b.get() == 0x12345678

    a.set(0)
    assert a.get() == 0
    assert b.get() == 0

    exit(0)
