# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, int32, DAQmx_Val_ChanForAllLines,\
    float64, DAQmx_Val_ChanPerLine
from htf.mixins.delegation import DelegateMixin


class DigitalOutputSingleLine(DelegateMixin):
    """
    ``DigitalOutputSingleLine`` is used to access a single digital output line
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`.
        """
        taskName = "DO_" + descriptor.replace("/", "_")
        self.state = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))
        self.written = int32(0)
        self.read = int32(0)
        self.task = Task(taskName)

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateDOChan(descriptor, None, DAQmx_Val_ChanPerLine)

    def get(self):
        """
        Read the current pin state.
        """
        self.task.ReadDigitalU32(self.N,
                                 float64(1.0),
                                 DAQmx_Val_ChanForAllLines,
                                 self.data,
                                 self.N,
                                 byref(self.read),
                                 None)
        self.task.StopTask()
        return bool(self.data[0])

    def set(self, state=True):
        """
        Set the pin state to ``state``.

        Args:
            state=True (bool): new state for output pin.
        """
        if state:
            self.data[0] = uInt32(0xFFFFFFFF)
            self.data[1] = uInt32(0xFFFFFFFF)
        else:
            self.data[0] = uInt32(0x00000000)
            self.data[1] = uInt32(0x00000000)

        self.task.WriteDigitalU32(self.N,
                                  True,
                                  float64(1.0),
                                  DAQmx_Val_ChanForAllLines,
                                  self.data,
                                  byref(self.written),
                                  None)
        self.task.StopTask()

    def reset(self):
        """
        Reset the pin state.
        """
        self.set(state=False)

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    import time
    a = DigitalOutputSingleLine("Dev1/port0")
    b = DigitalOutputSingleLine("Dev1/port1")

    a.reset()
    assert a.get() is False
    assert b.get() is False

    a.set()
    assert a.get() is True
    assert b.get() is True

    b.reset()
    assert a.get() is False
    assert b.get() is False

    b.set()
    assert a.get() is True
    assert b.get() is True

    a.set()

    a.close()
    b.close()

    exit(0)
    c = DigitalOutputSingleLine("cDaq4Mod1/line0")

    assert c.get() is True

    c.close()

    # a.set()
    # b.set()
    # a.get()
    # b.get()

    exit(0)

    l7 = DigitalOutputSingleLine("cDaq4Mod1/line0")
    l8 = DigitalOutputSingleLine("cDaq4Mod1/line0")

    for ii in range(10):
        l7.reset()
        l8.set()
        time.sleep(.5)

        l8.reset()
        l7.set()
        time.sleep(.5)
        print(ii)
