# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, float64, DAQmx_Val_CountUp, DAQmx_Val_CountDown, DAQmx_Val_Rising, DAQmx_Val_Falling
from htf.mixins.delegation import DelegateMixin


class EdgeCounterInput(DelegateMixin):
    """
    ``EdgeCounterInput`` is used to access a counter input on a DAQmx device.
    The counter is 32 bit wide.

    A DAQmx task is automatically created and deleted in the constructor and destructor.
    """

    def __init__(self, descriptor, initialCount=0, risingEdge=True, countUp=True):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/ctr0`.
            initialCount=0 (int): The initial count.
            risingEdge=True (bool): If set to True rising edges are counted
                else falling edges are counted.
            countUp=True (bool): If set to True the counter direction is up
                else it is down.
        """
        taskName = "CTR_" + descriptor.replace("/", "_")
        self.count = initialCount
        self.offset = 0
        self.data = uInt32(0)
        self.task = Task(taskName)

        self.edge = DAQmx_Val_Rising
        if not risingEdge:
            self.edge = DAQmx_Val_Falling

        self.countDirection = DAQmx_Val_CountUp
        if not countUp:
            self.countDirection = DAQmx_Val_CountDown

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateCICountEdgesChan(descriptor,
                                         None,
                                         edge=self.edge,
                                         initialCount=initialCount,
                                         countDirection=self.countDirection)
        self.task.start()

    def get(self):
        """
        Return the current edge count.
        """
        self.task.ReadCounterScalarU32(float64(1.0), byref(self.data), None)

        self.count = self.data.value

        result = self.count - self.offset
        return result

    def set(self, count):
        """
        Set the count to the desired value.

        Args:
            count (int): the new counter value
        """
        self.offset = self.count + count

    def reset(self):
        """
        Reset the counter to 0.
        """
        self.set(count=0)

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    import time
    ctr0 = EdgeCounterInput(descriptor="Dev1/ctr0", initialCount=0, risingEdge=True, countUp=True)

    for ii in range(100):
        print(ctr0.get())
        # exit(0)
        time.sleep(0.1)

    print("reset..")
    ctr0.reset()

    for ii in range(100):
        print(ctr0.get())
        # exit(0)
        time.sleep(0.1)
