# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

from htf.core import __version__
from htf.io.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.core import uInt32, float64, DAQmx_Val_Rising, DAQmx_Val_Falling,\
    DAQmx_Val_Hz, DAQmx_Val_LowFreq1Ctr
from htf.mixins.delegation import DelegateMixin


class FrequencyInput(DelegateMixin):
    """
    ``FrequencyInput`` is used to measure a low frequency in hertz on a DAQmx
    device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    def __init__(self, descriptor, fMin, fMax, measurementTime, risingEdge=True):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/ctr0`
            fMin (float): The minimum frequency in the input signal in Hertz.
            fMin (float): The maximum frequency in the input signal in Hertz.
            risingEdge=True (bool): If set to True rising edges are counted
                else falling edges are counted.
            measurementTime (float): The measurement time in seconds.
        """
        taskName = "FREQLOW_" + descriptor.replace("/", "_")
        self.fMin = fMin
        self.fMax = fMax
        self.measurementTime = measurementTime
        self.data = float64(0.0)
        self.task = Task(taskName)

        self.edge = DAQmx_Val_Rising
        if not risingEdge:
            self.edge = DAQmx_Val_Falling

        if not isinstance(descriptor, (bytes)):
            descriptor = descriptor.encode()

        self.task.CreateCIFreqChan(descriptor,
                                   None,
                                   minVal=self.fMin,
                                   maxVal=self.fMax,
                                   units=DAQmx_Val_Hz,
                                   edge=self.edge,
                                   measMethod=DAQmx_Val_LowFreq1Ctr,
                                   measTime=float64(self.measurementTime),
                                   divisor=4,
                                   customScaleName=None)
        self.task.start()

    def get(self):
        """
        Return the current frequency.
        """
        self.task.ReadCounterScalarF64(float64(1.0), byref(self.data), None)

        return self.data.value

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    import time

    f0 = FrequencyInput(descriptor="Dev1/ctr0", fMin=1000.0, fMax=10000.0, measurementTime=0.001, risingEdge=True)

    for ii in range(100):
        print(f0.get())
        time.sleep(0.1)
