# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import \
    absolute_import, division, print_function, unicode_literals

import jira
import functools
import json

import logging
logger = logging.getLogger(__name__)


def _create_flat_query_dict(d, separator="."):
    """
    Flatten a hierarchical dictionary. Every hierarchy is represented with a '.'.

    Attention: it is possible that outer keys are overwritten from new keys.

    Args:
        d (dict): A hierarchical dict with keys and sub keys

    Returns:
        dict: a flattened dict

    """

    def _get_flat_dict(data, prefix=""):
        ret = dict()

        for key, value in data.items():
            if not key.startswith("_"):
                if isinstance(value, dict):
                    ret.update(_get_flat_dict(value, prefix=prefix+key+separator))
                else:
                    ret["{}{}".format(prefix, key)] = value

        return ret

    flat = _get_flat_dict(d)

    return flat


def jira_connector_decorator(func):
    @functools.wraps(func)
    def _decorator(self, *args, **kwargs):
        replacements = {
            "test": self._htf_test_method_name,
            "testcase": self.__class__.__name__,
            "project": self._HTF_JIRA_PROJECT

        }

        query = self._HTF_JIRA_QUERY_TEMPLATE.format(**replacements)

        assert self._HTF_JIRA_AUTH in [None, "auth", "basic_auth", "oauth", "kerberos"]

        logger.info("Query JIRA issues")

        options = {'server': self._HTF_JIRA_URL}
        logger.debug("Using JIRA server {}".format(self._HTF_JIRA_URL))

        if self._HTF_JIRA_AUTH in ["auth", "basic_auth"]:
            options[self._HTF_JIRA_AUTH] = (self._HTF_JIRA_USERNAME, self._HTF_JIRA_PASSWORD)
            logger.debug("Using authentication scheme {} and username {}".format(self._HTF_JIRA_AUTH, self._HTF_JIRA_USERNAME))

        elif self._HTF_JIRA_AUTH in ["oauth"]:
            options["oauth"] = self._HTF_JIRA_OAUTH_DICT
            logger.debug("Using authentication scheme {}".format(self._HTF_JIRA_AUTH))

        elif self._HTF_JIRA_AUTH in ["kerberos"]:
            options["kerberos"] = True
            options["kerberos_options"] = self._HTF_JIRA_KERBEROS_OPTIONS
            logger.debug("Using authentication scheme {}".format(self._HTF_JIRA_AUTH))

        jira_api = jira.JIRA(**options)

        # search issues
        logger.debug("query: {}".format(query))
        issues = jira_api.search_issues(query)

        for issue in sorted(issues, key=lambda issue: issue.key):
            flat_issue = _create_flat_query_dict(issue.raw, "/")
            logger.debug("raw issue: {}".format(json.dumps(flat_issue, indent=4, sort_keys=True)))

            self.__data__["links"].append({
                "title": self._HTF_JIRA_TITLE_TEMPLATE.format(**flat_issue),
                "href": "{}/browse/{}".format(self._HTF_JIRA_URL, issue.key)})

        return func(*args, **kwargs)
    return _decorator


class JIRAConnector:
    """
    The ``htf.JIRAConnector`` searches JIRA using the rest api to find issues
    for the current test.
    Found issues are append to the test and put into the test report.
    """
    _HTF_JIRA_URL = None
    _HTF_JIRA_PROJECT = None
    _HTF_JIRA_USERNAME = None
    _HTF_JIRA_PASSWORD = None
    _HTF_JIRA_AUTH = None
    _HTF_JIRA_OAUTH_DICT = None
    _HTF_JIRA_KERBEROS_OPTIONS = None
    _HTF_JIRA_QUERY_TEMPLATE = 'project={project} and (summary~"{testcase}.{test}" or description~"{testcase}.{test}")'
    _HTF_JIRA_TITLE_TEMPLATE = "{fields/issuetype/name} {key} ({fields/status/name}): {fields/summary}"

    def __init__(self, *args, **kwargs):  # pragma: no cover
        super(JIRAConnector, self).__init__(*args, **kwargs)

    def setup_jira_connector(self, url, project,
                             username=None, password=None, auth=None,
                             oauth_dict=None, kerberos_options=None,
                             query_template=None,
                             title_template=None):
        """
        Set up the ``JIRAConnector`` so that it will search you JIRA installation
        for issues for the current test.

        Args:
            url (str): the url to you JIRA installation
            project (str): the JIRA project
            username=None (str or None): the username to be used for authentication
            password=None (str or None): the password to be used for authentication
            auth=None (str or None): possible values are:

                - ``None``: no authentication is used
                - ``auth``: cookie based authentication
                - ``basic_auth``: HTTP basic authentication
                - ``oauth``: OAuth authentication is used (you have to supply ``oauth_dict``, too)
                - ``kerberos``: Kerberos authentication (you have to supply ``kerberos_options``, too)

            oauth_dict=None (dict): options for ``oauth`` authentication
            kerberos_options=None (dict): options for ``kerberos`` authentication

            query_template="default" (str): the template for the JIRA query string. supported replacements are
                - ``{testcase}``: is replaced with the TestCase's class name
                - ``{test}``: is replaced with the current test name
                - ``{project}``: is replaced with the supplied project

                default: "project={project} and (summary~\"{testcase}.{test}\" or description~\"{testcase}.{test}\")"
            title_template="default" (str): The template for the resulting title of the link in the test report.
                The issue is flattend using ``'/'`` as the separator. So the template may include keys like
                ``{fields/status/name}`` to show the name of the current ticket status, etc.
        """

        self._HTF_JIRA_URL = url
        self._HTF_JIRA_PROJECT = project
        self._HTF_JIRA_USERNAME = username
        self._HTF_JIRA_PASSWORD = password
        self._HTF_JIRA_AUTH = auth
        self._HTF_JIRA_OAUTH_DICT = oauth_dict
        self._HTF_JIRA_KERBEROS_OPTIONS = kerberos_options

        if query_template is not None:
            self._HTF_JIRA_QUERY_TEMPLATE = query_template

        if title_template is not None:
            self._HTF_JIRA_TITLE_TEMPLATE = title_template

        # decorate test method with jira_connector_decorator
        try:
            test_method = self.__getattribute__(self._htf_test_method_name)
            logger.info("decorate {}".format(self._htf_test_method_name))
            decorated_test_method = jira_connector_decorator(test_method)
            self.__setattr__(self._htf_test_method_name, decorated_test_method.__get__(self, self.__class__))
        except AttributeError:  # pragma: no cover
            raise ValueError("test method not found in %s (did you run setUp() first?!)" %
                             (self.__class__))