# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import base64
import json
import copy


try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

import requests

from htf.termcolors import system


class JSONUnicodeEncoder(json.JSONEncoder):
    def default(self, obj):
        if not isinstance(obj, bytes):
            return obj.encode("utf-8", errors="ignore")

        return json.JSONEncoder.default(self, obj)


class ReportServerReport(object):
    """
    ``ReportServerReport`` report a test report to a report server.

    Usage::

        from __future__ import absolute_import, division, print_function, unicode_literals
        from htf import TestLoader, TestRunner, JSONTestReport

        suite = TestLoader().loadTestsFromTestCase(Test)
        result = TestRunner(title="A test", reports=JSONTestReport("test.json"), verbosity=2).run(suite)
    """

    def __init__(self, url):
        """
        Initialize a ``JSONTestReport`` instance.

        Args:
            filename (str): the filename for the json test report.
        """
        self.url = url
        self.report_type = "REPORT SERVER"

    def create_data_url(self, mimetype, data):
        if not isinstance(data, type(b"")):
            data = data.encode()

        encoded = base64.b64encode(data)

        if not isinstance(encoded, type("")):
            encoded = encoded.decode()

        data_url = "data:{};base64,".format(mimetype) + encoded

        return data_url

    def render(self, data):
        """
        Generates and sends a test report.

        Args:
            data (dict): test result data
        """
        data = copy.deepcopy(data)

        # convert tags into a list
        for test in data["tests"]:
            try:
                test["data"]["tags"] = list(sorted(test["data"]["tags"]))
            except KeyError:
                pass

        # escape all data items
        def safe_escape(data):
            if isinstance(data, dict):
                ret = {k: safe_escape(v) for k, v in data.items()}
                data.update(ret)
                return data
            elif isinstance(data, list):
                return [safe_escape(item) for item in data]
            else:
                if isinstance(data, bytes):
                    return data.encode("utf-8", errors="ignore")
                return data

        data = safe_escape(data)

        # read attachments
        for test in data["tests"]:
            # read attachments and encode them to data urls
            try:
                for attachment in test["data"]["attachments"]:
                    try:
                        with open(attachment["filename_abs"], "rb") as fp:
                            file_contents = fp.read()
                            data_url = self.create_data_url(attachment["mimetype"], file_contents)
                            if not isinstance(data_url, type("")):
                                data_url = data_url.encode()

                            attachment["data_url"] = data_url
                    except KeyError:
                        pass
            except KeyError:
                pass

        self.write({'origin': 'htf', 'report': data})

    def write(self, data):
        verify = True

        if "HTF_INSECURE" in os.environ:
            verify = False
        try:
            response = requests.post(self.url, json=data, verify=verify)
            assert response.status_code == 204, "The report server {} returned {}".format(self.url,
                                                                                          response.status_code)
        except Exception as e:
            print(system("An error occurred when sending sending the test report to {}: {}".format(self.url, e)))
        else:
            print(system("Sent report to {}".format(self.url)))
