# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import time
from functools import wraps
from contextlib import contextmanager


def attempts_generator(attempts=2, delay=None):
    """
    A generator for attempts with a configurable delay.

    Args:
        attempts=2 (int): the number of attempts
        delay=None (float or int): the delay between attempts in seconds

    Returns:
        a generator that generates ``attempts`` attempts with the given delay between them
    """

    delay = delay or 0

    for attempt in range(attempts):
        yield attempt

        if attempt < attempts - 1:
            time.sleep(delay)


def retry(action, args=(), kwargs={},
          attempts=2, delay=None,
          retry_exceptions=(BaseException,)):
    """
    Retry ``action`` with a given number of attempts and a given delay.

    Args:
        action (callable): the action to be retried
        args=() (tuple): the positional arguments passed to ``action``
        kwargs={} (dict): the keyword arguments passed to ``action``
        attempts=2 (int): the number of attempts
        delay=None (float or int): the delay between attempts in seconds
        retry_exceptions=(BaseException,) (tuple): the exceptions to be catched to retry the action

    Returns:
        the return value of ``action(*args, **kwargs)``
    """

    assert callable(action)

    n = 1

    for _ in attempts_generator(attempts=attempts, delay=delay):
        try:
            return action(*args, **kwargs)
        except retry_exceptions:
            if n == attempts:
                raise
        finally:
            n += 1


def retriable(attempts=2, delay=None,
              retry_exceptions=(BaseException,)):
    """
    A decorator factory for retry(). Wrap your function in @retriable(...) to
    give it retry powers!

    Args:
        attempts=2 (int): the number of attempts
        delay=None (float or int): the delay between attempts in seconds
        retry_exceptions=(BaseException,) (tuple): the exceptions to be catched to retry the action

    Returns:
        a function decorator supporting retries
    """
    def _retriable_factory(func):
        @wraps(func)
        def _retriable_wrapper(*args, **kwargs):
            return retry(action=func, args=args, kwargs=kwargs,
                         attempts=attempts, delay=delay,
                         retry_exceptions=retry_exceptions)
        return _retriable_wrapper
    return _retriable_factory


@contextmanager
def retrying(action,
             attempts=2, delay=None,
             retry_exceptions=(BaseException,)):
    """
    A context manager for wrapping ``action`` with retry functionality
    with a given number of attempts and a given delay.

    Args:
        action (callable): the action to be retried
        attempts=2 (int): the number of attempts
        delay=None (float or int): the delay between attempts in seconds
        retry_exceptions=(BaseException,) (tuple): the exceptions to be catched to retry the action

    Returns:
        a context manager supporting retries
    """

    yield retriable(attempts=attempts, delay=delay,
                    retry_exceptions=retry_exceptions)(action)

