# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import ply.lex as lex
import ply.yacc as yacc
from htf.exceptions import ParsingError


class _TagsParser(object):

    tokens = (
        'NAME',
        'AND', 'OR', 'XOR',
        'NOT',
        'LPAREN', 'RPAREN',
    )

    # Tokens

    t_AND = r'&'
    t_OR = r'\|'
    t_NOT = r'!'
    t_XOR = r'\^'
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_NAME = r'[a-zA-Z0-9_\-\.]+'

    # Ignored characters
    t_ignore = " \t\n\r"

    def t_error(self, t):
        raise ParsingError("Illegal character '%s'" % t.value[0])

    # Parsing rules

    precedence = (
        ('left', 'AND', 'OR', 'XOR'),
        ('right', 'NOT'),
    )

    def p_statement_expr(self, t):
        'statement : expression'
        t[0] = t[1]

    def p_expression_binop(self, t):
        '''expression : expression AND expression
                      | expression OR expression
                      | expression XOR expression'''
        if t[2] == '&':
            t[0] = t[1] & t[3]
        elif t[2] == '|':
            t[0] = t[1] | t[3]
        elif t[2] == '^':
            t[0] = t[1] ^ t[3]

    def p_expression_not(self, t):
        '''expression : NOT expression'''
        t[0] = not t[2]
        #print("p_expression_not", "not", t[2], "->", t[0])

    def p_expression_group(self, t):
        'expression : LPAREN expression RPAREN'
        t[0] = t[2]

    def p_expression_name(self, t):
        'expression : NAME'
        t[0] = t[1] in self.tag_list
        #print("p_expression_name", t[1], "in", self.tag_list, "->", t[0])

    def p_error(self, t):
        try:
            raise ParsingError("Syntax error at '%s'" % t.value)
        except AttributeError:
            raise ParsingError("Syntax error in '%s'" % self._expression)

    def __init__(self, expression):
        self.tag_list = None
        self.lexer = lex.lex(module=self)
        self.parser = yacc.yacc(module=self, debug=False, write_tables=False)
        self._expression = expression

    def match(self, tags):
        self.tag_list = tags
        result = self.parser.parse(self._expression)
        self.tag_list = None
        return result
