# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#


import itertools
import json
from collections import OrderedDict

import yaml
import csv


def data(iterable, unpack=False):
    """
    Create a data driven test using data.

    Args:
        iterable: an iterable used as parameter data source. Can also be a callable.
        unpack=False (bool): if set to ``True`` iterables are unpacked.
    """
    
    def wrapper(func):
        # call _test_generator if callable
        if callable(iterable):
            it = iterable()
        else:
            it = iterable

        if hasattr(func, "_test_generator"):
            raise Exception("Data decorators can only be used single.")

        func._test_generator = it
        func._test_generator_unpack = unpack

        return func
    return wrapper


class JSONFileIterator:
    """
    JSON file iterator.

    Args:
        filename (str): the filename to be read.

    Returns:
        iterator: iterator for the file contents.
    """
    def __init__(self, filename):
        self.filename = filename

    def __iter__(self):
        with open(self.filename, "r") as fp:
            data = json.load(fp)
        return iter(data)


def json_data(filename, unpack=False):
    """
    Create a data driven test using data from a json file.

    Args:
        filename (str): the filename to be read.
        unpack=False (bool): if set to ``True`` iterables are unpacked.
    """
    def wrapper(func):
        func._test_generator = JSONFileIterator(filename)
        func._test_generator_unpack = unpack
        return func

    return wrapper


class YAMLFileIterator:
    """
    YAML file iterator.

    Args:
        filename (str): the filename to be read.

    Returns:
        iterator: iterator for the file contents.
    """
    def __init__(self, filename):
        self.filename = filename

    def __iter__(self):
        with open(self.filename, "r") as fp:
            data = yaml.load(fp, Loader=yaml.FullLoader)
        return iter(data)


def yaml_data(filename, unpack=False):
    """
    Create a data driven test using data from a YAML file.

    Args:
        filename (str): the filename to be read.
        unpack=False (bool): if set to ``True`` iterables are unpacked.
    """
    def wrapper(func):
        func._test_generator = YAMLFileIterator(filename)
        func._test_generator_unpack = unpack
        return func

    return wrapper


class CSVFileIterator:
    """
    CSV file iterator.

    Args:
        filename (str): the filename to be read.

    Returns:
        iterator: iterator for the file contents as ``dict``.
    """
    def __init__(self, filename):
        self.filename = filename

    def __iter__(self):
        with open(self.filename, "r") as fp:
            reader = csv.reader(fp)
            data = list(reader)
            data_dict = []
            for line in data[1:]:
                d = OrderedDict()
                for index, value in enumerate(line):
                    d[data[0][index]] = value
                data_dict.append(d)
            data = data_dict

        return iter(data)


def csv_data(filename):
    """
    Create a data driven test using data from a CSV file.
    The first line of the CSV data is a header. Each column from the first
    is used as the parameter name of the decorated method.
    Each line is supplied as a ``dict``.

    Args:
        filename (str): the filename to be read.
    """
    def wrapper(func):
        func._test_generator = CSVFileIterator(filename)
        func._test_generator_unpack = False
        return func

    return wrapper


def product(*iterables, **repeat):
    """
    Create a data driven test using :py:func:`itertools.product`
    with a cartesian product of input ``*iterables``.

    Args:
        *iterables: iterables to be passed to :py:func:`itertools.product`
        repeat=1: number of repetitions.
    """
    return data(itertools.product(*iterables, **repeat),
                unpack=True)


def permutations(iterable, r=None):
    """
    Create a data driven test using :py:func:`itertools.permutations`
    with successive ``r`` length permutations of elements in the ``iterable``.

    Args:
        iterable: iterable to be passed to :py:func:`itertools.permutations`
        r=None (int or None): length of permutations
    """
    return data(itertools.permutations(iterable, r=r),
                unpack=True)


def combinations(iterable, r):
    """
    Create a data driven test using :py:func:`itertools.combinations`
    with successive ``r`` length subsequences of elements from ``iterable``.

    Args:
        iterable: iterable to be passed to :py:func:`itertools.combinations`
        r: length of combinations
    """
    return data(itertools.combinations(iterable, r=r),
                unpack=True)


def combinations_with_replacement(iterable, r):
    """
    Create a data driven test using :py:func:`itertools.combinations`
    with successive ``r`` length subsequences of elements from ``iterable``
    allowing individual elements to be repeated more than once.

    Args:
        iterable: iterable to be passed to :py:func:`itertools.combinations_with_replacement`
        r: length of combinations
    """
    return data(itertools.combinations_with_replacement(iterable, r=r),
                unpack=True)
